module linmod
  !module for linear algebra subroutines. Can be replaced with linpack or the like if available.
  implicit none
  private
  logical :: linmod_error
  public :: matinv,matinv_lapack,linmod_error
contains
  function matinv_lapack(A) result(res)
    implicit none
    real(8), dimension(:,:), intent(in) :: A
    real(8), dimension(size(A,1),size(A,2)) :: res
    real(8) :: work(size(A,1))
    integer, dimension(size(A,1)) :: ipiv
    integer :: n, info
    res=a
    n=size(A,1)
    call dgetrf(n,n,res,n,ipiv,info)
    if(info/=0) stop'Matrix is numerically singular'
    call dgetri(n,res,n,ipiv,work,n,info)
    if(info/=0) stop'Matrix is numerically singular'
  end function matinv_lapack

  function matinv(a)
    REAL(8), DIMENSION(:,:), INTENT(IN) :: a
    logical :: error
    real(8), dimension(size(a,1),size(a,1)) :: matinv
    real(8) :: b(size(a,1),1)
    linmod_error=.false.
    b=1.0d0
    matinv=a
    call gaussj(matinv,b,error)
    if(error) then
       matinv=-1.0d0
       linmod_error=.true.
    end if
  end function matinv

  SUBROUTINE gaussj(a,b,error)
    REAL(8), DIMENSION(:,:), INTENT(INOUT) :: a,b
    logical, intent(out) :: error
    INTEGER(4), DIMENSION(size(a,1)) :: ipiv,indxr,indxc
    LOGICAL, DIMENSION(size(a,1)) :: lpiv
    REAL(8) :: pivinv
    REAL(8), DIMENSION(size(a,1)) :: dumc
    INTEGER(4), TARGET :: irc(2)
    INTEGER(4) :: i,l,n
    INTEGER(4), POINTER :: irow,icol
    error=.false.
    n=assert_eq(size(a,1),size(a,2),size(b,1),'gaussj')
    irow => irc(1)
    icol => irc(2)
    ipiv=0
    do i=1,n
       lpiv = (ipiv == 0)
       irc=maxloc(abs(a),outerand(lpiv,lpiv))
       ipiv(icol)=ipiv(icol)+1
       if (ipiv(icol) > 1) then
          write(6,'("Fatal(gaussj): Matrix inversion failed (1).")')
          error=.true.
          return
       end if
       if (irow /= icol) then
          call swap(a(irow,:),a(icol,:))
          call swap(b(irow,:),b(icol,:))
       end if
       indxr(i)=irow
       indxc(i)=icol
       if (a(icol,icol) == 0.0d0) then
          write(6,'("Fatal(gaussj): Matrix inversion failed (2).")')
          error=.true.
          return
       end if
       pivinv=1.0d0/a(icol,icol)
       a(icol,icol)=1.0d0
       a(icol,:)=a(icol,:)*pivinv
       b(icol,:)=b(icol,:)*pivinv
       dumc=a(:,icol)
       a(:,icol)=0.0d0
       a(icol,icol)=pivinv
       a(1:icol-1,:)=a(1:icol-1,:)-outerprod(dumc(1:icol-1),a(icol,:))
       b(1:icol-1,:)=b(1:icol-1,:)-outerprod(dumc(1:icol-1),b(icol,:))
       a(icol+1:,:)=a(icol+1:,:)-outerprod(dumc(icol+1:),a(icol,:))
       b(icol+1:,:)=b(icol+1:,:)-outerprod(dumc(icol+1:),b(icol,:))
    end do
    do l=n,1,-1
       call swap(a(:,indxr(l)),a(:,indxc(l)))
    end do
  END SUBROUTINE gaussj

  FUNCTION assert_eq(n1,n2,n3,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER(4), INTENT(IN) :: n1,n2,n3
    INTEGER(4) :: assert_eq
    if (n1 == n2 .and. n2 == n3) then
       assert_eq=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq3'
    end if
  END FUNCTION assert_eq

  SUBROUTINE nrerror(string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    write (*,*) 'nrerror: ',string
    STOP 'program terminated by nrerror'
  END SUBROUTINE nrerror

  FUNCTION outerand(a,b)
    LOGICAL, DIMENSION(:), INTENT(IN) :: a,b
    LOGICAL, DIMENSION(size(a),size(b)) :: outerand
    outerand = spread(a,dim=2,ncopies=size(b)) .and. &
         spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerand

  FUNCTION outerprod(a,b)
    REAL(8), DIMENSION(:), INTENT(IN) :: a,b
    REAL(8), DIMENSION(size(a),size(b)) :: outerprod
    outerprod = spread(a,dim=2,ncopies=size(b)) * &
         spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerprod

  SUBROUTINE swap(a,b)
    REAL(8), DIMENSION(:), INTENT(INOUT) :: a,b
    REAL(8), DIMENSION(SIZE(a)) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap
end module linmod

module marsaglia
  !module for the random number generation.
  implicit none
  private
  type kisseed
     integer(4) :: x,y,z,w
  end type kisseed
  public :: kiss, kisset,kissuni,kissgaus,kisseed, setseed
  INTEGER(4) :: x=123456789, y=362436069, z=521288629, w=916191069
contains
  function kiss()
    integer(4) :: kiss
    ! The  KISS (Keep It Simple Stupid) random number generator. Combines:
    ! (1) The congruential generator x(n)=69069*x(n-1)+1327217885, period 2^32.
    ! (2) A 3-shift shift-register generator, period 2^32-1,
    ! (3) Two 16-bit multiply-with-carry generators, period 597273182964842497>2^59
    !  Overall period>2^123;  Default seeds x,y,z,w.
    !  Set your own seeds with statement i=kisset(ix,iy,iz,iw).

    x = 69069 * x + 1327217885
    y = m (m (m (y, 13), - 17), 5)
    z = 18000 * iand (z, 65535) + ishft (z, - 16)
    w = 30903 * iand (w, 65535) + ishft (w, - 16)
    kiss = x + y + ishft (z, 16) + w
  contains
    function m(k, n)
      integer :: m, k, n
      m = ieor (k, ishft (k, n) )
    end function m
  end function kiss

  function kisset (inp)
    integer(4) :: kisset
    type(kisseed), intent(in) :: inp
    x = inp%x
    y = inp%y
    z = inp%z
    w = inp%w
    kisset = 1
  end function kisset

  function kissuni()
    !Produces one of 2^32 possible realizations in the (0,1) interval (not including 0 and 1).
    real(8) :: kissuni
    kissuni=(dble(kiss())+2147483649.0d0)/4294967297.0d0
  end function kissuni

  function kissgaus()
    !normal rng based on simple box muller transformation. Slow.
    REAL(8) :: kissgaus
    REAL(8) :: rsq,v1,v2
    REAL(8), SAVE :: g
    LOGICAL, SAVE :: gaus_stored=.false.
    logical :: notdone
    if (gaus_stored) then
       kissgaus=g
       gaus_stored=.false.
    else
       notdone=.true.
       do while(notdone)
          v1=kissuni()
          v2=kissuni()
          v1=2.0d0*v1-1.0d0
          v2=2.0d0*v2-1.0d0
          rsq=v1**2+v2**2
          if (rsq > 0.0d0 .and. rsq < 1.0d0) then
             rsq=sqrt(-2.0d0*log(rsq)/rsq)
             if(rsq==rsq) notdone=.false.
          end if
       end do
       kissgaus=v1*rsq
       g=v2*rsq
       gaus_stored=.true.
    end if
  end function kissgaus

  subroutine setseed(ms)
    !kisseed(2345678,87656789,5434567,43212345)
    type(kisseed), intent(in) :: ms !Set master seed here.
    type(kisseed) :: mys
    integer :: i,j,imc
    logical :: tryagain
    i=kisset(ms)
    do i=1,10000
       j=kiss()
    end do
    tryagain=.true.
    do while (tryagain)
       do i=1,4 !x and y seed can be 32 bit integers. z and w seeds no more than 31 bit integers and not multiples of 7559.
          mys%x=kiss()
          mys%y=kiss()
          mys%z=ishft(kiss(),-2)
          mys%w=ishft(kiss(),-2)
       end do
       if(mod(mys%z,7559)/=0 .and. mod(mys%w,7559)/=0) tryagain=.false.
    end do
    i=kisset(mys) !seed the process
    do i=1,10000 !warm up the sequence. We are set to go.
       j=kiss()
    end do
  end subroutine setseed
end module marsaglia

module tools
  implicit none
contains
  function findndx(arr,val) !Determine the indx such that arr(ndx)<=val<arr(ndx+1). arr must be sorted in ascending order
    real(8), dimension(:), intent(in) :: arr
    real(8), intent(in) :: val
    integer :: findndx
    integer :: lb,ub,dim,i
    dim=size(arr)
    if(dim<2) stop'Fatal(findndx): dimension of input array <2'
    lb=1
    ub=dim
    if(arr(lb)>val .or. arr(ub)<val) then
       write(6,'(f20.10)'),arr
       write(6,'(" ")')
       write(6,'(f20.10)'),val
       stop'Fatal(findndx): val not in arr support.'
    end if
    if(arr(ub) == val) then
       findndx=ub
       return
    end if
    findndx=(ub+lb)/2
    i=1
    do while ((arr(findndx) .gt. val .or. arr(findndx+1) .le. val) .and. i<100) !notice, a bit of ambiguity in the return
       !if val hits exactly a knot point.
       if(arr(findndx) .gt. val) then
          ub=findndx
       else
          lb=findndx
       end if
       findndx=(lb+ub)/2
       i=i+1
    end do
    if(i==100) then
       write(6,'(f20.10)'),arr
       write(6,'(" ")')
       write(6,'(f20.10)'),val
       stop'Fatal(findndx): solver did not reach criterion.'
    end if
  end function findndx
end module tools

module optim
  private
  public dzbrent,lowtri,brent,dfpmin,numgrad_cent,getrs_nr
  INTERFACE assert_eq
     MODULE PROCEDURE assert_eq2,assert_eq3,assert_eq4,assert_eqn
  END INTERFACE assert_eq
  INTERFACE swap
     MODULE PROCEDURE swap_i,swap_r,swap_d,swap_rv,swap_dv, &
          masked_swap_rs,masked_swap_ds,masked_swap_rv,masked_swap_rm
  END INTERFACE swap
  INTERFACE outerprod
     MODULE PROCEDURE outerprod_r,outerprod_d
  END INTERFACE outerprod
  INTERFACE imaxloc
     MODULE PROCEDURE imaxloc_d,imaxloc_r,imaxloc_i
  END INTERFACE imaxloc


contains

  subroutine getrs_nr(mata,dim,matb)
    implicit none
    integer, intent(in) :: dim
    real(8), dimension(dim,dim), intent(in) :: mata
    real(8), dimension(dim), intent(inout) :: matb
    integer(4), dimension(dim) :: indx
    real(8), dimension(dim,dim) :: my_mata
    real(8) :: my_d
    my_mata=mata
    call ludcmp(my_mata,indx,my_d) !LU decomposition returned in my_mata and indx (my_d is not used)
    call lubksb(my_mata,indx,matb) !solve AX=B, where matb is B vector. Solution is returned in matb.
  end subroutine getrs_nr

  SUBROUTINE lubksb(a,indx,b)
    IMPLICIT NONE
    REAL(8), DIMENSION(:,:), INTENT(IN) :: a
    INTEGER(4), DIMENSION(:), INTENT(IN) :: indx
    REAL(8), DIMENSION(:), INTENT(INOUT) :: b
    INTEGER(4) :: i,n,ii,ll
    REAL(8) :: summ
    n=assert_eq(size(a,1),size(a,2),size(indx),'lubksb')
    ii=0
    do i=1,n
       ll=indx(i)
       summ=b(ll)
       b(ll)=b(i)
       if (ii /= 0) then
          summ=summ-dot_product(a(i,ii:i-1),b(ii:i-1))
       else if (summ /= 0.0d0) then
          ii=i
       end if
       b(i)=summ
    end do
    do i=n,1,-1
       b(i) = (b(i)-dot_product(a(i,i+1:n),b(i+1:n)))/a(i,i)
    end do
  END SUBROUTINE lubksb

  SUBROUTINE ludcmp(a,indx,d)
    IMPLICIT NONE
    REAL(8), DIMENSION(:,:), INTENT(INOUT) :: a
    INTEGER(4), DIMENSION(:), INTENT(OUT) :: indx
    REAL(8), INTENT(OUT) :: d
    REAL(8), DIMENSION(size(a,1)) :: vv
    REAL(8), PARAMETER :: TINY=1.0e-20
    INTEGER(4) :: j,n,imax
    n=assert_eq(size(a,1),size(a,2),size(indx),'ludcmp')
    d=1.0d0
    vv=maxval(abs(a),dim=2)
    if (any(vv == 0.0d0)) call nrerror('singular matrix in ludcmp')
    vv=1.0d0/vv
    do j=1,n
       imax=(j-1)+imaxloc(vv(j:n)*abs(a(j:n,j)))
       if (j /= imax) then
          call swap(a(imax,:),a(j,:))
          d=-d
          vv(imax)=vv(j)
       end if
       indx(j)=imax
       if (a(j,j) == 0.0d0) a(j,j)=TINY
       a(j+1:n,j)=a(j+1:n,j)/a(j,j)
       a(j+1:n,j+1:n)=a(j+1:n,j+1:n)-outerprod(a(j+1:n,j),a(j,j+1:n))
    end do
  END SUBROUTINE ludcmp

  function numgrad_cent(x,func) !calculate numerical gradient (central) of func at x.
    implicit none
    real(8), dimension(:), intent(in) :: x
    real(8), dimension(size(x)) :: numgrad_cent
    INTERFACE
       FUNCTION func(xx) !function evaluation
         IMPLICIT NONE
         REAL(8), DIMENSION(:), INTENT(IN) :: xx
         REAL(8) :: func
       END FUNCTION func
    END INTERFACE
    real(8), parameter :: h=.5e-7
    real(8), dimension(size(x)) :: vec
    integer :: i
    do i=1,size(x)
       vec=0
       vec(i)=1.0d0
       numgrad_cent(i)=(func(x+vec*h)-func(x-vec*h))/(2.0d0*h)
    end do
  end function numgrad_cent

  SUBROUTINE dfpmin(p,gtol,iter,fret,func,dfunc,ok) !minimization. initial guess, p. gtol is tolerance criterion on gradient
    !iter is number of iterations performed, fret is minimized value. func is function to be minimized. dfunc is gradient of func.
    IMPLICIT NONE
    INTEGER(4), INTENT(OUT) :: iter
    REAL(8), INTENT(IN) :: gtol
    REAL(8), INTENT(OUT) :: fret
    REAL(8), DIMENSION(:), INTENT(INOUT) :: p
    logical, intent(out) :: ok
    INTERFACE
       FUNCTION func(p) !function evaluation
         IMPLICIT NONE
         REAL(8), DIMENSION(:), INTENT(IN) :: p
         REAL(8) :: func
       END FUNCTION func
       FUNCTION dfunc(p) !gradient
         IMPLICIT NONE
         REAL(8), DIMENSION(:), INTENT(IN) :: p
         REAL(8), DIMENSION(size(p)) :: dfunc
       END FUNCTION dfunc
    END INTERFACE
    INTEGER(4), PARAMETER :: ITMAX=10000
    REAL(8), PARAMETER :: STPMX=100.0d0,EPS=epsilon(p),TOLX=4.0d0*EPS
    INTEGER(4) :: its
    LOGICAL :: check
    REAL(8) :: den,fac,fad,fae,fp,stpmax,sumdg,sumxi
    REAL(8), DIMENSION(size(p)) :: dg,g,hdg,pnew,xi
    REAL(8), DIMENSION(size(p),size(p)) :: hessin
    ok=.true.
    fp=func(p)
    g=dfunc(p)
    call unit_matrix(hessin)
    xi=-g
    stpmax=STPMX*max(vabs(p),dble(size(p)))
    do its=1,ITMAX
       iter=its
       call lnsrch(p,fp,g,xi,pnew,fret,stpmax,check,func,ok)
       if(.not. ok) return
       fp=fret
       xi=pnew-p
       p=pnew
       if (maxval(abs(xi)/max(abs(p),1.0d0)) < TOLX) RETURN !if the updated p is not moving, then exit.
       dg=g
       g=dfunc(p)
       den=max(fret,1.0d0)
       if (maxval(abs(g)*max(abs(p),1.0d0)/den) < gtol) RETURN !if gradient is zero, then exit.
       dg=g-dg
       hdg=matmul(hessin,dg)
       fac=dot_product(dg,xi)
       fae=dot_product(dg,hdg)
       sumdg=dot_product(dg,dg)
       sumxi=dot_product(xi,xi)
       if (fac > sqrt(EPS*sumdg*sumxi)) then
          fac=1.0d0/fac
          fad=1.0d0/fae
          dg=fac*xi-fad*hdg
          hessin=hessin+fac*outerprod(xi,xi)-fad*outerprod(hdg,hdg)+fae*outerprod(dg,dg)
       end if
       xi=-matmul(hessin,g)
    end do
    call nrerror('dfpmin: too many iterations')
  END SUBROUTINE dfpmin

  FUNCTION dzbrent(func,x1,x2,tol)
    IMPLICIT NONE
    REAL(8), INTENT(IN) :: x1,x2,tol
    REAL(8) :: dzbrent
    INTERFACE
       FUNCTION func(x)
         IMPLICIT NONE
         REAL(8), INTENT(IN) :: x
         REAL(8) :: func
       END FUNCTION func
    END INTERFACE
    INTEGER(4), PARAMETER :: ITMAX=100
    REAL(8), PARAMETER :: EPS=epsilon(x1)
    INTEGER(4) :: iter
    REAL(8) :: a,b,c,d,e,fa,fb,fc,p,q,r,s,tol1,xm
    a=x1
    b=x2
    fa=func(a)
    fb=func(b)
    if ((fa > 0.0d0 .and. fb > 0.0d0) .or. (fa < 0.0d0 .and. fb < 0.0d0)) then
       write(6,'("root must be bracketed for dzbrent",4f20.10)'),a,b,fa,fb
       stop
    end if
    c=b
    fc=fb
    do iter=1,ITMAX
       if ((fb > 0.0d0 .and. fc > 0.0d0) .or. (fb < 0.0d0 .and. fc < 0.0d0)) then
          c=a
          fc=fa
          d=b-a
          e=d
       end if
       if (abs(fc) < abs(fb)) then
          a=b
          b=c
          c=a
          fa=fb
          fb=fc
          fc=fa
       end if
       tol1=2.0d0*EPS*abs(b)+0.5d0*tol
       xm=0.5d0*(c-b)
       if (abs(xm) <= tol1 .or. fb == 0.0d0) then
          dzbrent=b
          RETURN
       end if
       if (abs(e) >= tol1 .and. abs(fa) > abs(fb)) then
          s=fb/fa
          if (a == c) then
             p=2.0d0*xm*s
             q=1.0d0-s
          else
             q=fa/fc
             r=fb/fc
             p=s*(2.0d0*xm*q*(q-r)-(b-a)*(r-1.0d0))
             q=(q-1.0d0)*(r-1.0d0)*(s-1.0d0)
          end if
          if (p > 0.0d0) q=-q
          p=abs(p)
          if (2.0d0*p  <  min(3.0d0*xm*q-abs(tol1*q),abs(e*q))) then
             e=d
             d=p/q
          else
             d=xm
             e=d
          end if
       else
          d=xm
          e=d
       end if
       a=b
       fa=fb
       b=b+merge(d,sign(tol1,xm), abs(d) > tol1 )
       fb=func(b)
    end do
    write(6,'("dzbrent: exceeded maximum iterations")')
    dzbrent=b
  END FUNCTION dzbrent

  subroutine lowtri(dim,a,b,x) !solves the lower triangular system ax=b
    implicit none
    integer, intent(in) :: dim
    real(8), intent(in) :: a(dim,dim) !this is obviously wasteful since a little less than half of the matrix contains zeros.
    real(8), intent(in) :: b(dim)
    real(8), intent(out) :: x(dim)
    integer :: i,j
    real(8) :: sum
    !solve by forward substitution
    do i=1,dim
       sum=0.0d0
       do j=1,i-1
          sum=sum+a(i,j)*x(j)
       end do
       x(i)=(b(i)-sum)/a(i,i)
    end do
  end subroutine lowtri

  FUNCTION brent(ax,bx,cx,func,tol,xmin) !minimizer
    IMPLICIT NONE
    REAL(8), INTENT(IN) :: ax,bx,cx,tol
    REAL(8), INTENT(OUT) :: xmin
    REAL(8) :: brent
    INTERFACE
       FUNCTION func(x)
         IMPLICIT NONE
         REAL(8), INTENT(IN) :: x
         REAL(8) :: func
       END FUNCTION func
    END INTERFACE
    INTEGER(4), PARAMETER :: ITMAX=100
    REAL(8), PARAMETER :: CGOLD=0.3819660,ZEPS=1.0e-3*epsilon(ax)
    INTEGER(4) :: iter
    REAL(8) :: a,b,d,e,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm
    a=min(ax,cx)
    b=max(ax,cx)
    v=bx
    w=v
    x=v
    e=0.0d0
    fx=func(x)
    fv=fx
    fw=fx
    do iter=1,ITMAX
       xm=0.5d0*(a+b)
       tol1=tol*abs(x)+ZEPS
       tol2=2.0d0*tol1
       if (abs(x-xm) <= (tol2-0.5d0*(b-a))) then
          xmin=x
          brent=fx
          RETURN
       end if
       if (abs(e) > tol1) then
          r=(x-w)*(fx-fv)
          q=(x-v)*(fx-fw)
          p=(x-v)*q-(x-w)*r
          q=2.0d0*(q-r)
          if (q > 0.0d0) p=-p
          q=abs(q)
          etemp=e
          e=d
          if (abs(p) >= abs(0.5d0*q*etemp) .or. &
               p <= q*(a-x) .or. p >= q*(b-x)) then
             e=merge(a-x,b-x, x >= xm )
             d=CGOLD*e
          else
             d=p/q
             u=x+d
             if (u-a < tol2 .or. b-u < tol2) d=sign(tol1,xm-x)
          end if
       else
          e=merge(a-x,b-x, x >= xm )
          d=CGOLD*e
       end if
       u=merge(x+d,x+sign(tol1,d), abs(d) >= tol1 )
       fu=func(u)
       if (fu <= fx) then
          if (u >= x) then
             a=x
          else
             b=x
          end if
          call shft(v,w,x,u)
          call shft(fv,fw,fx,fu)
       else
          if (u < x) then
             a=u
          else
             b=u
          end if
          if (fu <= fw .or. w == x) then
             v=w
             fv=fw
             w=u
             fw=fu
          else if (fu <= fv .or. v == x .or. v == w) then
             v=u
             fv=fu
          end if
       end if
    end do
    write(6,'("brent: exceed maximum iterations")')
  CONTAINS
    SUBROUTINE shft(a,b,c,d)
      REAL(8), INTENT(OUT) :: a
      REAL(8), INTENT(INOUT) :: b,c
      REAL(8), INTENT(IN) :: d
      a=b
      b=c
      c=d
    END SUBROUTINE shft
  END FUNCTION brent

  FUNCTION assert_eq2(n1,n2,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER, INTENT(IN) :: n1,n2
    INTEGER :: assert_eq2
    if (n1 == n2) then
       assert_eq2=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq2'
    end if
  END FUNCTION assert_eq2

  FUNCTION assert_eq3(n1,n2,n3,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER, INTENT(IN) :: n1,n2,n3
    INTEGER :: assert_eq3
    if (n1 == n2 .and. n2 == n3) then
       assert_eq3=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq3'
    end if
  END FUNCTION assert_eq3

  FUNCTION assert_eq4(n1,n2,n3,n4,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER, INTENT(IN) :: n1,n2,n3,n4
    INTEGER :: assert_eq4
    if (n1 == n2 .and. n2 == n3 .and. n3 == n4) then
       assert_eq4=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq4'
    end if
  END FUNCTION assert_eq4
  !BL
  FUNCTION assert_eqn(nn,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER, DIMENSION(:), INTENT(IN) :: nn
    INTEGER :: assert_eqn
    if (all(nn(2:) == nn(1))) then
       assert_eqn=nn(1)
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eqn'
    end if
  END FUNCTION assert_eqn

  SUBROUTINE nrerror(string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    write (*,*) 'nrerror: ',string
    STOP 'program terminated by nrerror'
  END SUBROUTINE nrerror

  SUBROUTINE swap_i(a,b)
    INTEGER(4), INTENT(INOUT) :: a,b
    INTEGER(4) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_i

  SUBROUTINE swap_r(a,b)
    REAL(4), INTENT(INOUT) :: a,b
    REAL(4) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_r

  SUBROUTINE swap_d(a,b)
    REAL(8), INTENT(INOUT) :: a,b
    REAL(8) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_d

  SUBROUTINE swap_dv(a,b)
    REAL(8), DIMENSION(:), INTENT(INOUT) :: a,b
    REAL(8), DIMENSION(SIZE(a)) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_dv

  SUBROUTINE swap_rv(a,b)
    REAL(4), DIMENSION(:), INTENT(INOUT) :: a,b
    REAL(4), DIMENSION(SIZE(a)) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_rv

  SUBROUTINE masked_swap_rs(a,b,mask)
    REAL(4), INTENT(INOUT) :: a,b
    LOGICAL, INTENT(IN) :: mask
    REAL(4) :: swp
    if (mask) then
       swp=a
       a=b
       b=swp
    end if
  END SUBROUTINE masked_swap_rs

  SUBROUTINE masked_swap_ds(a,b,mask)
    REAL(8), INTENT(INOUT) :: a,b
    LOGICAL, INTENT(IN) :: mask
    REAL(8) :: swp
    if (mask) then
       swp=a
       a=b
       b=swp
    end if
  END SUBROUTINE masked_swap_ds

  SUBROUTINE masked_swap_rv(a,b,mask)
    REAL(4), DIMENSION(:), INTENT(INOUT) :: a,b
    LOGICAL, DIMENSION(:), INTENT(IN) :: mask
    REAL(4), DIMENSION(size(a)) :: swp
    where (mask)
       swp=a
       a=b
       b=swp
    end where
  END SUBROUTINE masked_swap_rv

  SUBROUTINE masked_swap_rm(a,b,mask)
    REAL(4), DIMENSION(:,:), INTENT(INOUT) :: a,b
    LOGICAL, DIMENSION(:,:), INTENT(IN) :: mask
    REAL(4), DIMENSION(size(a,1),size(a,2)) :: swp
    where (mask)
       swp=a
       a=b
       b=swp
    end where
  END SUBROUTINE masked_swap_rm

  FUNCTION outerprod_r(a,b)
    REAL(4), DIMENSION(:), INTENT(IN) :: a,b
    REAL(4), DIMENSION(size(a),size(b)) :: outerprod_r
    outerprod_r = spread(a,dim=2,ncopies=size(b)) * spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerprod_r

  FUNCTION outerprod_d(a,b)
    REAL(8), DIMENSION(:), INTENT(IN) :: a,b
    REAL(8), DIMENSION(size(a),size(b)) :: outerprod_d
    outerprod_d = spread(a,dim=2,ncopies=size(b)) * spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerprod_d

  FUNCTION imaxloc_r(arr)
    REAL(4), DIMENSION(:), INTENT(IN) :: arr
    INTEGER(4) :: imaxloc_r
    INTEGER(4), DIMENSION(1) :: imax
    imax=maxloc(arr(:))
    imaxloc_r=imax(1)
  END FUNCTION imaxloc_r

  FUNCTION imaxloc_d(arr)
    REAL(8), DIMENSION(:), INTENT(IN) :: arr
    INTEGER :: imaxloc_d
    INTEGER, DIMENSION(1) :: imax
    imax=maxloc(arr(:))
    imaxloc_d=imax(1)
  END FUNCTION imaxloc_d

  FUNCTION imaxloc_i(iarr)
    INTEGER(4), DIMENSION(:), INTENT(IN) :: iarr
    INTEGER(4), DIMENSION(1) :: imax
    INTEGER(4) :: imaxloc_i
    imax=maxloc(iarr(:))
    imaxloc_i=imax(1)
  END FUNCTION imaxloc_i

  SUBROUTINE unit_matrix(mat)
    REAL(8), DIMENSION(:,:), INTENT(OUT) :: mat
    INTEGER(4) :: i,n
    n=min(size(mat,1),size(mat,2))
    mat(:,:)=0.0d0
    do i=1,n
       mat(i,i)=1.0d0
    end do
  END SUBROUTINE unit_matrix

  FUNCTION vabs(v)
    REAL(8), DIMENSION(:), INTENT(IN) :: v
    REAL(8) :: vabs
    vabs=sqrt(dot_product(v,v))
  END FUNCTION vabs

  SUBROUTINE lnsrch(xold,fold,g,p,x,f,stpmax,check,func,ok)
    IMPLICIT NONE
    REAL(8), DIMENSION(:), INTENT(IN) :: xold,g
    REAL(8), DIMENSION(:), INTENT(INOUT) :: p
    REAL(8), INTENT(IN) :: fold,stpmax
    REAL(8), DIMENSION(:), INTENT(OUT) :: x
    REAL(8), INTENT(OUT) :: f
    LOGICAL, INTENT(OUT) :: check
    logical, intent(out) :: ok
    INTERFACE
       FUNCTION func(x)
         IMPLICIT NONE
         REAL(8) :: func
         REAL(8), DIMENSION(:), INTENT(IN) :: x
       END FUNCTION func
    END INTERFACE
    REAL(8), PARAMETER :: ALF=1.0e-4,TOLX=epsilon(x)
    INTEGER(4) :: ndum
    REAL(8) :: a,alam,alam2,alamin,b,disc,f2,pabs,rhs1,rhs2,slope,tmplam
    ndum=assert_eq(size(g),size(p),size(x),size(xold),'lnsrch')
    check=.false.
    ok=.true.
    pabs=vabs(p(:))
    if (pabs > stpmax) p(:)=p(:)*stpmax/pabs
    slope=dot_product(g,p)
    if (slope >= 0.0d0) then
       ok=.false.
       return
!!$       call nrerror('roundoff problem in lnsrch')
    end if
    alamin=TOLX/maxval(abs(p(:))/max(abs(xold(:)),1.0d0))
    alam=1.0d0
    do
       x(:)=xold(:)+alam*p(:)
       f=func(x)
       if (alam < alamin) then
          x(:)=xold(:)
          check=.true.
          RETURN
       else if (f <= fold+ALF*alam*slope) then
          RETURN
       else
          if (alam == 1.0d0) then
             tmplam=-slope/(2.0d0*(f-fold-slope))
          else
             rhs1=f-fold-alam*slope
             rhs2=f2-fold-alam2*slope
             a=(rhs1/alam**2-rhs2/alam2**2)/(alam-alam2)
             b=(-alam2*rhs1/alam**2+alam*rhs2/alam2**2)/&
                  (alam-alam2)
             if (a == 0.0d0) then
                tmplam=-slope/(2.0d0*b)
             else
                disc=b*b-3.0d0*a*slope
                if (disc < 0.0d0) then
                   tmplam=0.5d0*alam
                else if (b <= 0.0d0) then
                   tmplam=(-b+sqrt(disc))/(3.0d0*a)
                else
                   tmplam=-slope/(b+sqrt(disc))
                end if
             end if
             if (tmplam > 0.5d0*alam) tmplam=0.5d0*alam
          end if
       end if
       alam2=alam
       f2=f
       alam=max(tmplam,0.1d0*alam)
    end do
  END SUBROUTINE lnsrch

end module optim

module NM !Simple Nelder-Meade minimization
  implicit none
  INTERFACE swap
     MODULE PROCEDURE swap_d,swap_dv
  END INTERFACE
  public NMmin
Contains
  subroutine NMmin(inp,outp,func,tol,err) !Simple call routine for damoeba. Makes choices for precision, itmax, etc.
    implicit none
    real(8), dimension(:), intent(in) :: inp !guess
    real(8), dimension(size(inp)), intent(out) :: outp !arg min of func
    real(8), intent(in) :: tol
    integer, intent(out) :: err !error flag
    integer :: i,ndim
    real(8), dimension(size(inp)+1,size(inp)) :: p
    real(8), dimension(size(inp)+1) :: y
    integer, parameter :: itmax=1000
    INTERFACE
       FUNCTION func(x)
         IMPLICIT NONE
         REAL(8), DIMENSION(:), INTENT(IN) :: x
         REAL(8) :: func
       END FUNCTION func
    END INTERFACE
    err=0
    ndim=size(inp)
    p(1,:)=inp
    do i=1,ndim
       p(i+1,:)=inp
       p(i+1,i)=p(i+1,i)*1.5d0 !add 1.0d0 to element i for each vector in the simplex
    end do
    do i=1,ndim+1
       y(i)=func(p(i,:)) !function values for each column in the simplex
    end do
    call damoeba(p,y,tol,func,i,itmax)
    if(i == itmax) err=1 !exit at max iterations
    outp=sum(p,dim=1)/dble(ndim+1) !let the argmin output be the average over the simplex.
  end subroutine NMmin

  SUBROUTINE damoeba(p,y,ftol,func,iter,itmax)
    IMPLICIT NONE
    INTEGER, INTENT(OUT) :: iter
    REAL(8), INTENT(IN) :: ftol
    REAL(8), DIMENSION(:), INTENT(INOUT) :: y
    REAL(8), DIMENSION(:,:), INTENT(INOUT) :: p
    integer, intent(in) :: itmax
    INTERFACE
       FUNCTION func(x)
         IMPLICIT NONE
         REAL(8), DIMENSION(:), INTENT(IN) :: x
         REAL(8) :: func
       END FUNCTION func
    END INTERFACE
    REAL(8), PARAMETER :: TINY=1.0e-16
    INTEGER(4) :: ihi,ndim
    REAL(8), DIMENSION(size(p,2)) :: psum
    call damoeba_private
  CONTAINS

    SUBROUTINE damoeba_private
      IMPLICIT NONE
      INTEGER(4) :: i,ilo,inhi,j
      REAL(8) :: rtol,ysave,ytry,ytmp
      ndim=assert_eq(size(p,2),size(p,1)-1,size(y)-1,'damoeba')
      iter=0
      psum(:)=sum(p(:,:),dim=1)
      do
         ilo=iminloc(y(:))
         ihi=imaxloc(y(:))
         ytmp=y(ihi)
         y(ihi)=y(ilo)
         inhi=imaxloc(y(:))
         y(ihi)=ytmp
         rtol=dble(2.0)*abs(y(ihi)-y(ilo))/(abs(y(ihi))+abs(y(ilo))+TINY)
         if (rtol < ftol) then
            call swap(y(1),y(ilo))
            call swap(p(1,:),p(ilo,:))
            RETURN
         end if
         if (iter >= ITMAX) RETURN
         ytry=damotry(dble(-1.0))
         iter=iter+1
         if (ytry <= y(ilo)) then
!!$            open(10,file='/home/raslentz/damoeba_improvement.csv')
!!$            do j=1,ndim
!!$               write(10,'(f30.15)'),p(ihi,j)
!!$            end do
!!$            close(10)
            ytry=damotry(dble(2.0))
            iter=iter+1
         else if (ytry >= y(inhi)) then
            ysave=y(ihi)
            ytry=damotry(dble(0.5))
            iter=iter+1
            if (ytry >= ysave) then
               p(:,:)=dble(0.5)*(p(:,:)+spread(p(ilo,:),1,size(p,1)))
               do i=1,ndim+1
                  if (i /= ilo) y(i)=func(p(i,:))
               end do
               iter=iter+ndim
               psum(:)=sum(p(:,:),dim=1)
            end if
         end if
      end do
    END SUBROUTINE damoeba_private

    FUNCTION damotry(fac)
      IMPLICIT NONE
      REAL(8), INTENT(IN) :: fac
      REAL(8) :: damotry
      REAL(8) :: fac1,fac2,ytry
      REAL(8), DIMENSION(size(p,2)) :: ptry
      fac1=(dble(1.0)-fac)/ndim
      fac2=fac1-fac
      ptry(:)=psum(:)*fac1-p(ihi,:)*fac2
      ytry=func(ptry)
      if (ytry < y(ihi)) then
         y(ihi)=ytry
         psum(:)=psum(:)-p(ihi,:)+ptry(:)
         p(ihi,:)=ptry(:)
      end if
      damotry=ytry
    END FUNCTION damotry
  END SUBROUTINE damoeba

  FUNCTION assert_eq(n1,n2,n3,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER, INTENT(IN) :: n1,n2,n3
    INTEGER :: assert_eq
    if (n1 == n2 .and. n2 == n3) then
       assert_eq=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq'
    end if
  END FUNCTION assert_eq

  FUNCTION iminloc(arr)
    REAL(8), DIMENSION(:), INTENT(IN) :: arr
    INTEGER, DIMENSION(1) :: imin
    INTEGER :: iminloc
    imin=minloc(arr(:))
    iminloc=imin(1)
  END FUNCTION iminloc

  FUNCTION imaxloc(arr)
    REAL(8), DIMENSION(:), INTENT(IN) :: arr
    INTEGER :: imaxloc
    INTEGER, DIMENSION(1) :: imax
    imax=maxloc(arr(:))
    imaxloc=imax(1)
  END FUNCTION imaxloc

  SUBROUTINE nrerror(string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    write (*,*) 'nrerror: ',string
    STOP 'program terminated by nrerror'
  END SUBROUTINE nrerror

  SUBROUTINE swap_d(a,b)
    REAL(8), INTENT(INOUT) :: a,b
    REAL(8) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_d

  SUBROUTINE swap_dv(a,b)
    REAL(8), DIMENSION(:), INTENT(INOUT) :: a,b
    REAL(8), DIMENSION(SIZE(a)) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap_dv
end module NM

module sort3
  implicit none
  private
  interface hpsort
     module procedure hpsort_real, hpsort_int, hpsort_real_pert, hpsort_int_pert
  end interface
  public HPSORT

contains

  !*****************************************************
  !*  Sorts an array RA of length N in ascending order *
  !*                by the Heapsort method             *
  !* ------------------------------------------------- *
  !* INPUTS:                                           *
  !*	    N	  size of table RA                     *
  !*         RA	  table to be sorted            	    *
  !* OUTPUT:                                           *
  !*	    RA    table sorted in ascending order     *
  !*                                                   *
  !* NOTE: The Heapsort method is a N Log2 N routine,  *
  !*       and can be used for very large arrays.      *
  !*****************************************************
  SUBROUTINE HPSORT_real(N,RA)
    implicit none
    integer, intent(in) :: N
    real(8), intent(inout) :: RA(N)
    integer :: L,IR,I,J
    real(8) :: RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.
10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
    else
       RRA=RA(IR)
       RA(IR)=RA(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    goto 10
  END SUBROUTINE HPSORT_real

  SUBROUTINE HPSORT_real_pert(N,RA,INDX) !
    implicit none
    integer, intent(in) :: N
    real(8), intent(inout) :: RA(N)
    integer, intent(inout) :: INDX(N)
    integer :: L,IR,I,J,IRRA
    real(8) :: RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.

!!$    do i=1,N !Initial order
!!$	indx(i)=i
!!$    end do

10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
       irra=indx(L)
    else
       RRA=RA(IR)
       irra=indx(IR)
       RA(IR)=RA(1)
       indx(IR)=indx(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
	   indx(1)=irra
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
	   indx(I)=indx(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    indx(I)=irra
    goto 10
  END SUBROUTINE HPSORT_real_pert

  SUBROUTINE HPSORT_int_pert(N,RA,INDX) !
    implicit none
    integer, intent(in) :: N
    integer, intent(inout) :: RA(N)
    integer, intent(inout) :: INDX(N)
    integer :: L,IR,I,J,IRRA
    integer :: RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.

!!$    do i=1,N !Initial order
!!$	indx(i)=i
!!$    end do

10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
       irra=indx(L)
    else
       RRA=RA(IR)
       irra=indx(IR)
       RA(IR)=RA(1)
       indx(IR)=indx(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
	   indx(1)=irra
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
	   indx(I)=indx(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    indx(I)=irra
    goto 10
  END SUBROUTINE HPSORT_int_pert

  SUBROUTINE HPSORT_int(N,RA)
    implicit none
    integer, intent(in) :: N
    integer(4), intent(inout) :: RA(N)
    integer :: L,IR,I,J,RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.
10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
    else
       RRA=RA(IR)
       RA(IR)=RA(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    goto 10
  END SUBROUTINE HPSORT_int
end module sort3

