module cem !This module contains the code for the CEM estimation
  use estim
  use paths
  use mpimod 
  use sort3 
  use marsaglia
  implicit none
  real(8) :: diff 
  integer, allocatable :: firm_size_order(:),ftype(:)
  integer :: mm_maxit,firmcount,cem_counter
  integer(4) :: mobcatcount(mobhetdim)
	logical :: e_settotruth=.false.
contains

  subroutine cem_estim(kk,ll,outname,seed,llout_BLM,llout_LPR) !this is the control of the CEM-algorithm using the C-, E-, and M-step subroutines.
    integer, intent(in) :: kk,ll
    character(*), intent(in) :: outname
    integer, intent(in) :: seed
    real(8), intent(out) :: llout_BLM, llout_LPR
    type(modelpar) :: oldep
    integer :: i, j, ccount
    integer, parameter :: maxit=50
    real(8), parameter :: tol=1.0e-8, tolc= 1.0e-8
    real(8) :: loglike,oldll, diffc, diffl
    logical :: nanflag
    integer :: ii,jj, mob, firm_alive, percentile, step
    real(8) :: btwvar, withinvar

    if(allocated(ftype)) deallocate(ftype)
    allocate(ftype(-1:ll))

    if(master)  then
       open(11,file=outpath//outname//'LPR.log',status='replace')
       write(11,'("Log file for the log likelihood evolution across EM steps")')
    end if

    call estim_ini(kk,ll,dim_I,dim_J,hetdim) !produce ep. Minor initialization

		!====== Here, set initial firm classification guess =====
    
    
	!Initial guess of firm types assigned by ordering of average wages
	call inifirmgroup
	
    !Perform rest of initialization
    call initialize(seed)
    call firmorder

    if(master)  then
       write(11,'("Initial guess:")')
		   call writemodelpar(ep,11,.TRUE.,.TRUE.)
    end if
		
    oldep=new_modelpar(ep%k,ep%l)	  
		!============ Done with initialization ===============    

    !=== Estimation is primed =====
		mm_maxit=1
    ccount=1

    call e_step(loglike,.true.)
    call m_step(.false.,.true.) !m_step arg determines if restricted (true) or unrestricted
		call c_step(.true.)
		
		call mpi_barrier(mpi_comm_world,ierr)
    if(master)  then
       write(11,'("After emc_step:")')
		   call writemodelpar(ep,11,.TRUE.,.TRUE.)
		   write(6,'("Initialization and priming done.")')
    end if
		!====== Done with priming =====
		
		call CEM_unrestr(100) 	!call unrestricted mobility model CEM
    if(master) then
       call  writemodelpar(ep,11,.true.,.true.)
    end if

		call CEM_restr(2) 		!call restricted mobility model CEM.
    if(master) then
       call  writemodelpar(ep,11,.true.,.true.)
       close(11)
    end if

    !======== Final outputs ======
    if(master) then !write final estimate
			write(6,'("All done. Now write final outputs.")')
      open(10,file=outpath//outname//'LPR.txt',status='replace')
      call  writemodelpar(ep,10,.false.,.true.)
      close(10)
    end if !end if master

    !========== Done ===============
    if(master) write(6,'("Estimation done. Exiting cem_estim.")')
  contains
  	subroutine CEM_restr(maxc) !=== Perform CEM estimation using LPR restricted mobility model. =====
			integer, intent(in) :: maxc !maximum number of c-steps
    	mm_maxit=5 !number of MM step iterations.
    	ccount=1
    	diffc=1.d0
    	diff=1.d0
    	loglike=-huge(loglike)
    
    	do while((diffc>tolc .or. diff>tol) .and. ccount .le. maxc)
      	!======= Now, for EM iterations ======
       	diff=1.d0
       	itcount=1
       	cem_counter=1
       	do while(diff>tol .and. itcount<min(maxit,10+ccount*2))
        	if(master) write(6,'("Starting it ",i5,"...",$)') itcount
          oldll=loglike
          oldep=ep !oldep is set equal to ep.
          call e_step(loglike,.true.)
          call m_step(.true.,.true.)

          call NaNfix(nanflag) !fix NaN outcomes in the wageparameters.

          diff=modpar_dist(oldep,ep)
					call mpi_bcast(diff,1,mpi_double_precision,0,mpi_comm_world,ierr)
          if(master) then
             write(6,'(I5,2f30.10)') itcount, diff, loglike
             write(11,'(i10,f30.6,"  solution has NaN=",l2,". LPR Likelihood improved=",l2)') itcount,loglike,nanflag,(loglike >= oldll)
          end if
          itcount=itcount+1
          cem_counter=cem_counter+1
       end do ! end of while EM loop
       !========= EM Done ======================
				call mpi_barrier(mpi_comm_world,ierr)
       !========= Re-classify firm types ======================
       oldll=logl()
       call c_step(.true.)
       loglike=logl()
       diffc=2*(abs(loglike-oldll))/abs(loglike+oldll)
			 if(loglike<oldll) diffc=0.d0
			 call mpi_bcast(diffc,1,mpi_double_precision,0,mpi_comm_world,ierr)
       if(master) then
          write( 6,'("cstep_update no=",i4," done. LPR Log-likelihood improved=",l2,". Relative improvement=",f15.8,". Values:",2f15.4)') ccount,(loglike>oldll),diffc,oldll,loglike
          write(11,'("cstep_update no=",i4," done. LPR Log-likelihood improved=",l2,". Relative improvement=",f15.8,". Values:",2f15.4)') ccount,(loglike>oldll),diffc,oldll,loglike
       end if
       ccount = ccount+1       
   		end do  !end of the outer while loop

    	llout_LPR=loglike !here optimized log likelihood is obtained.  	
  	end subroutine cem_restr
  	
  	subroutine cem_unrestr(maxc)
 			integer, intent(in) :: maxc !maximum number of c-steps
 			ccount=1
    	diffc=1.d0
    	diff=1.d0
    	loglike=-huge(loglike)

    	do while((diffc>tolc .or. diff>tol) .and. ccount<maxc)
      	!======= Now, for EM iterations ======
       	diff=1.d0
       	itcount=1
       	cem_counter=1
       	do while(diff>tol .and. itcount<min(maxit,10+ccount*2))
          if(master) write(6,'("Starting it ",i5,"...",$)') itcount
          oldll=loglike
          oldep=ep !oldep is set equal to ep.
          call e_step(loglike,.true.)
          call m_step(.false.,.true.)

          call NaNfix(nanflag) !fix NaN outcomes in the wageparameters.

          diff=modpar_dist(oldep,ep)
					call mpi_bcast(diff,1,mpi_double_precision,0,mpi_comm_world,ierr)
          if(master) then
             write(6,'(I5,2f30.10)') itcount, diff, loglike
             write(11,'(i10,f30.6,"  solution has NaN=",l2,". BLM Likelihood improved=",l2)') itcount,loglike,nanflag,(loglike >= oldll)
          end if
          itcount=itcount+1
          cem_counter=cem_counter+1
       end do ! end of while EM loop
       !========= EM Done ======================
			 call mpi_barrier(mpi_comm_world,ierr)
       !========= Re-classify firm types ======================
       oldll=logl()
       call c_step(.true.)
       loglike=logl()
       diffc=2*(abs(loglike-oldll))/abs(loglike+oldll)
			 if(loglike<oldll) diffc=0.d0
			 call mpi_bcast(diffc,1,mpi_double_precision,0,mpi_comm_world,ierr)
       if(master) then
          write( 6,'("cstep_update no=",i4," done. BLM Log-likelihood improved=",l2,". Relative improvement=",f15.8,". Values:",2f15.4)') ccount,(loglike>oldll),diffc,oldll,loglike
          write(11,'("cstep_update no=",i4," done. BLM Log-likelihood improved=",l2,". Relative improvement=",f15.8,". Values:",2f15.4)') ccount,(loglike>oldll),diffc,oldll,loglike
       end if
       if(loglike<oldll .and. ccount>70) exit
       ccount = ccount+1
    end do  !end of the outer while loop

    llout_BLM=loglike !here optimized log likelihood is obtained.
   end subroutine cem_unrestr
  	
    
   recursive function wagecount(myspell) result(res)
    type(spell), pointer, intent(in) :: myspell
    integer(4) :: res
    res=0
    if(.not. associated(myspell)) then
       return
    else
       if(associated(myspell%wage) .and. firmtype(myspell%fid)>0) then
           res=size(myspell%wage)
       else
           res=0
       end if    
    end if
    if(associated(myspell%next)) res=res+wagecount(myspell%next)
  end function wagecount
   
    subroutine NaNfix(nanflag)
      logical, intent(out) :: nanflag
      real(8), parameter :: gamlb=dble(1.0e-5)
      integer :: ii,jj,tt
      nanflag=.false.
      do ii=1,ep%k
         do jj=1,ep%l
            do tt=1,hetdim
               if(isnan(ep%muw(ii,jj,tt))) then
                  nanflag=.true.
                  ep%muw(ii,jj,tt)=oldep%muw(ii,jj,tt) !revert back to old muw and hope for better times.
                  if(master) write(11,'("Found NaN in muw at",3i5)') ii,jj,tt
               end if
            end do
         end do
      end do
      
      do ii=1,ep%k
         do jj=1,ep%l
            do tt=1,sdhetdim
               if(isnan(ep%sigmaw(ii,jj,tt))) then
                  nanflag=.true.
                  ep%sigmaw(ii,jj,tt)=oldep%sigmaw(ii,jj,tt) !revert back to old sigmaw and hope for better times.
                  if(master) write(11,'("Found NaN in sigmaw at",3i5)') ii,jj,tt
               end if
            end do
         end do
      end do

      do ii=1,ep%k
         do jj=1,ep%l
            do tt=1,mobhetdim
               if(ep%gam(ii,jj,tt)<gamlb) then
                  nanflag=.true.
                  ep%gam(ii,jj,tt)=gamlb !force gamma strictly positive
                  if(master) write(11,'("gam forced at lower bound",3i5)') ii,jj,tt
               end if
            end do
         end do
      end do
    end subroutine NaNfix

    subroutine initialize(seed)
      integer, intent(in) :: seed
      integer :: i,j,tt
      real(8) :: sigvar,muvar,gamvar,lamvar,rhovar
      
      if(seed>0) then
         sigvar=0.1d0
         muvar=0.1d0
         gamvar=0.5d0
         lamvar=0.5d0
         rhovar=0.2d0
      else
         sigvar=0.d0
         muvar=0.d0
         gamvar=0.d0
         lamvar=0.d0
         rhovar=0.d0
      end if
      call setseed(kisseed(2345678+2*seed,87656789+2*seed,5434567+2*seed,43212345+2*seed))

      ! Estimation provided with the dimensions of the problem and an initial guess.
      ! Initial parameters
      do i=1,ep%k
      	do j=1,ep%l
      		ep%lam(i,j,:) =0.0001d0*exp(kissgaus()*lamvar)
      	end do
      end do
      
      do i=1,ep%k
         do j=1,ep%l
            ep%muw(i,j,:)=2.d0*dble(i-ep%k/2)/dble(ep%k)+2.d0*dble(j-ep%l/2)/dble(ep%l)+6.d0
            ep%sigmaw(i,j,:)=1.5d0*exp(kissgaus()*sigvar)+3.d0
         end do
         do j=1,ep%l
            ep%gam(i,j,:)= 1.d0*exp(kissgaus()*gamvar)
         end do
         do tt=1,mobhetdim
            ep%gam(i,:,tt)=ep%gam(i,:,tt)/sum(ep%gam(i,:,tt)) !normalize gam to have gam sum to one.
         end do
      end do

      ep%psi=1.d0/(dble(ep%l)*26.d0)
      ep%del=1.d0/200.d0
      
      ep%muw=ep%muw + kissgaus()*muvar
      
      ep%rho=0.4d0
      ep%rho=ep%rho + kissgaus()*rhovar

      call M_update(ep)

      ep%piw=1.d0/dble(zhetdim)
      ep%m=1.d0/dble(ep%k*(ep%l+1))

			call mpi_bcast(ep%gam,ep%k*ep%l*mobhetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%lam,ep%k*ep%l*mobhetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%muw,ep%k*ep%l*hetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%sigmaw,ep%k*ep%l*sdhetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%rho,1,mpi_double_precision,0,mpi_comm_world,ierr)
			if(master) write(6,'("Everybody has same guess.")')
    end subroutine initialize
  end subroutine cem_estim
  
  subroutine mobinput(myminput) !produce average size, inflow and outlfow per year
    real(8), intent(out) :: myminput(4,dim_J)
    integer, parameter :: yr=52 !number of weeks in a year
    integer :: mylife(2,dim_J), life(2,dim_J), i,j, n_persyr
    real(8) :: myfirmlife(dim_J)
    real(8) :: mysize(80,dim_J),myinflow(80,dim_J),myoutflow(80,dim_J) !maximum number of periods from 1935-2013
    real(8) :: fsize(80,dim_J),inflow(80,dim_J),outflow(80,dim_J) !maximum number of periods from 1935-2013
    type(spell), pointer :: myspell
    
    myminput=0.d0
    mysize=0.d0
    myfirmlife=0.d0
    myinflow=0.d0
    myoutflow=0.d0
    fsize=0.d0
    inflow=0.d0
    outflow=0.d0
    !initialize mylife(1) bithweek to be some large number, and deadthweek to be zero
    mylife(1,:)=10000
    mylife(2,:)=0
    life=0
    
    do j=1,dim_J
      if(.not. associated(firm(j)%first)) cycle
      mylife(:,j)=firmlife(firm(j)) !this gives first and last weeks of firm in this period 
    end do
    
    call mpi_allreduce(mylife(1,:),life(1,:),dim_J,mpi_integer,mpi_min,mpi_comm_world,ierr)
    call mpi_allreduce(mylife(2,:),life(2,:),dim_J,mpi_integer,mpi_max,mpi_comm_world,ierr)
    do j=1,dim_J
        if(.not. associated(firm(j)%first)) cycle
          call avefirm(firm(j)%first,life(1,j),life(2,j))
    end do
    call mpi_allreduce(mysize,fsize,80*dim_J,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(myinflow,inflow,80*dim_J,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(myoutflow,outflow,80*dim_J,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    
    do j=1,dim_J
      myfirmlife(j)=max(real(life(2,j)-firm(j)%bweek)/real(yr),1.d0)
      !average firm size per year
      if(myfirmlife(j)>0.d0) myminput(1,j)=sum(fsize(:,j))/dble(myfirmlife(j))
      if(myminput(1,j) > 0.d0 .and. myfirmlife(j)>0.d0) then
          myminput(2,j)=(sum(inflow(:,j))/dble(myfirmlife(j)))/myminput(1,j) !average inflow per year
          myminput(3,j)=(sum(outflow(:,j))/dble(myfirmlife(j)))/myminput(1,j) !average outflow per year
          myminput(4,j)=myfirmlife(j)
      else
          myminput(2,j)= 0.d0
          myminput(3,j)= 0.d0
          myminput(4,j)= 0.d0
      end if
    end do
   
  contains
  recursive subroutine avefirm(mysp,minwk,maxwk)
    type(spell), pointer, intent(in) :: mysp
    integer, intent(in) :: maxwk, minwk
    integer :: startyr, endyr
    
    if(.not. associated(mysp)) return
    startyr=max(ceiling(real(mysp%strt)/real(yr)),1)
    endyr=max(ceiling(real(mysp%end)/real(yr)),1)
    !firm size by year
    mysize(startyr:endyr,mysp%fid)=mysize(startyr:endyr,mysp%fid)+1.d0
    !inflow by year of entry 
    if(mysp%strt>=minwk) myinflow(startyr,mysp%fid)=myinflow(startyr,mysp%fid)+1.d0
    !outflow by year of exit  
    if(mysp%end<=maxwk) myoutflow(endyr,mysp%fid)=myoutflow(endyr+1,mysp%fid)+1.d0

    call avefirm(mysp%f_next,minwk,maxwk)
  end subroutine avefirm
  end subroutine mobinput
  
  subroutine inifirmgroup !determine initial firm grouping by average firm wage
    implicit none
    integer :: i,j,alive
    real(8) :: avgfwage(dim_J),myfwage(dim_J)
    integer :: nobs(dim_J),mynobs(dim_J)
    integer, allocatable :: order(:)
    real(8), allocatable :: wsort(:)
    type(spell), pointer :: mysp
    avgfwage=0.d0
    nobs=0.d0
    firmtype=0
    do j=1,dim_J !determine average firm wage
       mynobs(j)=0
       myfwage(j)=0.d0
       if(associated(firm(j)%first)) then
          mysp=>firm(j)%first
          do while(associated(mysp))
             if(associated(mysp%wage)) then
                myfwage(j)=myfwage(j)+sum(exp(mysp%wage))
                mynobs(j)=mynobs(j)+size(mysp%wage)
             end if
             mysp=>mysp%f_next
          end do
       end if
    end do
    call mpi_reduce(myfwage,avgfwage,dim_J,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr)
    call mpi_reduce(mynobs,nobs,dim_J,mpi_integer,mpi_sum,0,mpi_comm_world,ierr)
    if(master) then
       do j=1,dim_J
          if(nobs(j)>0) then
             avgfwage(j)=avgfwage(j)/dble(nobs(j))
          else
             firmtype(j)=-1
          end if
       end do
       alive=dim_J-count(firmtype==-1)
			 write(6,'("dim_J=",i10,"Alive=",i10)') dim_J,alive
       if(alive==0) stop'no firms have wage observations.'
       allocate(order(alive),wsort(alive))
       i=0
       do j=1,dim_J
          if(firmtype(j) == 0) then !firmtype is either 0 or -1 at this point.
             i=i+1
             order(i)=j
             wsort(i)=avgfwage(j)
          end if
       end do
       if(i/=alive) stop'fatal(inifirmgroup): inconsistency in ordering.'
       call hpsort(alive,wsort,order) !returns sorted wage and sorted order (i.e. firm ID who has smallest wage,...firm ID who has highest wage)
       do j=1,alive
          firmtype(order(j))=min(ep%l,int(dble(j)*dble(ep%l)/dble(alive))+1)
       end do
    end if

    call mpi_bcast(firmtype,dim_J+1,mpi_integer1,0,mpi_comm_world,ierr)
    firmtype(0)=0 !just to make sure.

		!Update pif
		ep%pif=0.d0
		ep%pif(0,0)=1.d0
		do j=1,dim_J
			if(firmtype(j)>0) ep%pif(firm(j)%s,firmtype(j))=ep%pif(firm(j)%s,firmtype(j))+1.d0
    end do
        
    do j=1,shetdim
    	do i=1,ep%l
				ep%pif(j,i)=max(0.0001d0,ep%pif(j,i))
      end do
    end do
        
		ep%ps=sum(ep%pif(1:shetdim,1:ep%l),dim=2)
		ep%ps=ep%ps/sum(ep%ps)
		do i=1,ep%l
			ep%pif(:,i)=ep%pif(:,i)/sum(ep%pif(:,i))
		end do
  
    !everybody sets q
    ftype=0
    do j=0,dim_J
       ftype(firmtype(j))=ftype(firmtype(j))+1
    end do
		firmcount=sum(ftype(1:ep%l))
    do j=1,ep%l
       ep%q(j)=dble(ftype(j))/dble(firmcount)
    end do

    !master reports on classification.
    if(master) then
       write(11,'("Initial firm grouping")')
       write(11,'(<ep%l+4>i10)') ftype,sum(ftype),firmcount
    end if
  end subroutine inifirmgroup

  subroutine firmorder !this is used in the C-step.
    integer :: j
    integer :: firm_wage_count(dim_J),myc(dim_J),mysize(dim_J), totalnt
		if(allocated(firm_size_order)) deallocate(firm_size_order)
    allocate(firm_size_order(dim_J))
    myc=0
    mysize=0
    do j=1,dim_J
       firm_size_order(j)=j
       call firmcount(firm(j)%first) !this populates myc and mysize
    end do
    call mpi_allreduce(mysize,firmsize,dim_J,mpi_integer,mpi_sum,mpi_comm_world,ierr) !firmsize now populated

    call mpi_reduce(myc,firm_wage_count,dim_J,mpi_integer,mpi_sum,0,mpi_comm_world,ierr)

    if(master) call hpsort(dim_J,firm_wage_count,firm_size_order)
    call mpi_bcast(firm_size_order,dim_J,mpi_integer,0,mpi_comm_world,ierr)
		!check
    do j=1,dim_J
    	if(firmtype(j)==-1 .and. firmsize(j)>0) stop'-1 firmtype has employment.'
    end do
    if(master) then
        totalnt=sum(firm_wage_count)
        open(10,file=outpath//"nt"//'.txt',status='replace')
         write(10,'(i15)') (totalnt)
         write(10,'("Number of size 0 firms",I10)') count(firmsize==0)
         write(10,'("Number of type -1 firms",I10)') count(firmtype==-1)
         write(10,'("Number of firm_size_order <0 firms",I10)') count(firm_size_order<0)
        close(10)
    end if

  contains
    recursive subroutine firmcount(mysp)
      type(spell), pointer, intent(in) :: mysp
      if(.not. associated(mysp)) return
      if(size(mysp%wage)<1) stop'fatal(firmcount): emp spell wiht degenerate wage vector.'
      myc(mysp%fid)=myc(mysp%fid)+size(mysp%wage)
      mysize(mysp%fid)=mysize(mysp%fid)+1
      call firmcount(mysp%f_next)
    end subroutine firmcount
  end subroutine firmorder

  recursive function calc_h(mysp,myk,fullh) result(res)
    !calculate individual, k-conditional log likelihood of mobility pattern
    type(spell), intent(in), pointer :: mysp
    integer, intent(in) :: myk
    logical, intent(in) :: fullh !if false, calculate only the mobility part.
    integer :: i
    real(8) :: res
    res=0.d0
    if(dur(mysp)<=0) stop'duration invalid'
    if(dur(mysp)>0 .and. fullh) then !check for valid data (weed out poorly constructed spells)
       do i=1,size(mysp%wage) !loop over years in spell.
          if(ep%Mbar(myk,firmtype(mysp%fid),mobhetmap(mysp,i))<=0.d0) then
                  if(master) write(6,'("bad mbar at",f2.5,3i5)') ep%Mbar(myk,firmtype(mysp%fid),mobhetmap(mysp,i)),myk,firmtype(mysp%fid),mobhetmap(mysp,i)
                  if(master) write(11,'("bad mbar at",f2.5,3i5)') ep%Mbar(myk,firmtype(mysp%fid),mobhetmap(mysp,i)),myk,firmtype(mysp%fid),mobhetmap(mysp,i)
                  ep%Mbar(myk,firmtype(mysp%fid),mobhetmap(mysp,i))=0.000001d0 !set mbar to some small number
          end if
          res=res+dble(staycount(mysp,i))*log(ep%Mbar(myk,firmtype(mysp%fid),mobhetmap(mysp,i))) !periods of staying times log prob of staying.
       end do
    end if
    i=size(mysp%wage) !Set i to last year (for hetmap)
    if(associated(mysp%next)) then
       if(ep%MM(myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i))<=0.d0) then
           if(master) write(6,'("bad mm at",f2.5,4i5)') ep%MM(myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i)),myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i)
           if(master) write(11,'("bad mm at",f2.5,4i5)') ep%MM(myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i)),myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i)
       end if
       res=res+&
            &log(ep%MM(myk,firmtype(mysp%fid),firmtype(mysp%next%fid),mobhetmap(mysp,i)))&
            &-log(ep%q(firmtype(mysp%next%fid)))-log(dble(firmcount))&
            &+calc_h(mysp%next,myk,fullh)
          if(isnan(res)) then
              if(master) write(6,'("bad calc_h at",f2.5,i5)') ep%q(firmtype(mysp%next%fid)), dble(firmcount)
              if(master) write(11,'("bad calc_h at",f2.5,i5)') ep%q(firmtype(mysp%next%fid)), dble(firmcount)
         end if
    end if
  end function calc_h
  
  recursive function calc_q(mysp,myk) result(res)
    !calculate individual, k-conditional log likelihood of wage observations.
    type(spell), intent(in), pointer :: mysp
    integer, intent(in) :: myk
    real(8) :: res
    res=0.d0
    if (associated(mysp%wage)) res=fdensity(mysp,myk,firmtype(mysp%fid))
    if(associated(mysp%next)) res=res+calc_q(mysp%next,myk)
  end function calc_q

  function staycount(mythis,myi) result(res) !this counts periods that the worker chooses to stay with firm
    !during yr myi (exlcudes last week, where the worker moves).
    type(spell), pointer, intent(in) :: mythis
    integer(4), intent(in) :: myi
    integer(4) :: res,myyr
    if(myi<1 .or. myi>size(mythis%wage)) stop'Fatal(staycount): myi not in support'
    myyr=myi+mythis%yr-1
    res=yrtoweek(myyr+1)-yrtoweek(myyr)
    if(myi==1) res=res-max(mythis%strt-yrtoweek(myyr),0)
    if(myi==size(mythis%wage)) res=res-max(yrtoweek(myyr+1)-mythis%end,0)
    if(res<0) then
       write(6,'(11i12)') mythis%wid,mythis%fid,mythis%yr,mythis%strt,mythis%end,myyr,size(mythis%wage),res,myi,yrtoweek(myyr+1),yrtoweek(myyr)
       stop'Fatal(staycount): spell data must be a single year where start>end.'
    end if
  end function staycount
  
  function logli(i,k) result(res) !k-conditional  individual loglike.
    integer, intent(in) :: i,k
    real(8) :: res
    res=0
    if(.not. associated(work(i)%first)) return
    res=log(ep%pif(firm(work(i)%first%fid)%s,firmtype(work(i)%first%fid)))+log(ep%piw(zhetmap(work(i)),k))+log(ep%m(k,firmtype(work(i)%first%fid),inihetmap(work(i)%first)))&
         &-log(ep%q(firmtype(work(i)%first%fid))*dble(firmcount))&
         &+calc_q(work(i)%first,k)+calc_h(work(i)%first,k,.true.) 
    if(isnan(res)) then
       write(6,'("L is Nan",3i10,2f20.4)') i,k,firmtype(work(i)%first%fid),calc_q(work(i)%first,k),calc_h(work(i)%first,k,.true.)
       stop
    end if
    
    if(res<-huge(res)) then
       write(6,'("L is -ve infty",3i10,2f20.4)') i,k,firmtype(work(i)%first%fid),calc_q(work(i)%first,k),calc_h(work(i)%first,k,.true.)
       stop
    end if
    
    if(res>huge(res)) then
       write(6,'("L is +ve infty",3i10,2f20.4)') i,k,firmtype(work(i)%first%fid),calc_q(work(i)%first,k),calc_h(work(i)%first,k,.true.)
       stop
    end if
  end function logli

  function logl() result(res)
    integer :: i,k
    real(8) :: myres, res
    real(16) :: lik,lik1
    myres=0.d0
    do i=1+myid,dim_I,numprocs
       if(.not. associated(work(i)%first)) cycle
       lik=0.d0
       do k=1,ep%k
       		lik1=logli(i,k)
          lik=lik+exp(lik1)
       end do
       myres=myres+log(lik)
    end do
    call mpi_allreduce(myres,res,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
  end function logl
  
  subroutine residual(seed)
    integer, intent(in) :: seed
    real(8) :: myek(ep%k), mye
    integer :: i, j, tt, k, period, typ, newfile, spellcount
    type(spell), pointer :: mythis
    character(20) :: filename
  
    newfile=1.d0
    do i = 1+myid,dim_I,numprocs
           spellcount=0.d0
           mythis => work(i)%first
           do while(associated(mythis))
              j=mythis%fid 
              spellcount=spellcount+1.d0
              if (j>0) then
              period=size(mythis%wage)
              if (period>1) then
                  do tt=1,period
                   typ=hetmap(mythis,tt)
                   mye=0.d0
                   do k=1,ep%k
                      myek(k)=mythis%wage(tt)-ep%muw(k,firmtype(j),typ)
                      mye=mye+myek(k)*pprobw(i,k)
                   end do !end of k type loop
                   !begin writing results
                   if (newfile==1) then
                       write(filename, '(a,i0,a,i0,a)') "residual",seed,"_",myid,".txt"
                       open(11,file=outpath//filename,status='replace')
                       write(22,'(5i10,f15.8)') i, j, typ, tt, spellcount, mye
                       newfile=0
                   else
                       write(filename, '(a,i0,a,i0,a)') "residual",seed,"_",myid,".txt"
                       open(11,file=outpath//filename,access='append')
                       write(22,'(5i10,f15.8)') i, j, typ, tt, spellcount, mye
                   end if   
                  end do !end of tt loop
              end if !end of condition if we have obs of stayers
              end if !end of condition if this firm is not unemployed spell
              !update mythis
              mythis => mythis%next
           end do !end of while loop
      end do !end of worker i loop    
  end subroutine residual
  
  subroutine e_step(loglike,verb) !this subroutine updates pprobw(:,:), given wage and mobility parameters in ep.
    real(8), intent(out) :: loglike
    logical, optional, intent(in) :: verb
    logical :: verbose
    integer :: i,k,j,tt
    real(8) :: mylike,maxLk,logss,logq,logh
    real(8) :: mym(ep%k,0:ep%l,inihetdim),mypi(zhetdim,ep%k) !for update of ep%m and ep%pi
    real(16) :: L(ep%k)
    type(spell), pointer :: mysp
		if(present(verb)) then
			verbose=verb
		else
			verbose=.false.
		end if
		if(verbose .and. master) write(11,'("e_step routine is called... ",$)')
   	if(master .and. verbose .and. e_settotruth) write(11,'("Worker posterior centered on simulation truth... ",$)')    

		call mpi_barrier(mpi_comm_world,ierr)

    ! Initialize the likelihood and density function:
    pprobw = 0.d0
    mym=0.d0
    mypi=0.d0
    mylike=0.0d0

    !Begin computing worker posterior prob (split over processors).

    do i=1+myid,dim_I,numprocs !each processor holds only its own pprobw... be careful in m step to replicate this worker division.
       if(.not. associated(work(i)%first)) cycle
       do k=1,ep%k
          L(k)=logli(i,k)
       end do
       if(sum(exp(L)) .lt. -huge(mylike)) write(6,'("Warning: individual likelihood huge", 2i10)') myid,i
       mylike=mylike+log(sum(exp(L))) !producing each processor's contribution to total log-likelihood.
       if(sum(exp(L))==0.d0) then !the likelihood calculation failed. Provide output.
          write(6,'("pprobw will fail for worker",i10)') i
          call write_spell(work(i),6,work(i)%first)
          stop
       end if
       do k=1,ep%k        !Individual i's posterior is updated here. Held only by i's processor.
       		pprobw(i,k)=max(0.000001d0,exp(L(k))/sum(exp(L))) !impose a lower bound on the posterior.
       end do
			 pprobw(i,:)=pprobw(i,:)/sum(pprobw(i,:))
			 !Here, if doing simulation check, can override with truth
			 if(e_settotruth) then
			 	pprobw(i,:)=0.d0
			 	pprobw(i,work(i)%k)=1.d0
			 end if
			 
       do j=1,ep%k !check that posterior is well defined.
          if(isnan(pprobw(i,j))) then
             write(6,'("Not a number (pprobw): "i10,<ep%k>f20.4)') i,pprobw(i,:)
             stop
          end if
       end do

       !now, do the summations for the update of ep%m and ep%pi
       mypi(zhetmap(work(i)),:)=mypi(zhetmap(work(i)),:)+pprobw(i,:)
       mym(:,firmtype(work(i)%first%fid),inihetmap(work(i)%first))=mym(:,firmtype(work(i)%first%fid),inihetmap(work(i)%first))+pprobw(i,:)
    end do
    
    
    ! Proportion of workers of each type
    call mpi_allreduce(mypi,ep%piw,ep%k*zhetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(mym,ep%m,ep%k*(ep%l+1)*inihetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)    
    call mpi_allreduce(mylike,loglike,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

		ep%pz=sum(ep%piw,dim=2) !sum over k dimension to get z marginal dist
		ep%pz=ep%pz/sum(ep%pz)

		ep%p=sum(ep%piw,dim=1) !sum over z dimension to get k marginal dist
		ep%p=ep%p/sum(ep%p)

    do i=1,ep%k !make piw sum to one over z dimension for given k
       ep%piw(:,i)=ep%piw(:,i)/sum(ep%piw(:,i))
    end do

    do j=1,inihetdim !make m sum to one over (k,l) dimension for given (x)
       if(sum(sum(ep%m(:,:,j),dim=2))>0.d0) ep%m(:,:,j)=ep%m(:,:,j)/sum(sum(ep%m(:,:,j),dim=2)) !if j cat is empty, don't update.
    end do

		call mpi_bcast(ep%piw,zhetdim*ep%k,mpi_double_precision,0,mpi_comm_world,ierr)
		call mpi_bcast(ep%pz,zhetdim,mpi_double_precision,0,mpi_comm_world,ierr)
		call mpi_bcast(ep%p,ep%k,mpi_double_precision,0,mpi_comm_world,ierr)
		call mpi_bcast(ep%m,ep%k*(1+ep%l)*inihetdim,mpi_double_precision,0,mpi_comm_world,ierr)				
    if(master .and. verbose) write(11,'("Exiting e_step.")')
  end subroutine e_step

  subroutine m_step(LPR,verb) !this subroutine updates wage and mobility parameters in ep given pprobw(:,:).
    logical, intent(in) :: LPR !if true, do MM else do BLM frequency
		logical, intent(in), optional :: verb
		logical :: verbose
    real(8) :: inill,endll,prec
    real(8), dimension(ep%k,0:ep%l,mobhetdim) :: nbar,mynbar
    real(8), dimension(ep%k,0:ep%l,0:ep%l,mobhetdim) :: n,myn
    integer :: iter, myi,myk
    logical :: ok
    if(present(verb)) then
			verbose=verb
		else
			verbose=.false.
		end if
		call mpi_barrier(mpi_comm_world,ierr)
    if(master .and. verbose) write(11,'("Entering M step. ",$)')		
    ! ==== 1. update of wage parameters =====
    call update_wagepar

    ! ==== 2. update of mobility parameters ====
    call update_mobpar(LPR)
    if(master .and. verbose) write(11,'("Done. ")')
  contains
    subroutine update_wagepar
      integer :: i,k,j,period,tt,t,typ,typlag,jlag,index1,index2,l,info, sdtyp
      integer :: xl_alive, index_alive(ep%l*hetdim)
      real(8), dimension(ep%k,ep%l*hetdim,ep%l*hetdim) :: mya, a
      real(8), dimension(ep%k,ep%l*hetdim) :: myb, b
      real(8), dimension(ep%k,ep%l,hetdim)  :: muw, muwdenom
      real(8), dimension(ep%k,ep%l,sdhetdim)  :: denom,mydenom,mysigmaw,sigmaw
      real(8) :: myrho, myrhodenom, rhodenom, lastwage, rho, amat(ep%l*hetdim,ep%l*hetdim), bmat(ep%l*hetdim)
      real(8), allocatable :: a_nmat(:,:),a_n(:,:,:),b_nmat(:),b_n(:,:)
      integer, allocatable :: pivot(:)
      type(spell), pointer :: mythis
     	if(master .and. verbose) write(11,'("Wagepar update... ",$)')
      
      denom=0.0d0
      mydenom=0.0d0
      muwdenom=0.d0
      muw=0.d0
      a=0.d0
      b=0.d0
      mya=0.d0
      myb=0.d0
      !MUW
      do i = 1+myid,dim_I,numprocs
        do k = 1,ep%k
           mythis => work(i)%first  !Reset the AR process across spells.
           do while(associated(mythis))
             j = mythis%fid   !firm ID
             if(j<0 .or. j>dim_J) then
             	write(6,'("(update_wagepar): firm id corruption.",2i15)') myid,i,j
             	stop
             end if
             if (firmtype(j)>0) then
                period = size(mythis%wage)  !number of periods in the spell
                if(period<1) stop'empty wage vector.'
                if(isnan(pprobw(i,k))) write(6,'("pprobw not a number",2i6)') i,k
                  do tt=1,period 
                      if(tt==1) then !for the first wage of the first spell
                         typ=hetmap(mythis,tt)
                         sdtyp=sdhetmap(mythis,tt)
                         index1=firmtype(j)+(typ-1)*ep%l
                         mya(k,index1,index1)=mya(k,index1,index1)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)
                         myb(k,index1)=myb(k,index1)+pprobw(i,k)*mythis%wage(tt)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)
                         mydenom(k,firmtype(j),sdtyp)=mydenom(k,firmtype(j),sdtyp)+pprobw(i,k) !this is for sigma calculation and whether to update muw
                      else 
                         typ=hetmap(mythis,tt)
                         typlag=hetmap(mythis,tt-1) !for tt==1, then typ is based on last typ of previous spell
                         sdtyp=sdhetmap(mythis,tt)
                         jlag=j !if reading subsequent wages within a spell then lag(firmtype)=current firm type
                         lastwage=mythis%wage(tt-1)
                         index1=firmtype(j)+(typ-1)*ep%l
                         index2=firmtype(jlag)+(typlag-1)*ep%l
                         if(index1==index2) then
                            mya(k,index1,index1)=mya(k,index1,index1)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*(1.d0-ep%rho)**2
                            myb(k,index1)=myb(k,index1)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*(1.d0-ep%rho)*(mythis%wage(tt)-ep%rho*lastwage)
                         else
                            mya(k,index1,index1)=mya(k,index1,index1)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)
                            mya(k,index1,index2)=mya(k,index1,index2)-pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*ep%rho
                            mya(k,index2,index1)=mya(k,index2,index1)-pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*ep%rho
                            mya(k,index2,index2)=mya(k,index2,index2)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*ep%rho**2
                            myb(k,index1)=myb(k,index1)+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*(mythis%wage(tt)-ep%rho*lastwage)
                            myb(k,index2)=myb(k,index2)-pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*ep%rho*(mythis%wage(tt)-ep%rho*lastwage)
                         end if !end if index1=index2 
                         mydenom(k,firmtype(j),sdtyp)=mydenom(k,firmtype(j),sdtyp)+pprobw(i,k) !this is for sigma calculation and whether to update muw
                      end if ! end if wage is first in the spell or not
                  end do ! end of tt loop
             end if ! end of if condition that this firm is unemployment or not
             ! update mythis
             mythis => mythis%next
           end do !end of while loop
         end do !end of k loop
       end do !end of i worker loop
       call mpi_barrier(mpi_comm_world,ierr)
       if(master) write(6,'("Done with my and myb calcs")')
        
       do k=1,ep%k
         do i=1,ep%l*hetdim
             if(isnan(myb(k,i)) .or. abs(myb(k,i))>huge(0.d0)) then
             if(master) print*, k,i
             stop'myb bad'
             end if
           do j=1,ep%l*hetdim
             if(isnan(mya(k,i,j)) .or. abs(mya(k,i,j))>huge(0.d0)) then
             if(master) print*, k,i,j
             stop'mya bad'
             end if
           end do
         end do
       end do !end of k loop
       
      call mpi_allreduce(myb,b,ep%k*ep%l*hetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      call mpi_allreduce(mya,a,ep%k*ep%l*hetdim*ep%l*hetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      call mpi_allreduce(mydenom,denom,ep%k*ep%l*sdhetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    !set muwdenom to short and long tenure of denom(k,l,y) where y has 1-4 rows for 0-1yr tenure, 5-8 for 1-2 tenure, and 9-12 for 2+ tenure
    !combine 0-1 yr tenure and 1-2 yr tenure as short tenure
			if(sdhetdim>1) then !in this case estimation is not on simulation
      	muwdenom(:,:,1:4)=denom(:,:,1:4)+denom(:,:,4:8)
      	muwdenom(:,:,5:8)=denom(:,:,9:12)
      else
      	muwdenom=denom
      end if
    
    !Reshape matrixes by removing empty(x,ell) combination. You can tell this from zeros in b or denom.
    if(allocated(a_n)) deallocate(a_n,a_nmat,b_n,b_nmat,pivot)
    xl_alive=0
    index_alive=0
    do i=1,ep%l*hetdim
        if(b(1,i)/=0.d0) then
            xl_alive=xl_alive+1
            index_alive(xl_alive)=i !keep tracp of original index
        end if
    end do
    if(master .and. ep%l*hetdim .ne. xl_alive)  write(6,'("total (x,l)=",i10,"Alive (x,l)=",i10)') ep%l*hetdim, xl_alive
    allocate(a_nmat(xl_alive,xl_alive),a_n(ep%k,xl_alive,xl_alive),b_nmat(xl_alive),b_n(ep%k,xl_alive),pivot(xl_alive))
    !now populate matrixes removing zero cells
    b_n=0.d0
    a_n=0.d0
    b_nmat=0.d0
    a_nmat=0.d0
    do k=1,ep%k
      do i=1,xl_alive
          b_n(k,i)=b(k,index_alive(i))
          do j=1,xl_alive
            a_n(k,i,j)=a(k,index_alive(i),index_alive(j)) 
          end do
        end do
    end do
    !Solve for b matrix
    do k=1,ep%k
        a_nmat=a_n(k,:,:)
        b_nmat=b_n(k,:)
        call dgesv(xl_alive,1,a_nmat,xl_alive,pivot,b_nmat,xl_alive,info)
        if(info /=0) then
         if(master) write(6,'("info: 0=success, -1=bad values, 1=zero values",1i6)') info
        end if
        do i=1,xl_alive
           index1=index_alive(i)
           call amatmap(index1,hetdim,typ,l)
           muw(k,l,typ)=b_nmat(i)
        end do ! end of i loop
    end do ! end of k loop
    
      do k=1,ep%k
            do l=1,ep%l
                do tt=1,hetdim
                    if(muwdenom(k,l,tt)>0.d0) ep%muw(k,l,tt)=muw(k,l,tt) !only update ep%muw if well defined.
                end do
            end do
      end do
      
      !SIGMA
      mysigmaw=0.d0
      do i = 1+myid, dim_I, numprocs
         mythis => work(i)%first
         do while(associated(mythis))
            j = mythis%fid   !firm ID
            if(j>0) then
               period = size(mythis%wage)  !number of periods in the first spell       
                 do k = 1, ep%k
                   do t = 1, period
                      sdtyp=sdhetmap(mythis,t)
                      typ=hetmap(mythis,t)
                      if(t==1) then
                        mysigmaw(k,firmtype(j),sdtyp)=mysigmaw(k,firmtype(j),sdtyp)+pprobw(i,k)*(mythis%wage(t) - ep%muw(k,firmtype(j),typ)) **2 
                      else
                        typlag=hetmap(mythis,t-1)
                        mysigmaw(k,firmtype(j),sdtyp)=mysigmaw(k,firmtype(j),sdtyp)+pprobw(i,k)*(mythis%wage(t) - ep%muw(k,firmtype(j),typ)- ep%rho*(mythis%wage(t-1) - ep%muw(k,firmtype(j),typlag)))**2
                      end if !end if t==1
                    end do ! end of t loop
                 end do ! end of k loop
            end if ! end of if condition that this firm is not unemployment
            ! update mythis
            mythis => mythis%next
         end do ! end of while loop
      end do  ! end of i loop

      call mpi_allreduce(mysigmaw,sigmaw,ep%k*ep%l*sdhetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      
      do k=1,ep%k
         do l=1,ep%l
            do t=1,hetdim
            if(muwdenom(k,l,t)==0.d0) cycle   !!! change denom here to have muw dimension  !!!!
               if (isnan(ep%muw(k,l,t)) .or. ep%muw(k,l,t)<-huge(0.d0) .or. ep%muw(k,l,t)>huge(0.d0) ) then
                  if(master) write(6,'("ep%muw update failed.",3I2,3L2)') k,l,t,isnan(ep%muw(k,l,t)),ep%muw(k,l,t)<-huge(0.d0),ep%muw(k,l,t)>huge(0.d0)
               end if
            end do
         end do
      end do
      
      do k=1,ep%k
         do l=1,ep%l
            do t=1,sdhetdim
            if(denom(k,l,t)==0.d0) cycle
               if (denom(k,l,t)>0.0d0 .and. sqrt(sigmaw(k,l,t))>0.0001d0) ep%sigmaw(k,l,t) = sqrt(sigmaw(k,l,t)/denom(k,l,t))
               if (isnan(ep%sigmaw(k,l,t)) .or. ep%sigmaw(k,l,t)<0.d0 .or. ep%sigmaw(k,l,t)>huge(0.d0) ) then
                  if(master) write(6,'("ep%sigmaw update failed.",3I2,3L2)') k,l,t,isnan(ep%sigmaw(k,l,t)),ep%sigmaw(k,l,t)<0.d0,ep%sigmaw(k,l,t)>huge(0.d0)
               end if
            end do
         end do
      end do
      
      !RHO
      myrho=0.d0
      myrhodenom=0.d0
      do i = 1+myid, dim_I, numprocs
         mythis => work(i)%first
         do while(associated(mythis))
            j = mythis%fid   !firm ID
            if(j>0) then
               period = size(mythis%wage)  !number of periods in the first spell       
               if(period>1) then
                 do k = 1, ep%k
                   do t = 2, period
                      typ=hetmap(mythis,t)
                      typlag=hetmap(mythis,t-1)
                      sdtyp=sdhetmap(mythis,t)
                      myrho=myrho+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*(mythis%wage(t) - ep%muw(k,firmtype(j),typ))*(mythis%wage(t-1) - ep%muw(k,firmtype(j),typlag))
                      myrhodenom=myrhodenom+pprobw(i,k)/max(ep%sigmaw(k,firmtype(j),sdtyp),0.0001d0)*(mythis%wage(t-1) - ep%muw(k,firmtype(j),typlag))**2
                   end do ! end of t loop
                 end do ! end of k loop
                end if  !end if period of first spell>1
            end if ! end of if condition that this firm is not unemployment
            ! update mythis
            mythis => mythis%next
         end do ! end of while loop
      end do  ! end of i loop
      call mpi_allreduce(myrhodenom,rhodenom,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      myrho=myrho/rhodenom !do the division by combined denom before summation, to avoid the possibly very large sum.
      call mpi_allreduce(myrho,rho,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      if (isnan(rho) .or. abs(rho)>huge(0.d0)) then
         if(master) write(6,'("rho update failed.",3L2)') isnan(rho),rho<-huge(0.d0),rho>huge(0.d0)
      else
          ep%rho = rho
      end if

			call mpi_bcast(ep%muw,ep%k*ep%l*hetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%sigmaw,ep%k*ep%l*sdhetdim,mpi_double_precision,0,mpi_comm_world,ierr)
			call mpi_bcast(ep%rho,1,mpi_double_precision,0,mpi_comm_world,ierr)
      if(master .and. verbose) write(11,'("Done. ",$)')
    end subroutine update_wagepar

    subroutine update_mobpar(LPR) !update mobility parameters.
      !first step is to calculate the mobility counts
      logical, intent(in) :: LPR !if true, do MM else do unrestricted
      integer :: ii,kk,ll,lll
     	if(master .and. verbose) then
     		if(LPR) write(11,'("Mobility par update, restricted... ",$)')
				if(.not. LPR) write(11,'("Mobility par update, unrestricted... ",$)')
			end if

      mynbar=0.0d0
      myn=0.0d0
			mobcatcount=0
      do globi=1+myid,dim_I,numprocs
         if(.not. associated(work(globi)%first)) cycle
         call mobcount(work(globi)%first)
      end do
      call mpi_allreduce(mynbar,nbar,ep%k*(ep%l+1)*mobhetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
      call mpi_allreduce(myn,n,ep%k*(ep%l+1)*(ep%l+1)*mobhetdim,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)      
			do ii=1,mobhetdim
				mobcatcount(ii)=mobcatcount(ii)+int(sum(sum(nbar(:,:,ii),dim=1)))+int(sum(sum(sum(n(:,:,:,ii),dim=1),dim=1)))
      end do
      !=== Done - all processes now hold updated stayer and separation counts in nbar and n.
      if(LPR) then
          call mobpar_MM(1)
      else
          call mobpar_BLM !Here, only the M(k,l,x) parameters are updated, not the restricted mobility model of LPR.
      end if
      if(master .and. verbose) write(11,'("Done. ",$)')
    end subroutine update_mobpar

    recursive subroutine mobcount(mysp)
      type(spell), pointer, intent(in) :: mysp
      !We know the type of the worker from mysp%wid
      !The start of mysp is found by the function call work(mysp%wid)%start(mysp).
      !This is equivalent to the function call: start(work(mysp%wid),mysp)
      !The end date of mysp is simply mysp%end
      !Number of staying periods is mysp%end-work(mysp%wid)%start(mysp)
      integer :: mtyp,i

      if(dur(mysp)<0) then
         if(associated(mysp%next)) call mobcount(mysp%next)
      else
         do i=1,size(mysp%wage)
            mtyp=mobhetmap(mysp,i)
            mynbar(:,firmtype(mysp%fid),mtyp)=mynbar(:,firmtype(mysp%fid),mtyp)+pprobw(globi,:)*dble(staycount(mysp,i)) !count of staying periods.
         end do

         if(associated(mysp%next)) then !there exists a transition
            mtyp=mobhetmap(mysp,size(mysp%wage))
            myn(:,firmtype(mysp%fid),firmtype(mysp%next%fid),mtyp)=myn(:,firmtype(mysp%fid),firmtype(mysp%next%fid),mtyp)+pprobw(globi,:)
            call  mobcount(mysp%next)
         end if
      end if
    end subroutine mobcount

    subroutine mobpar_BLM
      integer :: k,l,ll,tt
      real(8) :: num,denom,dist
    
    !update psi parameters
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            denom=nbar(k,0,tt)
            do ll=1,ep%l
               denom=denom+n(k,0,ll,tt)
            end do
            do l=1,ep%l
               if(denom>0.d0) then
               	ep%psi(k,l,tt)=max(0.00001d0,n(k,0,l,tt)/denom)
               else
               	ep%psi(k,l,tt)=0.00001d0
               end if
            end do
         end do
      end do
      
    !update delta 
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
             do l=1,ep%l
                denom=nbar(k,l,tt)
                do ll=0,ep%l
                  denom=denom+n(k,l,ll,tt)
                end do !end of ll loop
                if(denom>0.d0) then
               	  ep%del(k,l,tt)=max(0.00001d0,n(k,l,0,tt)/denom)
                else
               	  ep%del(k,l,tt)=0.00001d0
                end if
             end do !end of l loop
         end do !end of k loop
      end do !end of tt loop
    
      !update capM parameters
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
             do l=0,ep%l
                denom=nbar(k,l,tt)
                do ll=0,ep%l
                  denom=denom+n(k,l,ll,tt)
                end do !end of ll denom loop
                if(denom>0.d0) then
                  do ll=0,ep%l
                    ep%MM(k,l,ll,tt)=max(0.00001d0,n(k,l,ll,tt)/denom)
                  end do !end of ll of ep%MM loop
                else
                  ep%MM(k,l,:,tt)=0.00001d0
                end if
             end do !end of l loop
             ep%MM(k,0,0,tt)=0.0d0  !move from 0 to 0 should be 0 prob.
         end do !end of k loop
      end do !end of tt loop
      
    !update Mbar
    do tt=1,mobhetdim
       do k=1,ep%k
          do l=0,ep%l
            ep%mbar(k,l,tt)=1.d0-sum(ep%MM(k,l,:,tt))
            if(ep%mbar(k,l,tt)<0) then
               write(6,'("mbar is negtive",3i15)') k, l, tt
    		stop
            end if
          end do
       end do
    end do
    end subroutine mobpar_BLM
    
    recursive subroutine mobpar_MM(it)
      integer, intent(in) :: it
      real(8), parameter :: tol=1.0e-8
      integer :: k,l,ll,tt,lll
      real(8) :: num,denom,dist,temp
      real(8), dimension(ep%k,ep%l,mobhetdim) :: mylam,mygam,mydel
      real(8), dimension(ep%l,mobhetdim) :: mygam_temp
      real(8), dimension(ep%k,ep%l,ep%l,mobhetdim) :: tilden,staypr
      if(master .and. verbose) write(11,'(i3,$)') it
      mygam=ep%gam
      mylam=ep%lam
      mydel=ep%del
      
      do tt=1,mobhetdim !first define tilde{n}.
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            do l=1,ep%l
               do ll=1,ep%l
               		if(mygam(k,l,tt)+mygam(k,ll,tt)>0.d0) then
                  	staypr(k,l,ll,tt)=mygam(k,l,tt)/(mygam(k,l,tt)+mygam(k,ll,tt))
                  else
                  	staypr(k,l,ll,tt)=1.d0
                  end if
                  tilden(k,l,ll,tt)=mylam(k,ll,tt)*staypr(k,l,ll,tt)
               end do
               if(ep%Mbar(k,l,tt)>0.d0) then
               		tilden(k,l,:,tt)=tilden(k,l,:,tt)*nbar(k,l,tt)/ep%Mbar(k,l,tt)
               else
               		tilden(k,l,:,tt)=huge(tilden(1,1,1,1))
               end if
            end do
         end do
      end do
            
      !update psi parameters
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            denom=nbar(k,0,tt)
            do ll=1,ep%l
               denom=denom+n(k,0,ll,tt)
            end do
            do l=1,ep%l
               if(denom>0.d0) then
               	ep%psi(k,l,tt)=max(0.00001d0,n(k,0,l,tt)/denom)
               else
               	ep%psi(k,l,tt)=0.00001d0
               end if
            end do
         end do
      end do
      
      !update gam parameters.
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            do l=1,ep%l
               num=0.0d0
               denom=0.0d0
               do ll=1,ep%l
                  num=num+tilden(k,l,ll,tt)+n(k,ll,l,tt)
                  denom=denom+(tilden(k,l,ll,tt)+n(k,l,ll,tt)+tilden(k,ll,l,tt)+n(k,ll,l,tt))/(mygam(k,l,tt)+mygam(k,ll,tt))
               end do
               if(denom<=0.d0) then
                  if(master) write(6,'("Denom<0 happened for gam")') denom
                  mygam_temp(l,tt)=0.0001d0
               else
!                  if(master .and. num/denom<0.0001d0) write(6,'("lower censoring happened")')
                  mygam_temp(l,tt)=max(0.0001d0,num/denom)  !impose a lower bound
               end if
            end do
            do l=1,ep%l
               ep%gam(k,l,tt)=max(0.0001d0,mygam_temp(l,tt)/sum(mygam_temp(:,tt))) !that takes care of gamma update
            end do
         end do
      end do

      !update delta
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            num=0.d0
            denom=0.d0
            do l=1,ep%l
               num=num+nbar(k,l,tt)*(1.0d0-mydel(k,l,tt)-sum(mylam(k,:,tt)))/ep%mbar(k,l,tt)+n(k,l,0,tt)
               denom=denom+nbar(k,l,tt)*(1.0d0-mydel(k,l,tt)-sum(mylam(k,:,tt)))/ep%mbar(k,l,tt)+n(k,l,0,tt)
               do ll=1,ep%l
                  denom=denom+tilden(k,l,ll,tt)+n(k,l,ll,tt)
               end do
            end do
            do l=1,ep%l
                temp=n(k,l,0,tt)/(nbar(k,l,tt)*(1.0d0-mydel(k,l,tt)-sum(mylam(k,:,tt)))/ep%mbar(k,l,tt)+n(k,l,0,tt))
               if(denom>0.d0 .and. abs(temp)<huge(0.d0)) then
               	  ep%del(k,l,tt)=(num/denom)*temp 
               else
                  if(master) write(6,'("Denom<0 happened for del")') denom
                  ep%del(k,l,tt)=0.00001d0
              end if
            end do
         end do
      end do
     
    !update lam
     do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty
         do k=1,ep%k
            do ll=1,ep%l
                num=0.d0
                denom=0.d0
                do l=1,ep%l
                   num=num+tilden(k,l,ll,tt)+n(k,l,ll,tt) 
                   denom=denom+nbar(k,l,tt)*(1.0d0-mydel(k,l,tt)-sum(mylam(k,:,tt)))/ep%mbar(k,l,tt)+n(k,l,0,tt)
                   do lll=1,ep%l
                      denom=denom+tilden(k,l,lll,tt)+n(k,l,lll,tt)
                   end do
                end do
                if(denom>0.d0) then
                    ep%lam(k,ll,tt)=num/denom
                else
                    if(master) write(6,'("Denom<0 happened for lam")') denom
                    ep%lam(k,ll,tt)=0.00001d0
                end if
            end do
         end do
      end do
      
      ep%del=max(0.0001d0,ep%del)
      ep%lam=max(0.0001d0,ep%lam)
      
      !All mobility parameters have been updated. Now, update ep%MM and ep%Mbar
      call M_update(ep)

      !==== ep has been fully updated. Now decide if another iteration should be done =====
      dist=0.d0
      do tt=1,mobhetdim
				 if(mobcatcount(tt)==0) cycle !don't update if category is empty     
         dist=dist+sum(sum((ep%gam(:,:,tt)-mygam(:,:,tt))**2,dim=1))+sum(sum((ep%lam(:,:,tt)-mylam(:,:,tt))**2,dim=1))
      end do
      if(dist>tol .and. it<mm_maxit .and. cem_counter>5) call mobpar_MM(it+1) !Begin to do multiple MM iterations once into 5 or more EM iterations.
    end subroutine mobpar_MM
  end subroutine m_step

  subroutine c_step(verb)
    implicit none
		logical, intent(in), optional :: verb
		logical :: verbose
    integer :: j,ii,perc
    integer(4) :: myftype(-1:ep%l), fcount(ep%l),reassigned
    integer(4) :: nl(ep%l) !this holds the spell by firm type counts.
		if(present(verb)) then
			verbose=verb
		else
			verbose=.false.
		end if
		reassigned=0
    if(master .and. verbose) write(11,'("C step. Reassigning firms: ",$)')

    !count firms in each type. Do not reassign type if there is only one member in the group.
    fcount=0
    do j=1,dim_J
       if(firmtype(j)>0) then
          fcount(firmtype(j))=fcount(firmtype(j))+1
       end if
    end do
    
    !first set nl
    nl=0
    do j=1,dim_J
 			if(firmtype(j)==-1 .or. firmsize(j)<=0) cycle
			if(firmtype(j)<1 .or. firmtype(j)>ep%l) stop'firmtype assignment corruption'
      nl(firmtype(j))=nl(firmtype(j))+firmsize(j)
    end do
    !then reassign firms
    !firm_size_order contains firm's ID in ascending order of firm's number of spells
		perc=1
    do j=dim_J,1,-1 !step through reassigning biggest firms first
    	if(firm_size_order(j)==0) stop'firmsize will be evaluated for fid=0'
     	if(firmtype(firm_size_order(j))==-1 .or. firmsize(firm_size_order(j))<=0) then
     		! write(6,'("C step cycled on firm ", i10)') j
     		cycle
     	end if
      call reassign_firm(firm_size_order(j)) !reassigned counts up if change in firmtype
			if(master .and. verbose) then
				if(j .le. dim_j*(10-perc)/10) then
					perc=perc+1
					write(11,'(".",$)') !do a dot for each decile in firm population
				end if
			end if
    end do
    if(master .and. verbose) write(11,'("! ",$)')
    
    !---Check integrity of firmtype classification stats ------
    myftype=0
    do j=0,dim_J
    	myftype(firmtype(j))=myftype(firmtype(j))+1
    end do
    do ii=-1,ep%l
    	if(myftype(ii) .ne. ftype(ii)) then
    		write(6,'("ftype is compromised",3i15)') ii,myftype(ii),ftype(ii)
    		stop
    	end if
    end do
    do j=1,dim_J
    	if(firmtype(j)==-1 .and. firmsize(j)>0) stop'firmtype -1 has positive size'
    end do
    if(firmcount .ne. sum(ftype(1:ep%l))) stop'firmcount out of date'
    
    !update pi's
    ep%pif=0.d0
    ep%pif(0,0)=1.d0
		do j=1,dim_J
			if(firmtype(j)>0) ep%pif(firm(j)%s,firmtype(j))=ep%pif(firm(j)%s,firmtype(j))+1.d0
    end do
        
    do j=1,shetdim
    	do ii=1,ep%l
				ep%pif(j,ii)=max(0.00001d0,ep%pif(j,ii))
     	end do
    end do
        
		ep%ps=sum(ep%pif(1:shetdim,1:ep%l),dim=2)
		ep%ps=ep%ps/sum(ep%ps)
		do ii=1,ep%l
			ep%pif(:,ii)=ep%pif(:,ii)/sum(ep%pif(:,ii))
		end do
		if(master .and. verbose) then
       write(11,'(" Done. Reassigned number of firms: ",I10)') reassigned
       write(11,'(<ep%l+3>i10)') ftype,sum(ftype)
    end if
  contains    
    subroutine reassign_firm(fid)
      implicit none
      integer, intent(in) :: fid
      integer :: myi,oldfirmtype
      real(8) :: myll(ep%l),ll(ep%l)
      oldfirmtype=firmtype(fid)
			if(firmtype(fid)<=0) stop'reassign firm has been called for a firmtype -1 or 0'
			if(fid<=0) stop'reassign firm has been called with fid<=0'
      myll=0.d0
      call ini_contr(fid,myll)
      call wage_contr(firm(fid)%first,myll)
      call mob_contr(firm(fid)%first,myll)

      call mpi_reduce(myll,ll,ep%l,mpi_double_precision,mpi_sum,0,mpi_comm_world,ierr) !only master knows ll
			do myi=1,ep%l
				if(isnan(ll(myi)) .and. master) write(6,'("NaN likelihood:",2I10)') fid,myi
			end do
      if(master) then !Master determines firm fid reclassification.
         ll=ll-ll(oldfirmtype)+omega(fid)
         firmtype(fid) = maxloc(ll,DIM = 1)
      end if
      call mpi_bcast(firmtype(fid),1,mpi_integer1,0,mpi_comm_world,ierr) !master broadcasts firmtype to all      
      if(firmtype(fid) .ne. oldfirmtype) then !update ftype, ep%q, and nl
      	 reassigned=reassigned+1 !count the number of reassigned firms.
         ftype(oldfirmtype)   = ftype(oldfirmtype)-1
         ftype(firmtype(fid)) = ftype(firmtype(fid))+1
         ep%q(1:ep%l)=dble(ftype(1:ep%l))/dble(sum(ftype(1:ep%l)))
         nl(oldfirmtype)=nl(oldfirmtype)-firmsize(fid)
         nl(firmtype(fid))=nl(firmtype(fid))+firmsize(fid)
      end if
!      if(master) write(6,'("!")')
    end subroutine reassign_firm

    function omega(myj) result(res) !Omega contribution subject to firm j changing type.
      integer, intent(in) :: myj
      real(8) :: res(ep%l)
      real(8) :: myq(ep%l)
      integer :: l,ll,mynl(ep%l)
			if(myj<=0) stop'omega called with myj<=0'
      res=0.d0
      do l=1,ep%l
         myq=dble(ftype(1:ep%l))/dble(sum(ftype(1:ep%l)))
         mynl=nl
         if(l .ne. firmtype(myj)) then
            myq(firmtype(myj))=dble(ftype(firmtype(myj))-1)/dble(sum(ftype(1:ep%l)))
            myq(l)            =dble(ftype(l)+1)/dble(sum(ftype(1:ep%l)))
            mynl(firmtype(myj))=nl(firmtype(myj))-firmsize(myj)
            mynl(l)            =nl(l)+firmsize(myj)
         end if

         do ll=1,ep%l
            res(l)=res(l)-log(myq(ll))*mynl(ll)
         end do
      end do
      res=res-res(firmtype(myj))
    end function omega
    
    subroutine ini_contr(fid,myll)
      implicit none
			integer, intent(in) :: fid
      real(8), intent(inout) :: myll(ep%l)
      type(spell), pointer :: mysp
      integer :: l,k !counters

      if(fid==0) return !only contribute if spell is an employment spell
			mysp=>firm(fid)%first
      do while (associated(mysp))
      	if(.not. associated(mysp%prev)) then
      		do l=1,ep%l
         		do k=1,ep%k
            myll(l)=myll(l)+pprobw(mysp%wid,k)*(log(ep%m(k,l,inihetmap(mysp))) &
                   +log(ep%piw(zhetmap(work(mysp%wid)),k)) &
                   +log(ep%pif(firm(work(mysp%wid)%first%fid)%s,l)))    
         		end do
      		end do
      	end if
      	mysp=>mysp%f_next
      end do
      return
    end subroutine ini_contr

    recursive subroutine wage_contr(mysp,myll)
      implicit none
      type(spell), pointer, intent(in) :: mysp
      real(8), intent(inout) :: myll(ep%l)
      integer :: l,k !counters
      if(.not. associated(mysp)) return
      if(mysp%fid>0) then
         do l=1,ep%l
            do k=1,ep%k
               myll(l)=myll(l)+pprobw(mysp%wid,k)*fdensity(mysp,k,l)
            end do
         end do
      end if
      call wage_contr(mysp%f_next,myll)
    end subroutine wage_contr

    recursive subroutine mob_contr(mysp,myll)
      implicit none
      type(spell), pointer, intent(in) :: mysp
      real(8), intent(inout) :: myll(ep%l)
      integer :: l,k,t,xtyp,sendtyp,recvtyp !counters
      real(8) :: stay
      if(.not. associated(mysp)) return
      if(mysp%fid==0) return

      !The stayer contribution
      do t=1,size(mysp%wage)
         stay=staycount(mysp,t) !number of periods staying in current firm.
         xtyp=mobhetmap(mysp,t) !worker's x type
         do l=1,ep%l
            do k=1,ep%k
               myll(l)=myll(l)+pprobw(mysp%wid,k)*stay*log(ep%Mbar(k,l,xtyp))
            end do
         end do
      end do

      !then for the moves
      xtyp=mobhetmap(mysp,size(mysp%wage))
      if(associated(mysp%next)) then !contribution for firm as sender.
         recvtyp=firmtype(mysp%next%fid)
         do l=1,ep%l
            do k=1,ep%k
               myll(l)=myll(l)+pprobw(mysp%wid,k)*log(ep%MM(k,l,recvtyp,xtyp))
            end do
         end do
      end if

      if(associated(mysp%prev)) then !contribution for firm as receiver.
         sendtyp=firmtype(mysp%prev%fid)
         do l=1,ep%l
            do k=1,ep%k
               myll(l)=myll(l)+pprobw(mysp%wid,k)*log(ep%MM(k,sendtyp,l,xtyp))
            end do
         end do
      end if

      !Do next worker in firm
      call mob_contr(mysp%f_next,myll)
    end subroutine mob_contr
  end subroutine c_step
end module cem
