module sparskit

private

public amux, atmux, amub, amubdg, csrcsc, cg

contains



  subroutine amux (n, x, y, a,ja,ia)

    real(8)  x(*), y(*), a(*) 

    integer n, ja(*), ia(*)

    !-----------------------------------------------------------------------

    !         A times a vector

    !----------------------------------------------------------------------- 

    ! multiplies a matrix by a vector using the dot product form

    ! Matrix A is stored in compressed sparse row storage.

    !

    ! on entry:

    !----------

    ! n     = row dimension of A

    ! x     = real array of length equal to the column dimension of

    !         the A matrix.

    ! a, ja,

    !    ia = input matrix in compressed sparse row format.

    !

    ! on return:

    !-----------

    ! y     = real array of length n, containing the product y=Ax

    !

    !-----------------------------------------------------------------------

    ! local variables

    !

    real(8) t

    integer i, k

    !-----------------------------------------------------------------------

    do i = 1,n !compute the inner product of row i with vector x

       t = 0.0d0

       do k=ia(i), ia(i+1)-1 

          t = t + a(k)*x(ja(k))

       end do !store result in y(i) 

       y(i) = t

    end do

  end subroutine amux

  !-------------end-of-amux-----------------------------------------------

  !-----------------------------------------------------------------------



  subroutine atmux (n, x, y, a, ja, ia)

    real(8) x(*), y(*), a(*) 

    integer n, ia(*), ja(*)

    !-----------------------------------------------------------------------

    !         transp( A ) times a vector

    !----------------------------------------------------------------------- 

    ! multiplies the transpose of a matrix by a vector when the original

    ! matrix is stored in compressed sparse row storage. Can also be

    ! viewed as the product of a matrix by a vector when the original

    ! matrix is stored in the compressed sparse column format.

    !-----------------------------------------------------------------------

    !

    ! on entry:

    !----------

    ! n     = row dimension of A

    ! x     = real array of length equal to the column dimension of

    !         the A matrix.

    ! a, ja,

    !    ia = input matrix in compressed sparse row format.

    !

    ! on return:

    !-----------

    ! y     = real array of length n, containing the product y=transp(A)*x

    !

    !-----------------------------------------------------------------------

    !     local variables 

    !

    integer i, k 

    !-----------------------------------------------------------------------

    !

    !     zero out output vector

    ! 

    do i=1,n

       y(i) = 0.0

    end do

    !

    ! loop over the rows

    !

    do i = 1,n

       do k=ia(i), ia(i+1)-1 

          y(ja(k)) = y(ja(k)) + x(i)*a(k)

       end do

    end do



    !-------------end-of-atmux---------------------------------------------- 

    !-----------------------------------------------------------------------

  end subroutine atmux



  subroutine amub (nrow,ncol,job,a,ja,ia,b,jb,ib,c,jc,ic,nzmax,iw,ierr) 

    real(8) a(*), b(*), c(*) 

    integer ja(*),jb(*),jc(*),ia(nrow+1),ib(*),ic(*),iw(ncol)

    !-----------------------------------------------------------------------

    ! performs the matrix by matrix product C = A B 

    !-----------------------------------------------------------------------

    ! on entry:

    ! ---------

    ! nrow  = integer. The row dimension of A = row dimension of C

    ! ncol  = integer. The column dimension of B = column dimension of C

    ! job   = integer. Job indicator. When job = 0, only the structure

    !                  (i.e. the arrays jc, ic) is computed and the

    !                  real values are ignored.

    !

    ! a,

    ! ja,

    ! ia   = Matrix A in compressed sparse row format.

    ! 

    ! b, 

    ! jb, 

    ! ib    =  Matrix B in compressed sparse row format.

    !

    ! nzmax = integer. The  length of the arrays c and jc.

    !         amub will stop if the result matrix C  has a number 

    !         of elements that exceeds exceeds nzmax. See ierr.

    ! 

    ! on return:

    !----------

    ! c, 

    ! jc, 

    ! ic    = resulting matrix C in compressed sparse row sparse format.

    !           

    ! ierr  = integer. serving as error message. 

    !         ierr = 0 means normal return,

    !         ierr .gt. 0 means that amub stopped while computing the

    !         i-th row  of C with i=ierr, because the number 

    !         of elements in C exceeds nzmax.

    !

    ! work arrays:

    !------------

    ! iw    = integer work array of length equal to the number of

    !         columns in A.

    ! Note: 

    !-------

    !   The row dimension of B is not needed. However there is no checking 

    !   on the condition that ncol(A) = nrow(B). 

    !

    !----------------------------------------------------------------------- 

    real(8) scal 

    logical values

    values = (job .ne. 0) 

    len = 0

    ic(1) = 1 

    ierr = 0

    ! initialize array iw.

    do j=1, ncol

       iw(j) = 0

    end do

    do ii=1, nrow 

       ! row i 

       do ka=ia(ii), ia(ii+1)-1 

          if (values) scal = a(ka)

          jj   = ja(ka)

          do kb=ib(jj),ib(jj+1)-1

             jcol = jb(kb)

             jpos = iw(jcol)

             if (jpos .eq. 0) then

                len = len+1

                if (len .gt. nzmax) then

                   ierr = ii

                   return

                endif

                jc(len) = jcol

                iw(jcol)= len

                if (values) c(len)  = scal*b(kb)

             else

                if (values) c(jpos) = c(jpos) + scal*b(kb)

             endif

          end do

       end do

       do k=ic(ii), len

          iw(jc(k)) = 0

       end do

       ic(ii+1) = len+1

    end do

    return

  end subroutine amub

  !-------------end-of-amub-----------------------------------------------

  !-----------------------------------------------------------------------



  subroutine amubdg (nrow,ncol,ncolb,ja,ia,jb,ib,ndegr,nnz,iw) 

    integer ja(*),jb(*),ia(nrow+1),ib(ncol+1),ndegr(nrow),iw(ncolb) 

    !-----------------------------------------------------------------------

    ! gets the number of nonzero elements in each row of A*B and the total 

    ! number of nonzero elements in A*B. 

    !-----------------------------------------------------------------------

    ! on entry:

    ! -------- 

    !

    ! nrow  = integer.  row dimension of matrix A

    ! ncol  = integer.  column dimension of matrix A = row dimension of 

    !                   matrix B.

    ! ncolb = integer. the colum dimension of the matrix B.

    !

    ! ja, ia= row structure of input matrix A: ja = column indices of

    !         the nonzero elements of A stored by rows.

    !         ia = pointer to beginning of each row  in ja.

    ! 

    ! jb, ib= row structure of input matrix B: jb = column indices of

    !         the nonzero elements of A stored by rows.

    !         ib = pointer to beginning of each row  in jb.

    ! 

    ! on return:

    ! ---------

    ! ndegr	= integer array of length nrow containing the degrees (i.e., 

    !         the number of nonzeros in  each row of the matrix A * B 

    !				

    ! nnz   = total number of nonzero elements found in A * B

    !

    ! work arrays:

    !-------------

    ! iw	= integer work array of length ncolb. 

    !-----------------------------------------------------------------------

    do k=1, ncolb 

       iw(k) = 0 

    end do



    do k=1, nrow

       ndegr(k) = 0 

    end do

    !     

    !     method used: Transp(A) * A = sum [over i=1, nrow]  a(i)^T a(i)

    !     where a(i) = i-th row of  A. We must be careful not to add  the

    !     elements already accounted for.

    !     

    do ii=1,nrow

       ! for each row of A

       ldg = 0 

       ! end-of-linked list

       last = -1 

       do j = ia(ii),ia(ii+1)-1

          ! row number to be added:

          jr = ja(j) 

          do k=ib(jr),ib(jr+1)-1

             jc = jb(k) 

             if (iw(jc) .eq. 0) then 

                ! add one element to the linked list 

                ldg = ldg + 1

                iw(jc) = last 

                last = jc

             endif

          end do

       end do

       ndegr(ii) = ldg

       ! reset iw to zero

       do k=1,ldg 

          j = iw(last) 

          iw(last) = 0

          last = j

       end do

       !-----------------------------------------------------------------------

    end do

    nnz = 0

    do ii=1, nrow 

       nnz = nnz+ndegr(ii) 

    end do

    return

  end subroutine amubdg

  !---------------end-of-amubdg ------------------------------------------

  !-----------------------------------------------------------------------



  subroutine csrcsc (n,job,ipos,a,ja,ia,ao,jao,iao)

    integer n,job,ipos,ia(n+1),iao(n+1),ja(*),jao(*)

    real*8  a(*),ao(*)

    !-----------------------------------------------------------------------

    ! Compressed Sparse Row     to      Compressed Sparse Column

    !

    ! (transposition operation)   Not in place. 

    !----------------------------------------------------------------------- 

    ! -- not in place --

    ! this subroutine transposes a matrix stored in a, ja, ia format.

    ! ---------------

    ! on entry:

    !----------

    ! n	= dimension of A.

    ! job	= integer to indicate whether to fill the values (job.eq.1) of the

    !         matrix ao or only the pattern., i.e.,ia, and ja (job .ne.1)

    !

    ! ipos  = starting position in ao, jao of the transposed matrix.

    !         the iao array takes this into account (thus iao(1) is set to ipos.)

    !         Note: this may be useful if one needs to append the data structure

    !         of the transpose to that of A. In this case use for example

    !                call csrcsc (n,1,ia(n+1),a,ja,ia,a,ja,ia(n+2)) 

    !	  for any other normal usage, enter ipos=1.

    ! a	= real array of length nnz (nnz=number of nonzero elements in input 

    !         matrix) containing the nonzero elements.

    ! ja	= integer array of length nnz containing the column positions

    ! 	  of the corresponding elements in a.

    ! ia	= integer of size n+1. ia(k) contains the position in a, ja of

    !	  the beginning of the k-th row.

    !

    ! on return:

    ! ---------- 

    ! output arguments:

    ! ao	= real array of size nzz containing the "a" part of the transpose

    ! jao	= integer array of size nnz containing the column indices.

    ! iao	= integer array of size n+1 containing the "ia" index array of

    !	  the transpose. 

    !

    !----------------------------------------------------------------------- 

    call csrcsc2 (n,n,job,ipos,a,ja,ia,ao,jao,iao) !This routine just calls csrcsc2

  end subroutine csrcsc



  !-----------------------------------------------------------------------



  subroutine csrcsc2 (n,n2,job,ipos,a,ja,ia,ao,jao,iao)

    integer ia(n+1),iao(n2+1),ja(*),jao(*)

    real*8  a(*),ao(*)

    !-----------------------------------------------------------------------

    ! Compressed Sparse Row     to      Compressed Sparse Column

    !

    ! (transposition operation)   Not in place. 

    !----------------------------------------------------------------------- 

    ! Rectangular version.  n is number of rows of CSR matrix,

    !                       n2 (input) is number of columns of CSC matrix.

    !----------------------------------------------------------------------- 

    ! -- not in place --

    ! this subroutine transposes a matrix stored in a, ja, ia format.

    ! ---------------

    ! on entry:

    !----------

    ! n	= number of rows of CSR matrix.

    ! n2    = number of columns of CSC matrix.

    ! job	= integer to indicate whether to fill the values (job.eq.1) of the

    !         matrix ao or only the pattern., i.e.,ia, and ja (job .ne.1)

    !

    ! ipos  = starting position in ao, jao of the transposed matrix.

    !         the iao array takes this into account (thus iao(1) is set to ipos.)

    !         Note: this may be useful if one needs to append the data structure

    !         of the transpose to that of A. In this case use for example

    !                call csrcsc2 (n,n,1,ia(n+1),a,ja,ia,a,ja,ia(n+2)) 

    !	  for any other normal usage, enter ipos=1.

    ! a	= real array of length nnz (nnz=number of nonzero elements in input 

    !         matrix) containing the nonzero elements.

    ! ja	= integer array of length nnz containing the column positions

    ! 	  of the corresponding elements in a.

    ! ia	= integer of size n+1. ia(k) contains the position in a, ja of

    !	  the beginning of the k-th row.

    !

    ! on return:

    ! ---------- 

    ! output arguments:

    ! ao	= real array of size nzz containing the "a" part of the transpose

    ! jao	= integer array of size nnz containing the column indices.

    ! iao	= integer array of size n+1 containing the "ia" index array of

    !	  the transpose. 

    !

    !----------------------------------------------------------------------- 

    !----------------- compute lengths of rows of transp(A) ----------------

    do i=1,n2+1

       iao(i) = 0

    end do

    do i=1, n

       do k=ia(i), ia(i+1)-1 

          j = ja(k)+1

          iao(j) = iao(j)+1

       end do

    end do

    !---------- compute pointers from lengths ------------------------------

    iao(1) = ipos 

    do i=1,n2

       iao(i+1) = iao(i) + iao(i+1)

    end do

    !--------------- now do the actual copying ----------------------------- 

    do i=1,n

       do k=ia(i),ia(i+1)-1 

          j = ja(k) 

          next = iao(j)

          if (job .eq. 1)  ao(next) = a(k)

          jao(next) = i

          iao(j) = next+1

       end do

    end do

    !-------------------------- reshift iao and leave ---------------------- 

    do i=n2,1,-1

       iao(i+1) = iao(i)

    end do

    iao(1) = ipos

  end subroutine csrcsc2

  !--------------- end of csrcsc2 ---------------------------------------- 

  !-----------------------------------------------------------------------



  subroutine cg(n, rhs, sol, ipar, fpar, w)

    implicit none

    integer n, ipar(16)

    real(8) rhs(n), sol(n), fpar(16), w(n,*)

    !-----------------------------------------------------------------------

    !     This is an implementation of the Conjugate Gradient (CG) method

    !     for solving linear system.

    !

    !     NOTE: This is not the PCG algorithm. It is a regular CG algorithm.

    !     To be consistent with the other solvers, the preconditioners are

    !     applied by performing Ml^{-1} A Mr^{-1} P in place of A P in the

    !     CG algorithm.  PCG uses the preconditioner differently.

    !

    !     fpar(7) is used here internally to store <r, r>.

    !     w(:,1) -- residual vector

    !     w(:,2) -- P, the conjugate direction

    !     w(:,3) -- A P, matrix multiply the conjugate direction

    !     w(:,4) -- temporary storage for results of preconditioning

    !     w(:,5) -- change in the solution (sol) is stored here until

    !               termination of this solver

    !-----------------------------------------------------------------------

    !external functions used: distdot, stopbis, brkdn, bisinit

    ! local variables

    integer i

    real*8 alpha

    logical lp,rp

    save



    ! check the status of the call

    if (ipar(1).le.0) ipar(10) = 0

    goto (10, 20, 40, 50, 60, 70, 80), ipar(10)





    ! initialization

    call bisinit(ipar,fpar,5*n,1,lp,rp,w)

    if (ipar(1).lt.0) return



    ! request for matrix vector multiplication A*x in the initialization

    ipar(1) = 1

    ipar(8) = n+1

    ipar(9) = ipar(8) + n

    ipar(10) = 1

    do i = 1, n

       w(i,2) = sol(i)

    enddo

    return

10  ipar(7) = ipar(7) + 1

    ipar(13) = 1

    do i = 1, n

       w(i,2) = rhs(i) - w(i,3)

    enddo

    fpar(11) = fpar(11) + n



    ! if left preconditioned

    if (lp) then

       ipar(1) = 3

       ipar(9) = 1

       ipar(10) = 2

       return

    end if



20  if (lp) then

       do i = 1, n

          w(i,2) = w(i,1)

       end do

    else

       do i = 1, n

          w(i,1) = w(i,2)

       end do

    end if



    fpar(7) = distdot(n,w,w)

    fpar(11) = fpar(11) + 2 * n

    fpar(3) = sqrt(fpar(7))

    fpar(5) = fpar(3)

    if (abs(ipar(3)).eq.2) then

       fpar(4) = fpar(1) * sqrt(distdot(n,rhs,rhs)) + fpar(2)

       fpar(11) = fpar(11) + 2 * n

    else if (ipar(3).ne.999) then

       fpar(4) = fpar(1) * fpar(3) + fpar(2)

    end if



    ! before iteration can continue, we need to compute A * p, which

    ! includes the preconditioning operations

30  if (rp) then

       ipar(1) = 5

       ipar(8) = n + 1

       if (lp) then

          ipar(9) = ipar(8) + n

       else

          ipar(9) = 3*n + 1

       end if

       ipar(10) = 3

       return

    end if



40  ipar(1) = 1

    if (rp) then

       ipar(8) = ipar(9)

    else

       ipar(8) = n + 1

    end if

    if (lp) then

       ipar(9) = 3*n+1

    else

       ipar(9) = n+n+1

    end if

    ipar(10) = 4

    return



50  if (lp) then

       ipar(1) = 3

       ipar(8) = ipar(9)

       ipar(9) = n+n+1

       ipar(10) = 5

       return

    end if



    !continuing with the iterations

60  ipar(7) = ipar(7) + 1

    alpha = distdot(n,w(1,2),w(1,3))

    fpar(11) = fpar(11) + 2*n

    if (brkdn(alpha,ipar)) goto 900

    alpha = fpar(7) / alpha

    do i = 1, n

       w(i,5) = w(i,5) + alpha * w(i,2)

       w(i,1) = w(i,1) - alpha * w(i,3)

    end do

    fpar(11) = fpar(11) + 4*n



    ! are we ready to terminate ?

    if (ipar(3).eq.999) then

       ipar(1) = 10

       ipar(8) = 4*n + 1

       ipar(9) = 3*n + 1

       ipar(10) = 6

       return

    end if



70  if (ipar(3).eq.999) then

       if (ipar(11).eq.1) goto 900

    else if (stopbis(n,ipar,1,fpar,w,w(1,2),alpha)) then

       goto 900

    end if



    ! continue the iterations

    alpha = fpar(5)*fpar(5) / fpar(7)

    fpar(7) = fpar(5)*fpar(5)

    do i = 1, n

       w(i,2) = w(i,1) + alpha * w(i,2)

    end do

    fpar(11) = fpar(11) + 2*n

    goto 30



    ! clean up -- necessary to accommodate the right-preconditioning

900 if (rp) then

       if (ipar(1).lt.0) ipar(12) = ipar(1)

       ipar(1) = 5

       ipar(8) = 4*n + 1

       ipar(9) = ipar(8) - n

       ipar(10) = 7

       return

    end if

80  if (rp) then

       call tidycg(n,ipar,fpar,sol,w(1,4))

    else

       call tidycg(n,ipar,fpar,sol,w(1,5))

    end if

    return

  end subroutine cg



  function distdot(n,dx,dy) !! DistDOT forms the dot product of two vectors

    !    Input, integer N, the number of entries in the vectors.

    !

    !    Input, real ( kind = 8 ) DX(*), the first vector.

    !

    !    Input, real ( kind = 8 ) DY(*), the second vector.

    !

    !    Output, real ( kind = 8 ) DDOT, the sum of the product of the 

    !    corresponding entries of DX and DY.

    !

    implicit none



    integer, intent(in) :: n

    real(8), intent(in) :: dx(n), dy(n)

    real(8) distdot

    integer i,m



    distdot = 0.0D+00



    if ( n <= 0 ) then

       return

    end if



    m = mod ( n, 5 )



    do i = 1, m

       distdot = distdot + dx(i) * dy(i)

    end do



    do i = m+1, n, 5



       distdot = distdot + dx(i  ) * dy(i  ) &

            &+ dx(i+1) * dy(i+1) &

            &+ dx(i+2) * dy(i+2) &

            &+ dx(i+3) * dy(i+3) &

            &+ dx(i+4) * dy(i+4)

    end do

  end function distdot



  subroutine tidycg(n,ipar,fpar,sol,delx)

    implicit none

    integer i,n,ipar(16)

    real*8 fpar(16),sol(n),delx(n)

    !-----------------------------------------------------------------------

    !     Some common operations required before terminating the CG routines

    !-----------------------------------------------------------------------

    real*8 zero

    parameter(zero=0.0D0)



    if (ipar(12).ne.0) then

       ipar(1) = ipar(12) 

    else if (ipar(1).gt.0) then

       if ((ipar(3).eq.999 .and. ipar(11).eq.1) .or. fpar(6) .le. fpar(4)) then

          ipar(1) = 0

       else if (ipar(7).ge.ipar(6) .and. ipar(6).gt.0) then

          ipar(1) = -1

       else

          ipar(1) = -10

       endif

    endif

    if (fpar(3).gt.zero .and. fpar(6).gt.zero .and. ipar(7).gt.ipar(13)) then

       fpar(7) = log10(fpar(3) / fpar(6)) / dble(ipar(7)-ipar(13))

    else

       fpar(7) = zero

    endif

    do i = 1, n

       sol(i) = sol(i) + delx(i)

    enddo

    return

  end subroutine tidycg

  !-----end-of-tidycg



  logical function brkdn(alpha, ipar)

    implicit none

    integer ipar(16)

    real*8 alpha, beta, zero, one

    parameter (zero=0.0D0, one=1.0D0)

    !-----------------------------------------------------------------------

    !     test whether alpha is zero or an abnormal number, if yes,

    !     this routine will return .true.

    !

    !     If alpha == 0, ipar(1) = -3,

    !     if alpha is an abnormal number, ipar(1) = -9.

    !-----------------------------------------------------------------------

    brkdn = .false.

    if (alpha.gt.zero) then

       beta = one / alpha

       if (.not. beta.gt.zero) then

          brkdn = .true.

          ipar(1) = -9

       endif

    else if (alpha.lt.zero) then

       beta = one / alpha

       if (.not. beta.lt.zero) then

          brkdn = .true.

          ipar(1) = -9

       endif

    else if (alpha.eq.zero) then

       brkdn = .true.

       ipar(1) = -3

    else

       brkdn = .true.

       ipar(1) = -9

    endif

    return

  end function brkdn

  !-----end-of-brkdn



  logical function stopbis(n,ipar,mvpi,fpar,r,delx,sx)

    implicit none

    integer n,mvpi,ipar(16)

    real*8 fpar(16), r(n), delx(n), sx

    !-----------------------------------------------------------------------

    !     function for determining the stopping criteria. return value of

    !     true if the stopbis criteria is satisfied.

    !-----------------------------------------------------------------------

    if (ipar(11) .eq. 1) then

       stopbis = .true.

    else

       stopbis = .false.

    endif

    if (ipar(6).gt.0 .and. ipar(7).ge.ipar(6)) then

       ipar(1) = -1

       stopbis = .true.

    endif

    if (stopbis) return



    ! computes errors

    fpar(5) = sqrt(distdot(n,r,r))

    fpar(11) = fpar(11) + 2 * n

    if (ipar(3).lt.0) then



       !     compute the change in the solution vector

       fpar(6) = sx * sqrt(distdot(n,delx,delx))

       fpar(11) = fpar(11) + 2 * n

       if (ipar(7).lt.mvpi+mvpi+1) then



          ! if this is the end of the first iteration, set fpar(3:4)

          fpar(3) = fpar(6)

          if (ipar(3).eq.-1) then

             fpar(4) = fpar(1) * fpar(3) + fpar(2)

          endif

       endif

    else

       fpar(6) = fpar(5)

    endif



    ! the test is struct this way so that when the value in fpar(6)

    ! is not a valid number, STOPBIS is set to .true.

    if (fpar(6).gt.fpar(4)) then

       stopbis = .false.

       ipar(11) = 0

    else

       stopbis = .true.

       ipar(11) = 1

    endif

    return

  end function stopbis

  !-----end-of-stopbis





  subroutine bisinit(ipar,fpar,wksize,dsc,lp,rp,wk)

    implicit none

    integer i,ipar(16),wksize,dsc

    logical lp,rp

    real*8  fpar(16),wk(*)

    !-----------------------------------------------------------------------

    !     some common initializations for the iterative solvers

    !-----------------------------------------------------------------------

    real*8 zero, one

    parameter(zero=0.0D0, one=1.0D0)



    ! ipar(1) = -2 inidcate that there are not enough space in the work

    ! array

    if (ipar(4).lt.wksize) then

       ipar(1) = -2

       ipar(4) = wksize

       return

    endif



    if (ipar(2).gt.2) then

       lp = .true.

       rp = .true.

    else if (ipar(2).eq.2) then

       lp = .false.

       rp = .true.

    else if (ipar(2).eq.1) then

       lp = .true.

       rp = .false.

    else

       lp = .false.

       rp = .false.

    endif

    if (ipar(3).eq.0) ipar(3) = dsc

    ! clear the ipar elements used

    ipar(7) = 0

    ipar(8) = 0

    ipar(9) = 0

    ipar(10) = 0

    ipar(11) = 0

    ipar(12) = 0

    ipar(13) = 0



    ! fpar(1) must be between (0, 1), fpar(2) must be positive,

    ! fpar(1) and fpar(2) can NOT both be zero

    ! Normally return ipar(1) = -4 to indicate any of above error

    if (fpar(1).lt.zero .or. fpar(1).ge.one .or. fpar(2).lt.zero .or. (fpar(1).eq.zero .and. fpar(2).eq.zero)) then

       if (ipar(1).eq.0) then

          ipar(1) = -4

          return

       else

          fpar(1) = 1.0D-6

          fpar(2) = 1.0D-16

       endif

    endif

    ! clear the fpar elements

    do i = 3, 10

       fpar(i) = zero

    enddo

    if (fpar(11).lt.zero) fpar(11) = zero

    ! clear the used portion of the work array to zero

    do i = 1, wksize

       wk(i) = zero

    enddo



    return

    !-----end-of-bisinit

  end subroutine bisinit

end module sparskit

