module linmod
  !module for linear algebra subroutines, including lapack call.
  implicit none
  private
  logical :: linmod_error
  public :: matinv,matinv_lapack,linmod_error
contains
  function kron(A,B) result(res) !Kronecker product of two matrices.
    real(8), dimension(:,:), intent(in) :: A
    real(8), dimension(:,:), intent(in) :: B
    real(8) :: res(size(A,dim=1)*size(B,dim=1),size(A,dim=2)*size(B,dim=2))
    integer :: i,j,m,n,p,q
    m=size(A,dim=1)
    n=size(A,dim=2)
    p=size(B,dim=1)
    q=size(B,dim=2)
    do i=1,m
       do j=1,n
          res(1+(i-1)*p:i*p,1+j*q:j*q)=A(i,j)*B
       end do
    end do
  end function kron

  function ones(dim) result(res) !generates a simple dim X dim matrix of ones.
    integer :: dim
    real(8) :: res(dim,dim)
    res=1.d0
  end function ones

  function eye(dim) result(res) !geneates a simple dim X dim identity matrix.
    integer :: dim
    real(8) :: res(dim,dim)
    integer :: i
    res=0.d0
    do i=1,dim
       res(i,i)=1.d0
    end do
  end function eye
  
  function matinv_lapack(A) result(res)
    implicit none
    real(8), dimension(:,:), intent(in) :: A
    real(8), dimension(size(A,1),size(A,2)) :: res
    real(8) :: work(size(A,1))
    integer, dimension(size(A,1)) :: ipiv
    integer :: n, info
    res=a
    n=size(A,1)
    call dgetrf(n,n,res,n,ipiv,info)
    if(info/=0) stop 'Matrix is numerically singular'
    call dgetri(n,res,n,ipiv,work,n,info)
    if(info/=0) stop 'Matrix is numerically singular'
  end function matinv_lapack

  function matinv(a)
    REAL(8), DIMENSION(:,:), INTENT(IN) :: a
    logical :: error
    real(8), dimension(size(a,1),size(a,1)) :: matinv
    real(8) :: b(size(a,1),1)
    linmod_error=.false.
    b=1.0d0
    matinv=a
    call gaussj(matinv,b,error)
    if(error) then
       matinv=-1.0d0
       linmod_error=.true.
    end if
  end function matinv

  SUBROUTINE gaussj(a,b,error)
    REAL(8), DIMENSION(:,:), INTENT(INOUT) :: a,b
    logical, intent(out) :: error
    INTEGER(4), DIMENSION(size(a,1)) :: ipiv,indxr,indxc
    LOGICAL, DIMENSION(size(a,1)) :: lpiv
    REAL(8) :: pivinv
    REAL(8), DIMENSION(size(a,1)) :: dumc
    INTEGER(4), TARGET :: irc(2)
    INTEGER(4) :: i,l,n
    INTEGER(4), POINTER :: irow,icol
    error=.false.
    n=assert_eq(size(a,1),size(a,2),size(b,1),'gaussj')
    irow => irc(1)
    icol => irc(2)
    ipiv=0
    do i=1,n
       lpiv = (ipiv == 0)
       irc=maxloc(abs(a),outerand(lpiv,lpiv))
       ipiv(icol)=ipiv(icol)+1
       if (ipiv(icol) > 1) then
          write(6,'("Fatal(gaussj): Matrix inversion failed (1).")')
          error=.true.
          return
       end if
       if (irow /= icol) then
          call swap(a(irow,:),a(icol,:))
          call swap(b(irow,:),b(icol,:))
       end if
       indxr(i)=irow
       indxc(i)=icol
       if (a(icol,icol) == 0.0d0) then
          write(6,'("Fatal(gaussj): Matrix inversion failed (2).")')
          error=.true.
          return
       end if
       pivinv=1.0d0/a(icol,icol)
       a(icol,icol)=1.0d0
       a(icol,:)=a(icol,:)*pivinv
       b(icol,:)=b(icol,:)*pivinv
       dumc=a(:,icol)
       a(:,icol)=0.0d0
       a(icol,icol)=pivinv
       a(1:icol-1,:)=a(1:icol-1,:)-outerprod(dumc(1:icol-1),a(icol,:))
       b(1:icol-1,:)=b(1:icol-1,:)-outerprod(dumc(1:icol-1),b(icol,:))
       a(icol+1:,:)=a(icol+1:,:)-outerprod(dumc(icol+1:),a(icol,:))
       b(icol+1:,:)=b(icol+1:,:)-outerprod(dumc(icol+1:),b(icol,:))
    end do
    do l=n,1,-1
       call swap(a(:,indxr(l)),a(:,indxc(l)))
    end do
  END SUBROUTINE gaussj

  FUNCTION assert_eq(n1,n2,n3,string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    INTEGER(4), INTENT(IN) :: n1,n2,n3
    INTEGER(4) :: assert_eq
    if (n1 == n2 .and. n2 == n3) then
       assert_eq=n1
    else
       write (*,*) 'nrerror: an assert_eq failed with this tag:', &
            string
       STOP 'program terminated by assert_eq3'
    end if
  END FUNCTION assert_eq

  SUBROUTINE nrerror(string)
    CHARACTER(LEN=*), INTENT(IN) :: string
    write (*,*) 'nrerror: ',string
    STOP 'program terminated by nrerror'
  END SUBROUTINE nrerror

  FUNCTION outerand(a,b)
    LOGICAL, DIMENSION(:), INTENT(IN) :: a,b
    LOGICAL, DIMENSION(size(a),size(b)) :: outerand
    outerand = spread(a,dim=2,ncopies=size(b)) .and. &
         spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerand

  FUNCTION outerprod(a,b)
    REAL(8), DIMENSION(:), INTENT(IN) :: a,b
    REAL(8), DIMENSION(size(a),size(b)) :: outerprod
    outerprod = spread(a,dim=2,ncopies=size(b)) * &
         spread(b,dim=1,ncopies=size(a))
  END FUNCTION outerprod

  SUBROUTINE swap(a,b)
    REAL(8), DIMENSION(:), INTENT(INOUT) :: a,b
    REAL(8), DIMENSION(SIZE(a)) :: dum
    dum=a
    a=b
    b=dum
  END SUBROUTINE swap
end module linmod

module marsaglia
  !module for the random number generation.
  implicit none
  private
  type kisseed
     integer(4) :: x,y,z,w
  end type kisseed
  public :: kiss, kisset,kissuni,kissgaus,kisseed, setseed
  INTEGER(4) :: x=123456789, y=362436069, z=521288629, w=916191069
contains
  function kiss()
    integer(4) :: kiss
    ! The  KISS (Keep It Simple Stupid) random number generator. Combines:
    ! (1) The congruential generator x(n)=69069*x(n-1)+1327217885, period 2^32.
    ! (2) A 3-shift shift-register generator, period 2^32-1,
    ! (3) Two 16-bit multiply-with-carry generators, period 597273182964842497>2^59
    !  Overall period>2^123;  Default seeds x,y,z,w.
    !  Set your own seeds with statement i=kisset(ix,iy,iz,iw).

    x = 69069 * x + 1327217885
    y = m (m (m (y, 13), - 17), 5)
    z = 18000 * iand (z, 65535) + ishft (z, - 16)
    w = 30903 * iand (w, 65535) + ishft (w, - 16)
    kiss = x + y + ishft (z, 16) + w
  contains
    function m(k, n)
      integer :: m, k, n
      m = ieor (k, ishft (k, n) )
    end function m
  end function kiss

  function kisset (inp)
    integer(4) :: kisset
    type(kisseed), intent(in) :: inp
    x = inp%x
    y = inp%y
    z = inp%z
    w = inp%w
    kisset = 1
  end function kisset

  function kissuni()
    !Produces one of 2^32 possible realizations in the (0,1) interval (not including 0 and 1).
    real(8) :: kissuni
    kissuni=(dble(kiss())+2147483649.0d0)/4294967297.0d0
  end function kissuni

  function kissgaus()
    !normal rng based on simple box muller transformation. Slow.
    REAL(8) :: kissgaus
    REAL(8) :: rsq,v1,v2
    REAL(8), SAVE :: g
    LOGICAL, SAVE :: gaus_stored=.false.
    logical :: notdone
    if (gaus_stored) then
       kissgaus=g
       gaus_stored=.false.
    else
       notdone=.true.
       do while(notdone)
          v1=kissuni()
          v2=kissuni()
          v1=2.0d0*v1-1.0d0
          v2=2.0d0*v2-1.0d0
          rsq=v1**2+v2**2
          if (rsq > 0.0d0 .and. rsq < 1.0d0) then
             rsq=sqrt(-2.0d0*log(rsq)/rsq)
             if(rsq==rsq) notdone=.false.
          end if
       end do
       kissgaus=v1*rsq
       g=v2*rsq
       gaus_stored=.true.
    end if
  end function kissgaus

  subroutine setseed(ms)
    !kisseed(2345678,87656789,5434567,43212345)
    type(kisseed), intent(in) :: ms !Set master seed here.
    type(kisseed) :: mys
    integer :: i,j,imc
    logical :: tryagain
    i=kisset(ms)
    do i=1,10000
       j=kiss()
    end do
    tryagain=.true.
    do while (tryagain)
       do i=1,4 !x and y seed can be 32 bit integers. z and w seeds no more than 31 bit integers and not multiples of 7559.
          mys%x=kiss()
          mys%y=kiss()
          mys%z=ishft(kiss(),-2)
          mys%w=ishft(kiss(),-2)
       end do
       if(mod(mys%z,7559)/=0 .and. mod(mys%w,7559)/=0) tryagain=.false.
    end do
    i=kisset(mys) !seed the process
    do i=1,10000 !warm up the sequence. We are set to go.
       j=kiss()
    end do
  end subroutine setseed
end module marsaglia

module tools
  implicit none
contains
  function findndx(arr,val,outp) !Determine the indx such that arr(ndx)<=val<arr(ndx+1). arr must be sorted in ascending order
    real(8), dimension(:), intent(in) :: arr
    real(8), intent(in) :: val
    integer, intent(in), optional :: outp
    integer :: findndx
    integer :: lb,ub,dim,i
    dim=size(arr)
    if(dim<2) stop 'Fatal(findndx): dimension of input array <2'
    lb=1
    ub=dim
    if(arr(lb)>val .or. arr(ub)<val) then
       write(6,'(f20.10)') arr
       write(6,'(" ")')
       write(6,'(f20.10)') val
       write(6,'(i10,f15.5)') dim,arr(dim)
       if(present(outp)) write(6,'(i5)') outp
       stop 'Fatal(findndx): val not in arr support.'
    end if
    if(arr(ub) == val) then
       findndx=ub
       return
    end if
    findndx=(ub+lb)/2
    i=1
    do while ((arr(findndx) .gt. val .or. arr(findndx+1) .le. val) .and. i<100) !notice, a bit of ambiguity in the return
       !if val hits exactly a knot point.
       if(arr(findndx) .gt. val) then
          ub=findndx
       else
          lb=findndx
       end if
       findndx=(lb+ub)/2
       i=i+1
    end do
    if(i==100) then
       write(6,'(f20.10)') arr
       write(6,'(" ")')
       write(6,'(f20.10)') val
       stop 'Fatal(findndx): solver did not reach criterion.'
    end if
  end function findndx

  function corr_wght(xdim,ydim,x,y,pdf) result(res)
    implicit none
    integer, intent(in) :: xdim,ydim
    real(8), intent(in) :: x(xdim),y(ydim),pdf(xdim,ydim)
    real(8) :: res,ex,ey,exx,eyy,exy
    integer :: i,j
    ex=0.d0
    ey=0.d0
    exx=0.d0
    eyy=0.d0
    exy=0.d0
    do i=1,xdim
       do j=1,ydim
          exy=exy+x(i)*y(j)*pdf(i,j)
          exx=exx+x(i)*x(i)*pdf(i,j)
          eyy=eyy+y(j)*y(j)*pdf(i,j)
          ex=ex+x(i)*pdf(i,j)
          ey=ey+y(j)*pdf(i,j)
       end do
    end do
    res=(exy-ex*ey)/(sqrt(exx-ex*ex)*sqrt(eyy-ey*ey))
  end function corr_wght

  function inttochar(dim,x) result(res)
    implicit none
    integer, intent(in) :: dim,x
    character(dim) :: res
    integer :: rem,i,dig
    character(1) :: num
    rem=x
    res=""
    do i=dim-1,0,-1
       dig=int(dble(rem)/dble(10.d0**i))
       write(num,'(i1)') dig
       res=trim(res)//num
       rem=rem-int(dble(dig)*(10.d0**i))
    end do
  end function inttochar

  function corr(x,y) result(res)
    implicit none
    real(8), dimension(:), intent(in) :: x,y
    real(8) :: res,ex,ey,exx,eyy,exy
    integer :: i,dim
    dim=size(x)
    if(dim .ne. size(y)) stop "x and y dimensions do not agree."
    ex=0.d0
    ey=0.d0
    exx=0.d0
    eyy=0.d0
    exy=0.d0
    do i=1,dim
       exy=exy+x(i)*y(i)/dble(dim)
       exx=exx+x(i)*x(i)/dble(dim)
       eyy=eyy+y(i)*y(i)/dble(dim)
       ex=ex+x(i)/dble(dim)
       ey=ey+y(i)/dble(dim)
    end do
    res=(exy-ex*ey)/(sqrt(exx-ex*ex)*sqrt(eyy-ey*ey))
  end function corr
end module tools

module sort3
  implicit none
  private
  interface hpsort
     module procedure hpsort_real, hpsort_int, hpsort_real_pert
  end interface
  public HPSORT

contains

  !*****************************************************
  !*  Sorts an array RA of length N in ascending order *
  !*                by the Heapsort method             *
  !* ------------------------------------------------- *
  !* INPUTS:                                           *
  !*	    N	  size of table RA                     *
  !*         RA	  table to be sorted            	    *
  !* OUTPUT:                                           *
  !*	    RA    table sorted in ascending order     *
  !*                                                   *
  !* NOTE: The Heapsort method is a N Log2 N routine,  *
  !*       and can be used for very large arrays.      *
  !*****************************************************         
  SUBROUTINE HPSORT_real(N,RA)
    implicit none
    integer, intent(in) :: N
    real(8), intent(inout) :: RA(N)
    integer :: L,IR,I,J
    real(8) :: RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR 
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.
10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
    else
       RRA=RA(IR)
       RA(IR)=RA(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    goto 10
  END SUBROUTINE HPSORT_real

  SUBROUTINE HPSORT_real_pert(N,RA,INDX) !
    implicit none
    integer, intent(in) :: N
    real(8), intent(inout) :: RA(N)
    integer, intent(inout) :: INDX(N)
    integer :: L,IR,I,J,IRRA
    real(8) :: RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR 
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.

!!$    do i=1,N !Initial order
!!$	indx(i)=i
!!$    end do

10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
       irra=indx(L)
    else
       RRA=RA(IR)
       irra=indx(IR)
       RA(IR)=RA(1)
       indx(IR)=indx(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
	   indx(1)=irra
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
	   indx(I)=indx(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    indx(I)=irra
    goto 10
  END SUBROUTINE HPSORT_real_pert

  SUBROUTINE HPSORT_int(N,RA)
    implicit none
    integer, intent(in) :: N
    integer(4), intent(inout) :: RA(N)
    integer :: L,IR,I,J,RRA
    L=N/2+1
    IR=N
    !The index L will be decremented from its initial value during the
    !"hiring" (heap creation) phase. Once it reaches 1, the index IR 
    !will be decremented from its initial value down to 1 during the
    !"retirement-and-promotion" (heap selection) phase.
10  continue
    if(L > 1)then
       L=L-1
       RRA=RA(L)
    else
       RRA=RA(IR)
       RA(IR)=RA(1)
       IR=IR-1
       if(IR.eq.1)then
          RA(1)=RRA
          return
       end if
    end if
    I=L
    J=L+L
20  if(J.le.IR)then
       if(J < IR)then
          if(RA(J) < RA(J+1))  J=J+1
       end if
       if(RRA < RA(J))then
          RA(I)=RA(J)
          I=J; J=J+J
       else
          J=IR+1
       end if
       goto 20
    end if
    RA(I)=RRA
    goto 10
  END SUBROUTINE HPSORT_int
end module sort3
