module class_spell
  implicit none !all objects in module inherit implicit none
  type spell
     integer(4) :: fid !Firm ID !Can go and look at characteristics data with IDs.
     integer(4) :: wid !Worker ID - Can go and look at characteristics with worker ID.
     real(4), pointer :: wage(:)   !can have multiple observations for each year the spell stretches over.
     integer(1), pointer :: tenure(:) !tenure status
     integer(2) :: strt !start week
     integer(2) :: end !end week
     integer(2) :: ten !tenure at wage(1) - measured in weeks.
     integer(1) :: yr !yr of spell start
     type(spell), pointer :: prev=>null(), next => null()
     type(spell), pointer :: f_prev=>null(), f_next=>null() !to navigate firm panels.
  end type spell
  type(spell), pointer :: this !current spell
  interface find_t
     module procedure find_t_d,find_t_i
  end interface find_t
contains
!!$  subroutine addspell !add a spell to this, and make this point to the newly added spell.
!!$    type(spell), pointer :: new
!!$    allocate(new)
!!$    new%prev=>this !create backward link
!!$    this%next=>new !create forward link
!!$    nullify(new%next)
!!$    this=>new
!!$  end subroutine addspell

  subroutine newspell !create a new spell, and make this point to it.
    type(spell), pointer :: new
    allocate(new)
    this=>new
  end subroutine newspell

  recursive subroutine remove_spell(mythis) !Removes this and all elements following it.
    type(spell), pointer :: mythis, next
    next => mythis%next
    do while(associated(next))
       call remove_spell(next)
    end do
    if(associated(mythis%wage)) then
       deallocate(mythis%wage,mythis%tenure) !step in and deallocate wage inside spell.
    end if
    deallocate(mythis)
  end subroutine remove_spell

  recursive function persyr_firm(myspell) result(res)
    type(spell), pointer, intent(in) :: myspell
    integer(4) :: res
    res=0
    if(.not. associated(myspell)) then
       return
    else
       if(associated(myspell%wage)) res=size(myspell%wage)
    end if
    if(associated(myspell%f_next)) res=res+persyr_firm(myspell%f_next)
  end function persyr_firm

  recursive subroutine find_t_i(myspell,myt)
    type(spell), pointer, intent(in) :: myspell
    integer, intent(in) :: myt
    this=>myspell
    if(myspell%strt < myt .and. associated(myspell%next)) call find_t(myspell%next,myt)
  end subroutine find_t_i

  recursive subroutine find_t_d(myspell,myt)
    type(spell), pointer, intent(in) :: myspell
    real(8), intent(in) :: myt
    this=>myspell
    if(myspell%strt < myt .and. associated(myspell%next)) call find_t_d(myspell%next,myt)
  end subroutine find_t_d

  function dur(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4) :: res
    res=mythis%end-mythis%strt+1
  end function dur
  
  function npvcal(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    real(8) :: res, r
	integer(4) :: t
	r=0.03 !discount rate
	res=0.d0
	do t=1,size(mythis%wage)
	  res = res+mythis%wage(t)/(1+r)**t
	end do
  end function npvcal  
end module class_spell

module panels
  use class_spell
  type workerlist
     real(8) :: akm_wage
     integer(8) :: id
     integer(4) :: entry
     integer(4) :: byr !birth year
     integer(4) :: gradyr !graduation yr
     integer(1) :: edu
     integer :: k !worker type (for simulation)
     integer(1) :: z !worker z
     logical :: male
     logical :: native
     logical :: akm_defined
     type(spell), pointer :: first=>null(),last=>null()
  end type workerlist

  type firmlist
     integer :: sector !private sector
     integer :: l !firm type (for simulation)
     integer(8) :: id
     type(spell), pointer :: first=>null(),last=>null()
  end type firmlist
contains
  subroutine add_worker(mythis,myspell)
    type(workerlist) :: mythis
    type(spell), pointer, intent(inout) :: myspell
    if(associated(mythis%last)) then
       myspell%prev=>mythis%last
       mythis%last%next=>myspell
       mythis%last=>myspell
    else
       mythis%first=>myspell
       mythis%last=>myspell
    end if
  end subroutine add_worker

  subroutine add_firm(mythis,myspell)
    type(firmlist) :: mythis
    type(spell), pointer, intent(inout) :: myspell
    if(associated(mythis%last)) then
       myspell%f_prev=>mythis%last
       mythis%last%f_next=>myspell
       mythis%last=>myspell
    else
       mythis%first=>myspell
       mythis%last=>myspell
    end if
  end subroutine add_firm

  function persyr(this) result(res)
    type(firmlist) :: this
    integer(4) :: res
    res=persyr_firm(this%first)
  end function persyr

  recursive subroutine write_spell(this,mych,mythis) !write out a spell to output channel mych
    type(workerlist) :: this
    integer, intent(in) :: mych
    type(spell), pointer, intent(in) :: mythis
    integer :: i
    write(mych,'(6i10)') mythis%wid, mythis%fid, mythis%yr, mythis%ten, mythis%strt,mythis%end
    write(mych,'(a15,*(i15))') " ",(i,i=1,size(mythis%wage))
    write(mych,'(a15,*(f15.4))') "Wage",(mythis%wage(i),i=1,size(mythis%wage))
    write(mych,'(a15,*(I15))') "tenure",(mythis%tenure(i),i=1,size(mythis%wage))
    write(mych,'(" ")')
    if(associated(mythis%next)) call write_spell(this,mych,mythis%next)
  end subroutine write_spell
end module panels

module estim
  use linmod
  use class_spell
  use panels
  use modelpar_class
  implicit none
  
  type(modelpar) :: ep !Here, hold the estimated model parameters.
!!$  type(modelpar) :: truth
  real(8),allocatable :: pprobw(:,:)  ! estimated posterior probabilities of worker types from step 2
  integer,allocatable :: firmtype(:)  !firm type assigned by max posterior prob
  integer,allocatable :: firmsize(:) !number of spells in firm.
  integer(4) :: itcount,globi,yrtoweek(80)

  !The major data panels
  type(workerlist), allocatable :: work(:) !these two structures hold the spell data
  type(firmlist), allocatable :: firm(:)
  integer :: dim_I,dim_J
  integer :: true_ck, true_cl 
  
  !procedure overloads.
!!$  interface mprob
!!$     module procedure mprob_i4,mprob_i1
!!$  end interface mprob
  interface find
     module procedure find_d,find_i
  end interface find
  interface fdensity
     module procedure fdensity_i4,fdensity_i1
  end interface fdensity  
contains
  recursive subroutine coworker_akm(mysp,ex_wid,akmsum,akmcount) !calculate a firm's average worker akm wage effects excluding ex_wid
    integer, intent(in) :: ex_wid
    type(spell), pointer, intent(in) :: mysp
    real(8), intent(inout) :: akmsum,akmcount
    if(.not. associated(mysp)) return
    if(mysp%wid .ne. ex_wid .and. work(mysp%wid)%akm_defined) then
       akmcount=akmcount+dble(size(mysp%wage))
       akmsum=akmsum+dble(size(mysp%wage))*work(mysp%wid)%akm_wage
    end if
    call coworker_akm(mysp%f_next,ex_wid,akmsum,akmcount)
  end subroutine coworker_akm

  subroutine estim_ini(mydim,dim_I,dim_J) !Estimation is initialized.
    type(pardim), intent(in) :: mydim
    integer, intent(in) :: dim_I,dim_J
    call modelpar_deallocate(ep)
    ep=new_modelpar_new(mydim)
    if(allocated(pprobw)) deallocate(pprobw)
    if(allocated(firmtype)) deallocate(firmtype)
    if(allocated(firmsize)) deallocate(firmsize)
    allocate(pprobw(dim_I,mydim%k),firmtype(0:dim_J),firmsize(1:dim_J))
  end subroutine estim_ini

  subroutine estim_term
    deallocate(pprobw,firmtype,firmsize)
    call modelpar_deallocate(ep)
  end subroutine estim_term

  function shorttenweeks(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer :: res
    if(mythis%fid==0) then
       res=max(0,min(mythis%end-1-mythis%strt,27-mythis%ten))
    else
       res=max(0,min(mythis%end-1-mythis%strt,100-mythis%ten))
    end if
  end function shorttenweeks

  function longtenweeks(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer :: res
    res=max(0,mythis%end-1-mythis%strt-shorttenweeks(mythis))
  end function longtenweeks

  function lastweekten(mythis) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer :: res
    if(mythis%fid==0) then
       if(mythis%end-mythis%strt>27-mythis%ten) then
          res=2
       else
          res=1
       end if
    else
       if(mythis%end-mythis%strt>100-mythis%ten) then
          res=2
       else
          res=1
       end if
    end if
  end function lastweekten
  
  function hetmap(mysp,myj) result(res) !maps into 8 integers (conditional on employment)
    type(spell), pointer :: mysp
    integer, intent(in) :: myj !wage observation number within spell
    integer :: res
    integer :: myexp

    !Potential experience mapping is set (simulation starts from 20 years old)
    res=1
    if(mysp%strt+52*myj .gt. 260) res=res+1
    if(mysp%strt+52*myj .gt. 520) res=res+1
    if(mysp%strt+52*myj .gt. 780) res=res+1
    if(myj<3) then
       res=res
    else
       res=res+4
    end if
    
    
    return
	
    ! res=1
    ! if(mysp%fid>0) res=mysp%tenure(myj)+1 !When wage observation use tenure variable
    ! if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+1 !nonemployed tenure. distinguish between first 6 months and everything else.
    ! return
  
    myexp=mysp%yr+myj-work(mysp%wid)%gradyr
    if(myexp<=5) res=1
    if(myexp>5 .and. myexp<=10) res=2
    if(myexp>10 .and. myexp<=15) res=3
    if(myexp>15) res=4

    !Time mapping is set (hetmap=1...4 is for 1987-1989)
    if(mysp%yr+myj>56 .and. mysp%yr+myj<60) res=res+4 	!1990-1992
    if(mysp%yr+myj>59 .and. mysp%yr+myj<63) res=res+8 	!1993-1995
    if(mysp%yr+myj>62 .and. mysp%yr+myj<66) res=res+12 	!1996-1998
    if(mysp%yr+myj>65 .and. mysp%yr+myj<69) res=res+16 	!1999-2001
    if(mysp%yr+myj>68 .and. mysp%yr+myj<72) res=res+20 	!2002-2004
    if(mysp%yr+myj>71 .and. mysp%yr+myj<75) res=res+24 	!2005-2007
    if(mysp%yr+myj>74 .and. mysp%yr+myj<78) res=res+28 	!2008-2010
    if(mysp%yr+myj>77) res=res+32 			!2011-2013

    !Tenure mapping is set
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+36 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+36 ! nonemployed tenure. distinguish between first 6 months and everything else.
  end function hetmap

  function inihetmap(mysp) result(res) !maps into [1,16] integers
    type(spell), pointer :: mysp
    integer :: res
    integer :: myexp
    res=1
    !Tenure mapping is set
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+1 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+1 ! nonemployed tenure. distinguish between first 6 months and everything else.
    return
    !Dimensions of heterogeneity: exp (4) - only for first time group, time (9), tenure (2) (24 groups)
    !Age mapping is set	
    myexp=mysp%yr-work(mysp%wid)%gradyr
    if(myexp<=5) res=1
    if(myexp>5 .and. myexp<=10) res=2
    if(myexp>10 .and. myexp<=15) res=3
    if(myexp>15) res=4

    !Tenure mapping is set
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+12 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+12 ! nonemployed tenure. distinguish between first 6 months and everything else.

    !entry year after first period can only be for low tenure category.
    if(mysp%yr>56 .and. mysp%yr<60) res=5
    if(mysp%yr>59 .and. mysp%yr<63) res=6
    if(mysp%yr>62 .and. mysp%yr<66) res=7
    if(mysp%yr>65 .and. mysp%yr<69) res=8
    if(mysp%yr>68 .and. mysp%yr<72) res=9
    if(mysp%yr>71 .and. mysp%yr<75) res=10
    if(mysp%yr>74 .and. mysp%yr<78) res=11
    if(mysp%yr>77) res=12
  end function inihetmap

  function mobhetmap(mysp,myj) result(res)
    type(spell), pointer :: mysp
    integer, intent(in) :: myj
    integer :: res,myexp

	
    res=1
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+1 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+1 ! nonemployed tenure. distinguish between first 6 months and everything else.
    return

    myexp=mysp%yr+myj-work(mysp%wid)%gradyr
    if(myexp<=5) res=1
    if(myexp>5 .and. myexp<=10) res=2
    if(myexp>10 .and. myexp<=15) res=3
    if(myexp>15) res=4

    !Time mapping is set (res=1...4 is for 1987-1989)
    if(mysp%yr+myj>56 .and. mysp%yr+myj<60) res=res+4 	!1990-1992, 1990 is year 56. myj lives on [1,size(wage)]
    if(mysp%yr+myj>59 .and. mysp%yr+myj<63) res=res+8 	!1993-1995
    if(mysp%yr+myj>62 .and. mysp%yr+myj<66) res=res+12 	!1996-1998
    if(mysp%yr+myj>65 .and. mysp%yr+myj<69) res=res+16 	!1999-2001
    if(mysp%yr+myj>68 .and. mysp%yr+myj<72) res=res+20 	!2002-2004
    if(mysp%yr+myj>71 .and. mysp%yr+myj<75) res=res+24 	!2005-2007
    if(mysp%yr+myj>74 .and. mysp%yr+myj<78) res=res+28 	!2008-2010
    if(mysp%yr+myj>77) res=res+32 			!2011-2013

    !Tenure mapping is set
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>100 .and. mysp%fid>0) res=res+36 ! employed tenure. distinguish between first two years and everything else.
    if(yrtoweek(mysp%yr+myj-1)-mysp%strt+mysp%ten>26 .and. mysp%fid==0) res=res+36 ! nonemployed tenure. distinguish between first 6 months and everything else.
  end function mobhetmap

  function zhetmap(mywork) result(res)
    type(workerlist), intent(in) :: mywork
    integer ::res
    res=mywork%z
    return
    if(mywork%edu<12) res=1
    if(mywork%edu==12) res=2
    if(mywork%edu>12) res=3
    if(mywork%male) res=res+3
  end function zhetmap

  function mobtonumap(myt) result(res)
    integer, intent(in) :: myt
    integer :: res
    res=1
    return
    res=0
    if((myt> 0 .and. myt< 5) .or. (myt>36 .and. myt<41)) res=1
    if((myt> 4 .and. myt< 9) .or. (myt>40 .and. myt<45)) res=2
    if((myt> 8 .and. myt<13) .or. (myt>44 .and. myt<49)) res=3
    if((myt>12 .and. myt<17) .or. (myt>48 .and. myt<53)) res=4
    if((myt>16 .and. myt<21) .or. (myt>52 .and. myt<57)) res=5
    if((myt>20 .and. myt<25) .or. (myt>56 .and. myt<61)) res=6
    if((myt>24 .and. myt<29) .or. (myt>60 .and. myt<65)) res=7
    if((myt>28 .and. myt<33) .or. (myt>64 .and. myt<69)) res=8
    if((myt>32 .and. myt<37) .or. (myt>68 .and. myt<73)) res=9
    if(res==0) stop 'fatal(mobtonumap): input outside of range'
  end function mobtonumap

!!$  subroutine readest(folder,name,res,myxdim,myinidim,myzdim,kdim,ldim)
!!$    use mpimod
!!$    use paths
!!$    use sim_ss_ten
!!$    use sort3
!!$    character(*), intent(in) :: folder, name
!!$    integer, intent(in) :: myxdim,myinidim,myzdim,kdim,ldim
!!$    type(modelpar), intent(out) :: res
!!$    integer :: ki,li,xi,zi,ii,jj
!!$    real(8) :: xx
!!$    integer, allocatable :: kmap(:),lmap(:)
!!$	real(8), allocatable :: temp(:,:,:)
!!$	
!!$	call modelpar_deallocate(res)
!!$	
!!$	if(allocated(kmap)) deallocate(kmap,lmap,temp)
!!$	allocate(kmap(kdim), lmap(0:ldim))
!!$	
!!$	
!!$    open(10,file=estpath//trim(folder)//'/inputs/'//trim(name)//"LPR.txt",status='old')
!!$	read(10,*) kmap
!!$	read(10,*) lmap
!!$    read(10,*) ki, li
!!$    
!!$	res=new_modelpar(ki,li,myxdim,myzdim,myinidim,myxdim/2,myxdim+4,36) 
!!$	!res has now been initialized. the last input to new_modelpar is dimension of industry*sector
!!$	
!!$	
!!$  read(10,*) (((res%muw(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%xdim)
!!$  read(10,*) (((res%sigmaw(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%sigdim)
!!$  read(10,*) (((res%gam(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%xdim)
!!$  read(10,*) (((res%lam(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%xdim)
!!$  read(10,*) (((res%psi(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%xdim)
!!$  read(10,*) (((res%del(ki,li,xi),li=1,res%l),ki=1,res%k),xi=1,res%xdim)
!!$  read(10,*) ((res%piw(xi,ki),xi=1,res%zdim),ki=1,res%k)
!!$  read(10,*) ((res%pif(xi,li),xi=1,res%sdim),li=1,res%l)
!!$  read(10,*) (((res%m(ki,li,xi),li=0,res%l),ki=1,res%k),xi=1,res%mdim)
!!$  read(10,*) (res%q(li),li=1,res%l)
!!$  read(10,*) (res%pk(jj), jj=1,res%k)
!!$  read(10,*) (res%pz(ii), ii=1,res%zdim)
!!$  read(10,*) (res%ps(ii), ii=1,res%sdim)
!!$  read(10,*) res%rho
!!$  close(10)
!!$   
!!$  
!!$  allocate(temp(res%k,0:res%l,res%xdim))
!!$  open(10,file=estpath//trim(folder)//'/inputs/'//trim(name)//'_mobpkl.txt',status='old')
!!$  read(10,*) xi, ki, li
!!$  read(10,*) (((temp(ki,li,xi), li=0,res%l),ki=1,res%k),xi=1,res%xdim) !p_{kl}(xi). Sums to one for given xi.
!!$  close(10)
!!$  
!!$  do xi=1,res%xdim
!!$  do ki=1,res%k
!!$  do li=0,res%l
!!$	    res%pkl(ki,li,xi)=temp(kmap(ki),lmap(li),xi)
!!$  end do
!!$  end do
!!$  end do
!!$  
!!$  
!!$  open(10,file=estpath//trim(folder)//'/inputs/'//trim(name)//"_hetdist.txt",status='old')
!!$  read(10,*) ((ki, xx, res%p(xi)),xi=1,res%xdim)
!!$  close(10)
!!$  
!!$  call M_update(res) !mobility matrices stored in MM and Mbar
!!$  call steadystate(res) !x conditional stationary distributions stored in ss
!!$
!!$  
!!$end subroutine readest

  
  function find_d(in1,in2) result(res)
    real(8), dimension(:), intent(in) :: in1
    real(8), intent(in) :: in2
    integer :: res
    integer :: dim,lb,ub
    if(in2<=in1(1)) then
       res=1
       return
    end if
    dim=size(in1)
    if(in2>in1(dim)) then
       res=0
       return
    end if
    lb=1
    ub=dim
    do while(ub-lb>1)
       res=(ub+lb)/2
       if(in1(res)<in2) lb=res
       if(in1(res)>=in2) ub=res
    end do
    res=ub
    if(in1(res-1)>=in2 .or. in1(res)<in2) stop 'find failed.'
  end function find_d

  function find_i(in1,in2) result(res)
    integer, dimension(:), intent(in) :: in1
    integer, intent(in) :: in2
    integer :: res
    integer :: dim,lb,ub
    if(in2<=in1(1)) then
       res=1
       return
    end if
    dim=size(in1)
    if(in2>in1(dim)) then
       res=0
       return
    end if
    lb=1
    ub=dim
    do while(ub-lb>1)
       res=(ub+lb)/2
       if(in1(res)<in2) lb=res
       if(in1(res)>=in2) ub=res
    end do
    res=ub
    if(in1(res-1)>=in2 .or. in1(res)<in2) stop 'find failed.'
  end function find_i

  function fdensity_onew(mythis,k,ll,t)
    type(spell), pointer, intent(in) :: mythis
    integer, intent(in) :: k, ll, t
    integer :: typ
    real(8) :: w
    real(8) :: fdensity_onew
    real(8) :: pi = 3.1415926535897932d0
    fdensity_onew = 0.d0     ! Initialize density
    if(ll==0) return
    w = mythis%wage(t)
    typ=hetmap(mythis,t)
    fdensity_onew = -log(ep%sigmaw(k,ll,typ))-0.5d0*(log(2.d0)+log(pi))-0.5d0*(((w-ep%muw(k,ll,typ))/ep%sigmaw(k,ll,typ))**2)
    if(fdensity_onew<-huge(fdensity_onew)) then
       write(6,'(2i10,2f15.5,4i10)') 1,mythis%wid,ep%muw(k,ll,typ),w,typ,k,ll,t
       fdensity_onew=-huge(fdensity_onew)
    end if
    if(isnan(fdensity_onew)) then
       print *,2,mythis%wid,w,typ,k,ll,t
       fdensity_onew=-huge(fdensity_onew)
    end if
  end function fdensity_onew

  function fdensity_i4(mythis,k,ll) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4), intent(in) :: k, ll
    real(8) :: res
    integer :: t
    res = 0.d0  ! Initialize density
    if(ll==0) return
    do t = 1, size(mythis%wage)   !do this for all periods in a given spell of each match
       res = res + fdensity_onew(mythis,k,ll,t)
    end do
    if(res<-huge(res)) res=-huge(res)
    if(isnan(res)) res=-huge(res)
  end function fdensity_i4

  function fdensity_i1(mythis,k,ll) result(res)
    type(spell), pointer, intent(in) :: mythis
    integer(4), intent(in) :: k
    integer(1), intent(in) :: ll
    real(8) :: res
    res=fdensity_i4(mythis,k,int(ll))
  end function fdensity_i1

  function fdensity_mat(mythis)
    type(spell), pointer, intent(in) :: mythis
    real(8) :: fdensity_mat(ep%k,ep%l)
    integer :: ll, k

    fdensity_mat = 0.d0 ! Initialize density
    do k = 1, ep%k
       do ll = 1, ep%l
          fdensity_mat(k,ll) = fdensity(mythis,k,ll)
       end do
    end do
  end function fdensity_mat
end module estim
