module postest
  use modelpar_class
  use tools
  use datbysim
  use linmod, only : matinv_lapack
  implicit none
  !folders
  character(*), parameter :: figfold='out/fig/'
  character(*), parameter :: tabfold='out/tab/'
  character(*), parameter :: outfold='out/'
  character(*), parameter :: estfold='estim/'
  
  type txcdf !by (x and time)
     integer :: ldim,kdim
     real(8), allocatable :: qkcdf(:,:)
     real(8), allocatable :: qcdf(:)
     real(8), allocatable :: lrank(:)
     real(8), allocatable :: eval(:,:)
  end type txcdf

  type tcdf !by time
     integer :: ldim,kdim
     real(8), allocatable :: krank(:),lrank(:),zcdf(:,:),expcdf(:,:),tencdf(:,:,:),pkcdf(:),plcdf(:)
     real(8) :: pz(6),pexp(4),pten(2,2)
     type(fullest), pointer :: est
   contains
     procedure :: zeval => zeval
     procedure :: expeval => expeval
     procedure :: eteneval => eteneval
     procedure :: uteneval => uteneval
     procedure :: muwev => muweval
  end type tcdf

  type fullest
     type(modelpar) :: par !holds model parameters
     real(8), allocatable :: p(:,:,:),ak(:),bl(:),het(:) !p(k,l,x) and group wage labels.
     real(8), allocatable :: varwx(:) !empirical x-conditional wage variances.
     real(8), allocatable :: npv(:,:,:) !NPV measures.
     real(8), allocatable :: wfe(:,:),ffe(:,:) !collection of fixed effects
     real(8) :: c !constant in linear wage projection.
   contains
     procedure :: pp => pp !map to par%dim%het x categorization.
     procedure :: pe => pe !employment conditional p. myx in 8 categorization
     procedure :: pklx_emp => pklx_emp !employment conditional p. myx in 8 categorization. Returns full matrix.
     procedure :: pel => pel !employment conditional l marginal
     procedure :: pek => pek !employment conditional k marginal
     procedure :: krank => krank !employment conditional k rank (has optional conditioning on (l,x))
     procedure :: lrank => lrank !employment conditional l rank.
  end type fullest
  type(fullest), target :: est(5) !Contains LPR estimates for all 5 periods.
contains
  subroutine postwork
    use datbysim
    character(1) :: ai
    integer :: i,t,k,l,x
    real(8) :: denom
    do i=1,5 !Read in LPR model estimates.
       write(ai,'(i1)') i
       open(12,file=estfold//'time'//ai//'est.txt',status="old")
       call readest(12,est(i))
       close(12)
       call linproj_common(est(i)%par%muw,est(i)%pklx_emp(),est(i)%ak,est(i)%bl,est(i)%c,est(i)%het) !fill in wage effects.
    end do

    !Either read in past NPV calculations or do new ones
    call NPVcalc(.false.) 

    !Table and figure production
    if(master) then
       call tab_px
       call fig1
       call fig2
       call fig3
       call fig_sort_time
       call table_vardecomp
       call vardecomp
       call tab_parcorr
       call tab_prefint
    end if

    !Produce cohort simulations and counterfactuals.
    call cohortsim_sort
  end subroutine postwork
  
  function emp(this) result(res)
    class(fullest), intent(in) :: this
    real(8) :: res
    res=1.d0-sum(sum(this%p(:,0,:),dim=1))
  end function emp

  function krank(this,myk,myl,myx) result(res)
    class(fullest), intent(in) :: this
    integer, intent(in) :: myk
    integer, intent(in), optional :: myl,myx
    real(8) :: res,denom,x
    integer :: k
    if(.not. present(myl)) then
       res=0.d0
       do k=1,myk
          res=res+pek(this,k)
       end do
    else
       res=0.d0
       denom=0.d0
       do k=1,this%par%dim%k
          x=pp(this,k,myl,myx)
          if(k .le. myk) res=res+x
          denom=denom+x
       end do
       res=res/denom
    end if
  end function krank

  function lrank(this,myl) result(res)
    class(fullest), intent(in) :: this
    integer, intent(in) :: myl
    real(8) :: res
    integer :: l
    res=0.d0
    do l=1,myl
       res=res+pel(this,l)
    end do
  end function lrank

  function pe(this,myk,myl,myx) result(res) !employment conditional p. myx is in 8 categorization.
    class(fullest), intent(in) :: this
    integer, intent(in) :: myk,myl,myx
    real(8) :: res
    res=pp(this,myk,myl,myx)/emp(this)
  end function pe

  function pklx_emp(this) result(res)
    class(fullest), intent(in) :: this
    real(8) :: res(this%par%dim%k,this%par%dim%l,this%par%dim%het)
    integer :: k,l,x
    do x=1,this%par%dim%het
       do l=1,this%par%dim%l
          do k=1,this%par%dim%k
             res(k,l,x)=pe(this,k,l,x)
          end do
       end do
    end do
  end function pklx_emp

  function pel(this,ndx,myx) result(res) !employment conditional l-marginal. Optional myx is in 8 categorization.
    class(fullest), intent(in) :: this
    integer, intent(in) :: ndx
    integer, intent(in), optional :: myx
    integer :: k,l
    real(8) :: res
    real(8) :: con
    con=0.d0
    res=0.d0
    if(present(myx)) then
       do k=1,this%par%dim%k
          res=res+pp(this,k,ndx,myx)
          do l=1,this%par%dim%l
             con=con+pp(this,k,l,myx)
          end do
       end do
       res=res/con
    else
       con=1.d0-sum(sum(this%p(:,0,:),dim=2))
       res=sum(sum(this%p(:,ndx,:),dim=2))/con
    end if
  end function pel

  function pek(this,ndx,myx) result(res) !employment conditional k-marginal. Optional myx is in 8 categorization.
    class(fullest), intent(in) :: this
    integer, intent(in) :: ndx
    integer, intent(in), optional :: myx
    integer :: k,l
    real(8) :: res
    real(8) :: con
    con=0.d0
    res=0.d0
    if(present(myx)) then
       do l=1,this%par%dim%l
          res=res+pp(this,ndx,l,myx)
          do k=1,this%par%dim%k
             con=con+pp(this,k,l,myx)
          end do
       end do
       res=res/con
    else
       con=1.d0-sum(sum(this%p(:,0,:),dim=2))
       res=sum(sum(this%p(ndx,1:this%par%dim%L,:),dim=2))/con
    end if
  end function pek

  function pp(this,myk,myl,myx) result(res) !p(k,l,x) for x in 8 categorization.
    class(fullest), intent(in) :: this
    integer, intent(in) :: myk,myl,myx
    real(8) :: res
    if(myx .gt. 8) stop 'myx outside of range.'
    if(myx .le. 4) then
       res=this%p(myk,myl,myx)+this%p(myk,myl,myx+4)
    else
       res=this%p(myk,myl,myx+4)
    end if
  end function pp

  pure function linint(x,y,myev) result(res)
    real(8), dimension(*), intent(in) :: x(:),y(:)
    real(8), intent(in) :: myev
    real(8) :: res
    real(8) :: lam
    integer :: i,dim
    dim=size(x)
    if(myev .le. x(1)) then
       res=y(1)
       return
    end if
    if(myev .ge. x(dim)) then
       res=y(dim)
       return
    end if
    i=1
    do while(x(i) .lt. myev)
       i=i+1
    end do
    lam=(myev-x(i-1))/(x(i)-x(i-1))
    res=(1.d0-lam)*y(i-1)+lam*y(i)
  end function linint

  function muweval(this,ndx,krev,lrev) result(res)
    class(tcdf), intent(in) :: this
    integer, intent(in) :: ndx
    real(8), intent(in) :: krev,lrev
    real(8) :: res,lint(2),lamk,laml,musq(2,2)
    integer :: kn(2),ln(2),k,l,i
    if(krev .le. this%krank(1)) then
       kn=(/1,2/)
    else
       i=1
       do while(this%krank(i) .lt. krev)
          i=i+1
          if(i .eq. this%kdim) exit
       end do
       kn=(/i-1,i/)
    end if
    lamk=(krev-this%krank(kn(1)))/(this%krank(kn(2))-this%krank(kn(1)))
    lamk=min(1.d0,max(0.d0,lamk)) !constrain between zero and one.
    if(lrev .le. this%lrank(1)) then
       ln=(/1,2/)
    else
       i=1
       do while(this%lrank(i) .lt. lrev)
          i=i+1
          if(i .eq. this%ldim) exit
       end do
       ln=(/i-1,i/)
    end if
    laml=(lrev-this%lrank(ln(1)))/(this%lrank(ln(2))-this%lrank(ln(1)))
    laml=min(1.d0,max(0.d0,laml)) !constrain between zero and one.
    do k=1,2
       do l=1,2
          musq(k,l)=this%est%par%muw(kn(k),ln(l),ndx)
       end do
    end do
    res=(1.d0-laml)*((1.d0-lamk)*musq(1,1)+lamk*musq(2,1))+laml*((1.d0-lamk)*musq(1,2)+lamk*musq(2,2))
  end function muweval

  function muwsmooth(ndx,ev,bndwdth,ck) result(res)
    use sort3
    integer, intent(in) :: ndx,ck
    real(8), intent(in) :: ev(2),bndwdth
    real(8) :: res,wght,sumwght
    integer :: t,k,l,i
    integer :: dim
    integer, allocatable :: order(:), list(:,:)
    real(8), allocatable :: val(:),ordval(:)
    !first determine the k nearest neighbors
    dim=0
    do t=1,5
       dim=dim+est(t)%par%dim%k*est(t)%par%dim%l
    end do
    allocate(order(dim),list(dim,3),val(dim),ordval(dim))
    i=0
    do t=1,5
       do k=1,est(t)%par%dim%k
          do l=1,est(t)%par%dim%l
             i=i+1
             list(i,:)=(/ t, k, l /)
             order(i)=i
             val(i)=-gausskrnl(ev,(/ est(t)%ak(k), est(t)%bl(l) /),1.d0)
          end do
       end do
    end do
    ordval=val
    call hpsort(dim,ordval,order)

    sumwght=0.d0
    res=0.d0
    do i=1,ck
       t=list(order(i),1)
       k=list(order(i),2)
       l=list(order(i),3)
       wght=gausskrnl(ev,(/ est(t)%ak(k), est(t)%bl(l) /),bndwdth)
       sumwght=sumwght+wght
       res=res+wght*(est(t)%par%muw(k,l,ndx)-est(t)%ak(k)-est(t)%bl(l)-est(t)%het(ndx)-est(t)%c) !residual
    end do
    res=res/sumwght
  end function muwsmooth

  function quantsmooth(ndx,ev,bndwdth,ck) result(res)
    use sort3
    integer, intent(in) :: ndx,ck
    real(8), intent(in) :: ev(2),bndwdth
    real(8) :: res,wght,sumwght
    integer :: t,k,l,i
    integer :: dim
    integer, allocatable :: order(:), list(:,:)
    real(8), allocatable :: val(:),ordval(:)
    !first determine the k nearest neighbors
    dim=0
    do t=1,5
       dim=dim+est(t)%par%dim%k*est(t)%par%dim%l
    end do
    allocate(order(dim),list(dim,3),val(dim),ordval(dim))
    i=0
    do t=1,5
       do k=1,est(t)%par%dim%k
          do l=1,est(t)%par%dim%l
             i=i+1
             list(i,:)=(/ t, k, l /)
             order(i)=i
             val(i)=-gausskrnl(ev,(/ est(t)%lrank(l), est(t)%krank(k) /),1.d0)
          end do
       end do
    end do
    ordval=val
    call hpsort(dim,ordval,order)

    sumwght=0.d0
    res=0.d0
    do i=1,ck
       t=list(order(i),1)
       k=list(order(i),2)
       l=list(order(i),3)
       wght=gausskrnl(ev,(/ est(t)%lrank(l), est(t)%krank(k) /),bndwdth)
       sumwght=sumwght+wght
       res=res+wght*est(t)%krank(k,l,ndx)
    end do
    res=res/sumwght
  end function quantsmooth

  function gausskrnl(a,b,bndwdth) result(res)
    real(8), dimension(:), intent(in) :: a,b
    real(8), intent(in) :: bndwdth
    real(8) :: res
    integer :: i
    res=0.d0
    do i=1,size(a)
       res=res+0.5d0*(((a(i)-b(i))/bndwdth)**2)
    end do
    res=((2.d0*3.141592653589793d0)**(-dble(size(a))/2.d0))*exp(-res)
  end function gausskrnl

  function zeval(this,ndx,myev) result(res)
    class(tcdf), intent(in) :: this
    integer, intent(in) :: ndx
    real(8), intent(in) :: myev
    real(8) :: res
    res=linint(this%krank,this%zcdf(:,ndx),myev)
  end function zeval

  function expeval(this,ndx,myev) result(res)
    class(tcdf), intent(in) :: this
    integer, intent(in) :: ndx
    real(8), intent(in) :: myev
    real(8) :: res
    res=linint(this%krank,this%expcdf(:,ndx),myev)
  end function expeval

  function eteneval(this,ndx,myev) result(res)
    class(tcdf), intent(in) :: this
    integer, intent(in) :: ndx
    real(8), intent(in) :: myev
    real(8) :: res
    res=linint(this%krank,this%tencdf(:,ndx,2),myev)
  end function eteneval

  function uteneval(this,ndx,myev) result(res)
    class(tcdf), intent(in) :: this
    integer, intent(in) :: ndx
    real(8), intent(in) :: myev
    real(8) :: res
    res=linint(this%krank,this%tencdf(:,ndx,1),myev)
  end function uteneval

  subroutine tab_px
    real(8) :: px(8,5)
    integer :: t,x
    character(10) :: yrlabel(5)
    yrlabel=(/ "1989-93", "1994-98", "1999-03", "2004-08", "2009-13" /)
    do t=1,5
       px(:,t)=sum(sum(est(t)%pklx_emp(),dim=1),dim=1)       
    end do
    
    open(10,file=tabfold//'tab_px.tex',status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')

    write(10,'("\begin{tabular}{c| *{4}{d{2.2}} | *{4}{d{2.2}}} ")')
    write(10,'("\hline")')
    write(10,'(" & \multicolumn{4}{c|}{Short tenure} & \multicolumn{4}{c}{Long Tenure}\tabularnewline")')
    write(10,'("Experience (yrs): & \mc{0-5} & \mc{5-10} & \mc{10-15} & ",&
         "\multicolumn{1}{c|}{15+} & \mc{0-5} & \mc{5-10} & \mc{10-15} & \mc{15+}\tabularnewline")')
    write(10,'("\hline")')
    do t=1,5
       write(10,'(a10,8(" & ",f6.2),"\tabularnewline")') yrlabel(t),(px(x,t),x=1,8)
    end do
    write(10,'("\hline")')
    write(10,'("\end{tabular}")')
    write(10,'("\end{document}")')
    close(10)
  end subroutine tab_px
  
  subroutine fig_sort_time
    !takes as given that there are 8 x categories
    real(8), allocatable :: pkl_con(:,:,:),pklx(:,:,:)
    real(8) :: wsort(5,8),misort(5,8),px(8)
    integer :: t,k,l,x,yr
    do t=1,5
       allocate(pkl_con(est(t)%par%dim%k,est(t)%par%dim%l,8),pklx(est(t)%par%dim%k,est(t)%par%dim%l,8))
       pklx=est(t)%pklx_emp() !returns the employment conditional joint pdf
       px=sum(sum(pklx,dim=1),dim=1)
       do x=1,8
          pkl_con(:,:,x)=pklx(:,:,x)/px(x)
          wsort(t,x)=corr(est(t)%ak,est(t)%bl,pkl_con(:,:,x))
          misort(t,x)=mi(pkl_con(:,:,x))
       end do
       deallocate(pkl_con,pklx)
    end do
    open(10,file=figfold//"wsort_time.dat",status='replace')
    write(10,'(9a10)') "year","one","two","three","four","five","six","seven","eight"
    do t=1,5
       do yr=1989+(t-1)*5,1989+t*5
          write(10,'(i10,8f10.6)') yr,(wsort(t,x),x=1,8)
       end do
    end do
    close(10)
    open(10,file=figfold//"misort_time.dat",status='replace')
    write(10,'(9a10)') "year","one","two","three","four","five","six","seven","eight"
    do t=1,5
       do yr=1989+(t-1)*5,1989+t*5
          write(10,'(i10,8f10.6)') yr,(misort(t,x),x=1,8)
       end do
    end do
    close(10)
  contains
    function corr(a,b,pkl) result(res)
      real(8), dimension(:,:), intent(in) :: pkl
      real(8), intent(in) :: a(size(pkl,dim=1)),b(size(pkl,dim=2))
      real(8) :: res
      integer :: k,l
      real(8) :: ea,eaa,eb,ebb,eab
      ea=0.d0
      eaa=0.d0
      eb=0.d0
      ebb=0.d0
      eab=0.d0
      do k=1,size(a)
         do l=1,size(b)
            ea=ea+a(k)*pkl(k,l)
            eaa=eaa+a(k)*a(k)*pkl(k,l)
            eb=eb+b(l)*pkl(k,l)
            ebb=ebb+b(l)*b(l)*pkl(k,l)
            eab=eab+a(k)*b(l)*pkl(k,l)
         end do
      end do
      res=(eab-ea*eb)/sqrt((eaa-ea*ea) * (ebb-eb*eb))
    end function corr

    function mi(pkl) result(res)
      real(8), dimension(:,:), intent(in) :: pkl
      real(8) :: res
      real(16) :: pk(size(pkl,dim=1)),pl(size(pkl,dim=2)),enk,enl
      real(16) :: acc
      integer :: k,l
      acc=0.d0
      pk=sum(pkl,dim=2)
      pl=sum(pkl,dim=1)
      enk=0.d0
      enl=0.d0
      do k=1,size(pk)
         enk=enk-pk(k)*log(pk(k))
      end do
      do l=1,size(pl)
         enl=enl-pl(l)*log(pl(l))
      end do
      
      do k=1,size(pk)
         do l=1,size(pl)
            if(log(pkl(k,l))<0.d0) then
               acc=acc+pkl(k,l)*(max(-huge(pkl(k,l)),log(pkl(k,l)))-log(pk(k))-log(pl(l))) !ensure that -infinity is evaluated as -huge.
            end if
         end do
      end do
      res=acc/min(enk,enl)
    end function mi
  end subroutine fig_sort_time

  subroutine vardecomp
    real(8) :: dec(6,10,5)
    integer :: t,ii,jj
    real(8), allocatable :: afix(:),bfix(:),xfix(:)
    real(8) :: con
    real(8) :: cor_v(6)
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%par%muw,est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,1)=afix
       est(t)%ffe(:,1)=bfix
       call vardec(est(t)%par%muw,est(t)%pklx_emp(),afix,bfix,xfix,con,dec(1,:,t))
       deallocate(afix,bfix,xfix)
    end do
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%par%del,est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,2)=afix
       est(t)%ffe(:,2)=bfix
       call vardec(est(t)%par%del,est(t)%pklx_emp(),afix,bfix,xfix,con,dec(2,:,t))
       deallocate(afix,bfix,xfix)
    end do
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%npv(:,1:est(t)%par%dim%l,:),est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,6)=afix
       est(t)%ffe(:,6)=bfix
       call vardec(est(t)%npv(:,1:est(t)%par%dim%l,:),est(t)%pklx_emp(),afix,bfix,xfix,con,dec(6,:,t))
       deallocate(afix,bfix,xfix)
    end do
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%par%lam,est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,3)=afix
       est(t)%ffe(:,3)=bfix
       call vardec(est(t)%par%lam,est(t)%pklx_emp(),afix,bfix,xfix,con,dec(3,:,t))
       deallocate(afix,bfix,xfix)
    end do
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%par%psi,est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,4)=afix
       est(t)%ffe(:,4)=bfix
       call vardec(est(t)%par%psi,est(t)%pklx_emp(),afix,bfix,xfix,con,dec(4,:,t))
       deallocate(afix,bfix,xfix)
    end do
    do t=1,5
       allocate(afix(est(t)%par%dim%k),bfix(est(t)%par%dim%l),xfix(est(t)%par%dim%het))
       call linproj_common(est(t)%par%gam,est(t)%pklx_emp(),afix,bfix,con,xfix)
       est(t)%wfe(:,5)=afix
       est(t)%ffe(:,5)=bfix
       call vardec(est(t)%par%gam,est(t)%pklx_emp(),afix,bfix,xfix,con,dec(5,:,t))
       deallocate(afix,bfix,xfix)
    end do

    !---- A quick study of the correlations in type characteristics -----
    write(6,'("Worker type correlations")')
    write(6,'(6a8)') "mu","del","lam","psi","gam","NPV"
    do ii=1,6,1
       do jj=1,ii-1,1
          write(6,'(a8,$)') " "
       end do
       write(6,'(f8.3,$)') 1.d0
       cor_v=0.d0
       do jj=ii+1,6,1
          do t=1,5
             cor_v(jj)=cor_v(jj)+corr(est(t)%wfe(:,ii),est(t)%wfe(:,jj))/dble(5) 
          end do
       end do
       if(ii<6) write(6,'(*(f8.3))') (cor_v(jj),jj=ii+1,6)
    end do
    write(6,'(" ")')
    write(6,'("Firm type correlations")')
    write(6,'(6a8)') "mu","del","lam","psi","gam","NPV"
    do ii=1,6,1
       do jj=1,ii-1,1
          write(6,'(a8,$)') " "
       end do
       write(6,'(f8.3,$)') 1.d0
       cor_v=0.d0
       do jj=ii+1,6,1
          do t=1,5
             cor_v(jj)=cor_v(jj)+corr(est(t)%ffe(:,ii),est(t)%ffe(:,jj))/dble(5)
          end do
       end do
       if(ii<6) write(6,'(*(f8.3))') (cor_v(jj),jj=ii+1,6)
    end do
    write(6,'(" ")')

    open(10,file=tabfold//'vardecom_all.tex',status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    call write_table_all
    write(10,'("\end{document}")')
    close(10)
    
    open(10,file=tabfold//'vardecom_muw.tex',status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    call write_table(1)
    write(10,'("\end{document}")')
    close(10)

    open(10,file=tabfold//'vardecom_del.tex',status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    call write_table(2)
    write(10,'("\end{document}")')
    close(10)
    
    open(10,file=tabfold//'vardecom_npv.tex',status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    call write_table(3)
    write(10,'("\end{document}")')
    close(10)    
  contains
    subroutine write_table(ndx)
      integer, intent(in) :: ndx
       write(10,'("\begin{tabular}{l *{5}{d{3.2}} }")')
       write(10,'("\hline")')
       write(10,'(" & \mc{$89-93$} & \mc{$94-98$} & \mc{$99-03$} & \mc{$04-08$} & \mc{$09-13$}\tabularnewline")')
       write(10,'("\hline")')
       write(10,'("Coefficient of variation",5(" & ",f6.2),"\tabularnewline")') dec(ndx,1,:)
       write(10,'("Variance decomposition (pct) & & & & & \tabularnewline")')
       write(10,'("\hspace{.3cm}worker effect, $a_{k}$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,2,:)
       write(10,'("\hspace{.3cm}match effect, $\widetilde{y}_{k\ell}(x)$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,3,:)
       write(10,'("\hspace{.3cm}firm effect, $b_{\ell}$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,4,:)
       write(10,'("\hspace{.3cm}sorting effect, $2\Cov(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,5,:)
       write(10,'("\hspace{.3cm}tenure and experience, $x$",5(" & ",f6.2),"\tabularnewline")') (sum(dec(ndx,6:8,t)),t=1,5)
       write(10,'("\hspace{.6cm}$\overline{y}(x)$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,6,:)
       write(10,'("\hspace{.6cm}$2\Cov(a_{k},\overline{y}(x))$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,7,:)
       write(10,'("\hspace{.6cm}$2\Cov(b_{\ell},\overline{y}(x))$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,8,:)
       write(10,'("\hline")')
       write(10,'("$\Cor(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') dec(ndx,10,:)
       write(10,'("\hline")')
       write(10,'("\end{tabular}")')
    end subroutine write_table

    subroutine write_table_all
      integer :: ndx
       write(10,'("\begin{tabular}{l *{5}{d{3.2}} }")')
       write(10,'("\hline")')
       write(10,'(" & \mc{$\mu_{k\ell}(x)$} & \mc{$\delta_{k\ell}(x)$} & \mc{$\lambda_{k\ell}(x)$}"//&
            &" & \mc{$\psi_{k\ell}(x)$} & \mc{$\gamma_{k\ell}(x)$} \tabularnewline")')
       write(10,'("\hline")')
       write(10,'("Coefficient of variation",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,1,:))/dble(5),ndx=1,5)
       write(10,'("Variance decomposition (\%) & & & & & \tabularnewline")')
       write(10,'("\hspace{.3cm}worker effect, $a^y_{k}$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,2,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.3cm}match effect, $\widetilde{y}_{k\ell}(x)$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,3,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.3cm}firm effect, $b^y_{\ell}$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,4,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.3cm}sorting effect, $2\Cov(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,5,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.3cm}tenure and experience, $x$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(sum(dec(ndx,6:8,:),dim=1))/dble(5),ndx=1,5)
       write(10,'("\hspace{.6cm}$\overline{y}(x)$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,6,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.6cm}$2\Cov(a_{k},\overline{y}(x))$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,7,:))/dble(5),ndx=1,5)
       write(10,'("\hspace{.6cm}$2\Cov(b_{\ell},\overline{y}(x))$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,8,:))/dble(5),ndx=1,5)
       write(10,'("\hline")')
       write(10,'("$\Cor(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') &
            & (sum(dec(ndx,10,:))/dble(5),ndx=1,5)
       write(10,'("\hline")')
       write(10,'("\end{tabular}")')
    end subroutine write_table_all
    
    subroutine vardec(mypar,pklx,afix,bfix,xfix,con,decom) !close copy of uncon_vardecom subroutine below.
      implicit none
      real(8), intent(out) :: decom(10)
      real(8), dimension(:,:,:), intent(in) :: mypar,pklx
      real(8), dimension(:), intent(in) :: afix,bfix,xfix
      real(8), intent(in) :: con
      real(8), allocatable :: res(:,:,:)
      real(8) :: vw, emuw2, emuw, emuwt2, emuwt, ea, eb, ehet,ea2, eb2,ehet2, eab, eahet, ebhet
      integer :: k,l,i,x
      integer :: ck,cl,cx
      ck=size(mypar,dim=1)
      cl=size(mypar,dim=2)
      cx=size(mypar,dim=3)
      allocate(res(ck,cl,cx))
      
      decom=0.d0
      
      do x=1,cx
         do l=1,cl
            do k=1,ck
               res(k,l,x)=mypar(k,l,x)-xfix(x)-con-afix(k)-bfix(l)
            end do
         end do
      end do
      
      vw=0.d0
      emuw2=0.d0
      emuw=0.d0
      emuwt2=0.d0
      emuwt=0.d0
      ea=0.d0
      eb=0.d0
      ehet=0.d0
      ea2=0.d0
      eb2=0.d0
      eb=0.d0
      ehet2=0.d0
      eab=0.d0
      eahet=0.d0
      ebhet=0.d0
      do x=1,cx
         do l=1,cl
            do k=1,ck
               emuw2=emuw2+pklx(k,l,x)*mypar(k,l,x)**2
               emuw=emuw+pklx(k,l,x)*mypar(k,l,x)
               emuwt2=emuwt2+pklx(k,l,x)*res(k,l,x)**2
               emuwt=emuwt+pklx(k,l,x)*res(k,l,x)
               ea=ea+pklx(k,l,x)*afix(k)
               eb=eb+pklx(k,l,x)*bfix(l)
               ehet=ehet+pklx(k,l,x)*xfix(x)
               ea2=ea2+pklx(k,l,x)*afix(k)**2
               eb2=eb2+pklx(k,l,x)*bfix(l)**2
               ehet2=ehet2+pklx(k,l,x)*xfix(x)**2
               eab=eab+pklx(k,l,x)*afix(k)*bfix(l)
               eahet=eahet+pklx(k,l,x)*afix(k)*xfix(x)
               ebhet=ebhet+pklx(k,l,x)*bfix(l)*xfix(x)
            end do
         end do
      end do

      vw=emuw2-(emuw)**2 !total variance of mypar
      decom(1)= sqrt(vw)/abs(emuw)  !CV
      decom(2)= (ea2-(ea)**2)/vw*100   !var a
      decom(3)= (emuwt2-(emuwt)**2)/vw*100 !muw tilde	
      decom(4)= (eb2-(eb)**2)/vw*100   !var b
      decom(5)= 2*(eab-ea*eb)/vw*100  !2*cov (a,b)
      decom(6)= (ehet2-(Ehet)**2)/vw*100   !var het or muw bar
      decom(7)= 2*(eahet-ea*ehet)/vw*100  !2*cov (a,het)
      decom(8)= 2*(ebhet-eb*ehet)/vw*100  !2*cov (b,het)
      decom(9)= (emuw2-(emuw)**2)/vw*100  !total between or sum of decom(2:8)
      decom(10) = (eab-ea*eb)/((ea2-(ea)**2) * (eb2-(eb)**2))**0.5 !corr(a,b)
    end subroutine vardec
  end subroutine vardecomp
  
  subroutine NPVcalc(getold)
    logical, intent(in) :: getold
    integer :: t,k,l,x
    real(8) :: avg(8)
    if(getold) then
       do t=1,5
          allocate(est(t)%npv(est(t)%par%dim%k,0:est(t)%par%dim%l,est(t)%par%dim%het))
          open(11,file=outfold//'npv'//inttochar(1,t)//'.txt',status='old')
          do x=1,est(t)%par%dim%het
             do k=1,est(t)%par%dim%k
                read(11,'(*(f12.6))') (est(t)%npv(k,l,x),l=0,est(t)%par%dim%l)
             end do
          end do
          close(11)
       end do
    else
       do t=1,5
          call modelpar_deallocate(par)
          par=new_modelpar_new(est(t)%par%dim)
          par=est(t)%par
          allocate(est(t)%npv(par%dim%k,0:par%dim%l,par%dim%het))
          call simNPV(est(t)%npv)
          if(master) then
             open(11,file=outfold//'npv'//inttochar(1,t)//'.txt',status='replace')
             do x=1,est(t)%par%dim%het
                do k=1,est(t)%par%dim%k
                   write(11,'(*(f12.6))') (est(t)%npv(k,l,x),l=0,est(t)%par%dim%l)
                end do
             end do
             close(11)
          end if
       end do
    end if
  end subroutine NPVcalc

  subroutine tab_prefint
    character(10) :: yrlabel(5)
    integer :: t
    yrlabel=(/ "1989-93", "1994-98", "1999-03", "2004-08", "2009-13" /)
    !this routine assumes 24 worker types in every period and that types are already ordered.
    open(10,file=tabfold//"tab_prefint.tex",status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    write(10,'("\begin{tabular}{ll| *{4}{d{1.2}} | *{4}{d{1.2}}} ")')
    write(10,'("\hline")')
    write(10,'(" & & \multicolumn{4}{c|}{Short tenure} & \multicolumn{4}{c}{Long Tenure}\tabularnewline")')
    write(10,'("\multicolumn{2}{c|}{Experience (yrs):} & \mc{0-5} & \mc{5-10} & \mc{10-15} & ",&
         "\multicolumn{1}{c|}{15+} & \mc{0-5} & \mc{5-10} & \mc{10-15} & \mc{15+}\tabularnewline")')
    write(10,'("\hline")')
    do t=1,5
       if(t==1) then
          write(10,'(" & Low",8(" & ",f5.2),"\tabularnewline")') avgkl_div(est(t)%par%gam(1:8,:,:))
       else
          write(10,'("\rule{0pt}{1.4em} & Low",8(" & ",f5.2),"\tabularnewline")') avgkl_div(est(t)%par%gam(1:8,:,:))
       end if
       write(10,'(a7," & Med",8(" & ",f5.2),"\tabularnewline")') yrlabel(t),avgkl_div(est(t)%par%gam(9:16,:,:))
       write(10,'(" & High",8(" & ",f5.2),"\tabularnewline")') avgkl_div(est(t)%par%gam(17:24,:,:))
    end do
    write(10,'("\hline")')
    write(10,'("\end{tabular}")')
    write(10,'("\end{document}")')
    close(10)
  contains
    function avgkl_div(mygam) result(res)
      real(8), dimension(:,:,:), intent(in) :: mygam
      real(8) :: res(size(mygam,dim=3))
      integer :: k,l,x
      res=0.d0
      do x=1,size(mygam,dim=3)
         do k=1,size(mygam,dim=1)
            res(x)=res(x)+dkl(mygam(k,:,x))/dble(size(mygam,dim=1))
         end do
      end do
    end function avgkl_div

    function dkl(x) result(res)
      real(8), intent(in) :: x(:)
      real(8) :: res
       res=log(dble(size(x)))+sum(x*log(x)) !provides kl divergence from uniformity
!!$       res=(res-log(dble(size(x)))-sum(log(x))/dble(size(x)))/dble(2) !activate this to make it symmetric (jeffrey's divergence)
     end function dkl
  end subroutine tab_prefint

  subroutine tab_parcorr
    real(8) :: avg(8)
    integer :: t
    open(10,file=tabfold//"tab_parcorr.tex",status='replace')
    write(10,'("\documentclass{standalone}")')
    write(10,'("\usepackage{basics}")')
    write(10,'("\begin{document}")')
    write(10,'("\centering")')
    write(10,'("\begin{tabular}{c| *{4}{d{2.2}} | *{4}{d{2.2}}} ")')
    write(10,'("\hline")')
    write(10,'(" & \multicolumn{4}{c|}{Short tenure} & \multicolumn{4}{c}{Long Tenure}\tabularnewline")')
    write(10,'("Experience (yrs): & \mc{0-5} & \mc{5-10} & \mc{10-15} & ",&
         "\multicolumn{1}{c|}{15+} & \mc{0-5} & \mc{5-10} & \mc{10-15} & \mc{15+}\tabularnewline")')
    write(10,'("\hline")')
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%npv(:,1:est(t)%par%dim%l,:),log(est(t)%par%gam))/dble(5)
    end do
    write(10,'("$\ln\gamma_{k\ell}(x),\text{NPV}_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%muw,est(t)%par%gam)/dble(5)
    end do
    write(10,'("\rule{0pt}{1.3em}$\mu_{k\ell}(x),\gamma_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%muw,est(t)%par%del)/dble(5)
    end do
    write(10,'("$\mu_{k\ell}(x),\delta_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%muw,est(t)%par%lam)/dble(5)
    end do
    write(10,'("$\mu_{k\ell}(x),\lambda_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%muw,est(t)%par%psi)/dble(5)
    end do
    write(10,'("$\mu_{k\ell}(x),\psi_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg

    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%gam,est(t)%par%del)/dble(5)
    end do
    write(10,'("\rule{0pt}{1.3em}$\gamma_{k\ell}(x),\delta_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%gam,est(t)%par%lam)/dble(5)
    end do
    write(10,'("$\gamma_{k\ell}(x),\lambda_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%gam,est(t)%par%psi)/dble(5)
    end do
    write(10,'("$\gamma_{k\ell}(x),\psi_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg

    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%del,est(t)%par%lam)/dble(5)
    end do
    write(10,'("\rule{0pt}{1.3em}$\delta_{k\ell}(x),\lambda_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%del,est(t)%par%psi)/dble(5)
    end do
    write(10,'("$\delta_{k\ell}(x),\psi_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg

    avg=0.d0
    do t=1,5
       avg=avg+avgcorr(est(t)%par%lam,est(t)%par%psi)/dble(5)
    end do
    write(10,'("\rule{0pt}{1.3em}$\lambda_{k\ell}(x),\psi_{k\ell}(x)$",8(" & ",f5.2),"\tabularnewline")') avg
    
    write(10,'("\hline")')
    write(10,'("\end{tabular}")')
    write(10,'("\end{document}")')
    close(10)
  contains
    function avgcorr(one,two) result(res)
      real(8), intent(in) :: one(:,:,:),two(:,:,:)
      real(8) :: res(size(one,dim=3))
      integer :: k,x
      res=0.d0
      do k=1,size(one,dim=1)
         do x=1,size(one,dim=3)
            res(x)=res(x)+corr(one(k,:,x),two(k,:,x))
         end do
      end do
      res=res/dble(size(one,dim=1))
    end function avgcorr
  end subroutine tab_parcorr
      
  subroutine cohortsim_sort
    integer :: simrep
    integer, parameter :: totalrep=1
    integer, parameter :: datxdim=8
    integer, parameter :: datzdim=6
    integer, parameter :: sim_time=30
    integer, parameter :: period=5
    integer, parameter :: counterf=14
    type multch
       character(4) :: order
       real(8) :: contrib(4)
    end type multch
    type(multch) :: tot(2,24) !tot(1,:) is corr, tot(2,:) is MI
    real(8), allocatable :: aa(:),bb(:)
    real(8), allocatable :: pk(:), pl(:),pkl(:,:),mypk(:),mypl(:),mypkl(:,:)
    integer :: i,j,ii,jj,xi,ki,li,ti, ch, klvar,myii
    real(8) :: enpk, enpl,ea,eb,va,vb,corrab(sim_time,period,0:counterf), mutual(sim_time,period,0:counterf)
    real(8) :: myea,myeb,myva,myvb,myab(sim_time),ab(sim_time)
    type(spell), pointer :: mysp
    do klvar=1,3  !for each of the three variations
       !Initializa correlation and mutual information
       corrab=0.d0
       mutual=0.d0

       do ti=1,period
          do ch=0,counterf !set counterfactual loop here
             if(master) print*, 'counterfactual', ch, 'out of', counterf

             !--- this replaces setsimpar ---
             call modelpar_deallocate(par) !make sure par is not allocated prior to allocation. par is global from estim.f90
             par=new_modelpar_new(est(ti)%par%dim) 
             par=est(ti)%par

             if(allocated(aa)) deallocate(aa,bb,pk,pl,pkl,mypk,mypl,mypkl)
             allocate(aa(par%dim%k),bb(par%dim%l),pk(par%dim%k),pl(par%dim%l),pkl(par%dim%k,par%dim%l),&
                  & mypk(par%dim%k),mypl(par%dim%l),mypkl(par%dim%k,par%dim%l))
             !---
             call SortCF(ch,klvar)
             call simdat_ten_exp(223987+2*myid+1000,.false.,par%dim%k,par%dim%l) !produce simulation that is divided across CPU's if ser=.false.

             call pklxcalc(par,aa,bb) !this also understands parallel

             do j=1,sim_time
                myii=0
                myab=0.d0
                myea=0.d0
                myeb=0.d0
                myva=0.d0
                myvb=0.d0
                mypk=0.d0
                mypl=0.d0
                mypkl=0.d0
                enpk=0.d0
                enpl=0.d0
                do i=1,dim_I
                   if (.not. associated(work(i)%first)) cycle
                   call find_t(work(i)%first,1+1560*dble(j)/dble(30)) !this is smallest this%strt such that this%strt>1560*dble(j)/dble(30)
                   if(.not. associated(this%prev)) then !only happens if first spell is also only spell.
                      if(this%strt>1) stop 'this should have been a single continous spell'
                   else
                      if(this%prev%end>=1560*dble(j)/dble(30)) then !this is the norm. Happens as long we are not in last spell.
                         this=>this%prev
                      else
                         if(this%strt > 1560*dble(j)/dble(30)) then !check that this is indeed last spell.
                            print *,myid,i,j,this%strt,this%end,this%prev%strt,this%prev%end
                            stop 'something went wrong at end of history.'
                         end if
                      end if
                   end if
                   if(this%fid>0) then
                      myab(j)=myab(j)+aa(work(this%wid)%k)*bb(firm(this%fid)%l)
                      myea=myea+aa(work(this%wid)%k)
                      myva=myva+aa(work(this%wid)%k)*aa(work(this%wid)%k)
                      myeb=myeb+bb(firm(this%fid)%l)
                      myvb=myvb+bb(firm(this%fid)%l)*bb(firm(this%fid)%l)
                      mypk(work(this%wid)%k)=mypk(work(this%wid)%k)+1
                      mypl(firm(this%fid)%l)=mypl(firm(this%fid)%l)+1
                      mypkl(work(this%wid)%k,firm(this%fid)%l)=mypkl(work(this%wid)%k,firm(this%fid)%l)+1
                      myii=myii+1
                   end if
                end do  !end of worker loop
                call mpi_allreduce(myab,ab,sim_time,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(myea,ea,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(myva,va,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(myeb,eb,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(myvb,vb,1,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(mypk,pk,par%dim%k,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(mypl,pl,par%dim%l,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(mypkl,pkl,par%dim%k*par%dim%l,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
                call mpi_allreduce(myii,ii,1,mpi_integer,mpi_sum,mpi_comm_world,ierr)

                ea=ea/dble(ii)
                va=va/dble(ii)-ea*ea
                eb=eb/dble(ii)
                vb=vb/dble(ii)-eb*eb
                corrab(j,ti,ch)=(ab(j)/dble(ii)-ea*eb)/sqrt(va*vb)
                pk=pk/sum(pk)
                pl=pl/sum(pl)
                pkl=pkl/sum(sum(pkl,1))
                do ki=1,par%dim%k
                   do li=1,par%dim%l
                      if(pkl(ki,li)>0) mutual(j,ti,ch)=mutual(j,ti,ch)&
                           &+pkl(ki,li)*log(pkl(ki,li)/(pk(ki)*pl(li)))
                   end do
                end do
                do ki=1,par%dim%k
                   enpk=enpk+pk(ki)*log(pk(ki))
                end do
                do li=1,par%dim%l
                   enpl=enpl+pl(li)*log(pl(li))
                end do
                enpk=enpk*(-1.d0)
                enpl=enpl*(-1.d0)
                mutual(j,ti,ch)=mutual(j,ti,ch)/min(enpl,enpl)
             end do !end of sim_period t
             call cleansim
          end do !end of counterfactual ch
       end do !end of period loop
       if(master) then
          call fig_cohort_sort
          call fig_cohort_counter
          call sort_contrib
       end if
    end do ! end of variation
  contains
    subroutine fig_cohort_sort
      integer :: ii,jj
      open(10,file=figfold//'misort_cohort.dat',status='replace')
      write(10,'(7a10)') "x","one","two","three","four","five","avg"
      do ii=1,sim_time
         write(10,'(7f10.4)') dble(ii)*dble(1560)/(dble(52)*dble(sim_time)),&
              &(mutual(ii,jj,0),jj=1,5),sum(mutual(ii,:,0))/dble(5)
      end do
      close(10)
      open(10,file=figfold//'wsort_cohort.dat',status='replace')
      write(10,'(7a10)') "x","one","two","three","four","five","avg"
      do ii=1,sim_time
         write(10,'(7f10.4)') dble(ii)*dble(1560)/(dble(52)*dble(sim_time)),&
              &(corrab(ii,jj,0),jj=1,5),sum(corrab(ii,:,0))/dble(5)
      end do
      close(10)
    end subroutine fig_cohort_sort

    subroutine fig_cohort_counter
      integer :: ii,jj
      open(10,file=figfold//'micount_'//inttochar(1,klvar)//'.dat',status='replace')
      write(10,'(5a10)') "x","one","two","three","four"
      do ii=1,sim_time
         write(10,'(5f10.4)') dble(ii)*dble(1560)/(dble(52)*dble(sim_time)),&
              &((sum(mutual(ii,:,0))-sum(mutual(ii,:,jj)))/sum(mutual(ii,:,0)),jj=1,4)
      end do
      close(10)
      open(10,file=figfold//'wcount_'//inttochar(1,klvar)//'.dat',status='replace')
      write(10,'(5a10)') "x","one","two","three","four"
      do ii=1,sim_time
         write(10,'(5f10.4)') dble(ii)*dble(1560)/(dble(52)*dble(sim_time)),&
              &((sum(corrab(ii,:,0))-sum(corrab(ii,:,jj)))/sum(corrab(ii,:,0)),jj=1,4)
      end do
      close(10)
    end subroutine fig_cohort_counter

    subroutine sort_contrib
      integer :: a,b,c,d,ii,jj
      ii=0  
      do a=1,4
         do b=1,4
            if(b .eq. a) cycle
            do c=1,4
               if(c .eq. a .or. c .eq. b) cycle
               do d=1,4
                  if(d .eq. a .or. d .eq. b .or. d .eq. c) cycle
                  ii=ii+1
                  !wage correlations
                  write(tot(1,ii)%order,'(4i1)') a,b,c,d
                  tot(1,ii)%contrib(a)=(sum(sum(corrab(:,:,0),dim=1))-sum(sum(corrab(:,:,cfmap(a)),dim=1)))
                  tot(1,ii)%contrib(b)=(sum(sum(corrab(:,:,cfmap(a)),dim=1))-sum(sum(corrab(:,:,cfmap(a,b)),dim=1)))
                  tot(1,ii)%contrib(c)=(sum(sum(corrab(:,:,cfmap(a,b)),dim=1))-sum(sum(corrab(:,:,cfmap(a,b,c)),dim=1)))
                  tot(1,ii)%contrib(d)=sum(sum(corrab(:,:,cfmap(a,b,c)),dim=1))
                  tot(1,ii)%contrib=tot(1,ii)%contrib/sum(sum(corrab(:,:,0),dim=1))
                  
                  !mutual information
                  write(tot(2,ii)%order,'(4i1)') a,b,c,d
                  tot(2,ii)%contrib(a)=(sum(sum(mutual(:,:,0),dim=1))-sum(sum(mutual(:,:,cfmap(a)),dim=1)))
                  tot(2,ii)%contrib(b)=(sum(sum(mutual(:,:,cfmap(a)),dim=1))-sum(sum(mutual(:,:,cfmap(a,b)),dim=1)))
                  tot(2,ii)%contrib(c)=(sum(sum(mutual(:,:,cfmap(a,b)),dim=1))-sum(sum(mutual(:,:,cfmap(a,b,c)),dim=1)))
                  tot(2,ii)%contrib(d)=sum(sum(mutual(:,:,cfmap(a,b,c)),dim=1))
                  tot(2,ii)%contrib=tot(2,ii)%contrib/sum(sum(mutual(:,:,0),dim=1))
               end do
            end do
         end do
      end do
      call tab_cohort_counter
    end subroutine sort_contrib

    subroutine tab_cohort_counter
      integer :: ii,jj,i,j
      character(250) :: fmt
      type(multch) :: avg(2),std(2)
      do jj=1,2
         avg(jj)%contrib=0.d0
         std(jj)%contrib=0.d0
         do ii=1,24
            avg(jj)%contrib=avg(jj)%contrib+tot(jj,ii)%contrib
            std(jj)%contrib=std(jj)%contrib+tot(jj,ii)%contrib*tot(jj,ii)%contrib
         end do
         avg(jj)%contrib=avg(jj)%contrib/dble(24)
         std(jj)%contrib=std(jj)%contrib/dble(24)-avg(jj)%contrib*avg(jj)%contrib
      end do
      open(10,file=tabfold//"tab_cohort_counter_"//inttochar(1,klvar)//".tex",status='replace')
      write(10,'("\documentclass{standalone}")')
      write(10,'("\usepackage{basics}")')
      write(10,'("\begin{document}")')
      write(10,'("\centering")')
      write(10,'("\begin{tabular}{c| *{4}{d{2.3}} | *{4}{d{2.3}}} ")')
      write(10,'("\hline")')
      write(10,'("Elimination & \multicolumn{4}{c|}{Wage sorting} & \multicolumn{4}{c}{Type sorting}\tabularnewline")')
      write(10,'("order & \mc{$\gamma$} & \mc{$\delta$} & \mc{$\lambda$} & ",&
           "\multicolumn{1}{c|}{$\psi$} & \mc{$\gamma$} & \mc{$\delta$} & \mc{$\lambda$} & \mc{$\psi$}\tabularnewline")')
      do i=1,4
         write(10,'("\hline")')
         do j=1,6
            fmt='(a4'
            do jj=1,2
               do ii=1,4
                  if(ii.eq.i) then
                     fmt=trim(fmt)//'," & \cellcolor{gray!10}",f6.3'
                  else
                     fmt=trim(fmt)//'," & ",f6.3'
                  end if
               end do
            end do
            fmt=trim(fmt)//',"\tabularnewline")'
            write(10,trim(fmt)) tot(1,(i-1)*6+j)%order,(tot(jj,(i-1)*6+j)%contrib,jj=1,2)
         end do
      end do
      write(10,'("\hline")')
      write(10,'("\rowcolor{gray!15}",a8,8(" & ",f6.3),"\tabularnewline")') "Average",(avg(jj)%contrib,jj=1,2)
      write(10,'("\rowcolor{gray!15}",a8,8(" & ",f6.3),"\tabularnewline")') "Std dev",(sqrt(std(jj)%contrib),jj=1,2)
      write(10,'("\hline")')
      write(10,'("\end{tabular}")')
      write(10,'("\end{document}")')
      close(10)
    end subroutine tab_cohort_counter

    function logtoint(x) result(res)
      logical :: x
      integer :: res
      res=0
      if(x) res=1
    end function logtoint

    function cfmap(ca,cb,cc) result(res)
      !expects that if cb is not present then cc is also not
      !It also expects that ca, cb, cc not to equal each other.
      integer, intent(in) :: ca
      integer, intent(in), optional :: cb,cc
      integer :: res
      integer :: two,three
      res=0
      if(.not. present(cb)) then
         res=ca
      else
         if(.not. present(cc)) then
            two=min(ca,cb)*10+max(ca,cb)
            if(two .eq. 12) res=5
            if(two .eq. 13) res=6
            if(two .eq. 14) res=7
            if(two .eq. 23) res=8
            if(two .eq. 24) res=9
            if(two .eq. 34) res=10
         else               
            three=min(ca,cb,cc)*100+max(ca,cb,cc)
            if(ca .ne. min(ca,cb,cc) .and. ca .ne. max(ca,cb,cc)) three=three+10*ca
            if(cb .ne. min(ca,cb,cc) .and. cb .ne. max(ca,cb,cc)) three=three+10*cb
            if(cc .ne. min(ca,cb,cc) .and. cc .ne. max(ca,cb,cc)) three=three+10*cc
            if(three .eq. 123) res=11
            if(three .eq. 124) res=12
            if(three .eq. 134) res=13
            if(three .eq. 234) res=14
         end if
      end if
    end function cfmap
  end subroutine cohortsim_sort

  subroutine table_vardecomp
    integer :: t,i
    real(8) :: decom(10,5)
    do t=1,5
       call uncond_vardecom(decom(:,t),est(t))
    end do
    open(10,file=tabfold//"tab_vardecomp.tex",status='replace')
      write(10,'("\documentclass{standalone}")')
      write(10,'("\usepackage{basics}")')
      write(10,'("\begin{document}")')
      write(10,'("\centering")')
      write(10,'("\begin{tabular}{l *{5}{d{3.2}} }")')
      write(10,'("\hline")')
      write(10,'(" & \mc{$89-93$} & \mc{$94-98$} & \mc{$99-03$} & \mc{$04-08$} & \mc{$09-13$}\tabularnewline")')
      write(10,'("\hline")')
      write(10,'("Residual",5(" & ",f6.2),"\tabularnewline")') decom(1,:)
      write(10,'("Total between",5(" & ",f6.2),"\tabularnewline")') decom(9,:)
      write(10,'("\hspace{.3cm}worker effect, $a_{k}$",5(" & ",f6.2),"\tabularnewline")') decom(2,:)
      write(10,'("\hspace{.3cm}match effect, $\widetilde{\mu}_{k\ell}(x)$",5(" & ",f6.2),"\tabularnewline")') decom(3,:)
      write(10,'("\hspace{.3cm}firm effect, $b_{\ell}$",5(" & ",f6.2),"\tabularnewline")') decom(4,:)
      write(10,'("\hspace{.3cm}sorting effect, $2\Cov(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') decom(5,:)
      write(10,'("\hspace{.3cm}tenure and experience, $x$",5(" & ",f6.2),"\tabularnewline")') (sum(decom(6:8,t)),t=1,5)
      write(10,'("\hspace{.6cm}$\overline{\mu}(x)$",5(" & ",f6.2),"\tabularnewline")') decom(6,:)
      write(10,'("\hspace{.6cm}$2\Cov(a_{k},\overline{\mu}(x))$",5(" & ",f6.2),"\tabularnewline")') decom(7,:)
      write(10,'("\hspace{.6cm}$2\Cov(b_{\ell},\overline{\mu}(x))$",5(" & ",f6.2),"\tabularnewline")') decom(8,:)
      write(10,'("\hline")')
      write(10,'("Wage correlation, $\Cor(a_{k},b_{\ell})$",5(" & ",f6.2),"\tabularnewline")') decom(10,:)
      write(10,'("\hline")')
      write(10,'("\end{tabular}")')
      write(10,'("\end{document}")')
      close(10)      
  end subroutine table_vardecomp

  subroutine uncond_vardecom(decom,my)
    implicit none
    type(fullest) :: my
    real(8), intent(out) :: decom(10)
    real(8) :: pklx_un(my%par%dim%k,my%par%dim%l,my%par%dim%het)
    real(8) :: muwtilde(my%par%dim%k,my%par%dim%l,my%par%dim%het)
    real(8) :: vw, emuw2, emuw, emuwt2, emuwt, ea, eb, ehet,ea2, eb2,ehet2, eab, eahet, ebhet
    integer :: k,l,i,x

    pklx_un=my%pklx_emp()

    decom=0.d0
    muwtilde=0.d0

    do x=1,my%par%dim%het
       do l=1,my%par%dim%l
          do k=1,my%par%dim%k
             muwtilde(k,l,x)=my%par%muw(k,l,x)-my%het(x)-my%c-my%ak(k)-my%bl(l)
          end do
       end do
    end do

    vw=0.d0
    emuw2=0.d0
    emuw=0.d0
    emuwt2=0.d0
    emuwt=0.d0
    ea=0.d0
    eb=0.d0
    ehet=0.d0
    ea2=0.d0
    eb2=0.d0
    eb=0.d0
    ehet2=0.d0
    eab=0.d0
    eahet=0.d0
    ebhet=0.d0
    do x=1,my%par%dim%het
       do l=1,my%par%dim%l
          do k=1,my%par%dim%k
             vw = vw+pklx_un(k,l,x)*my%varwx(x)
             emuw2=emuw2+pklx_un(k,l,x)*my%par%muw(k,l,x)**2
             emuw=emuw+pklx_un(k,l,x)*my%par%muw(k,l,x)
             emuwt2=emuwt2+pklx_un(k,l,x)*muwtilde(k,l,x)**2
             emuwt=emuwt+pklx_un(k,l,x)*muwtilde(k,l,x)
             ea=ea+pklx_un(k,l,x)*my%ak(k)
             eb=eb+pklx_un(k,l,x)*my%bl(l)
             ehet=ehet+pklx_un(k,l,x)*my%het(x)
             ea2=ea2+pklx_un(k,l,x)*my%ak(k)**2
             eb2=eb2+pklx_un(k,l,x)*my%bl(l)**2
             ehet2=ehet2+pklx_un(k,l,x)*my%het(x)**2
             eab=eab+pklx_un(k,l,x)*my%ak(k)*my%bl(l)
             eahet=eahet+pklx_un(k,l,x)*my%ak(k)*my%het(x)
             ebhet=ebhet+pklx_un(k,l,x)*my%bl(l)*my%het(x)
          end do
       end do
    end do
    
    !Table 7 in appendix, all in percentage of vw
    decom(1)= (vw- (emuw2-(emuw)**2))/vw*100  !Residual = VW - VE(variance of muw)
    decom(2)= (ea2-(ea)**2)/vw*100   !var a
    decom(3)= (emuwt2-(emuwt)**2)/vw*100 !muw tilde	
    decom(4)= (eb2-(eb)**2)/vw*100   !var b
    decom(5)= 2*(eab-ea*eb)/vw*100  !2*cov (a,b)
    decom(6)= (ehet2-(Ehet)**2)/vw*100   !var het or muw bar
    decom(7)= 2*(eahet-ea*ehet)/vw*100  !2*cov (a,het)
    decom(8)= 2*(ebhet-eb*ehet)/vw*100  !2*cov (b,het)
    decom(9)= (emuw2-(emuw)**2)/vw*100  !total between or sum of decom(2:8)
    decom(10) = (eab-ea*eb)/((ea2-(ea)**2) * (eb2-(eb)**2))**0.5 !corr(a,b)
  end subroutine uncond_vardecom

  subroutine pklxcalc(mypar,mya,myb)
    type(modelpar), intent(in) :: mypar
    real(8), intent(out) :: mya(mypar%dim%k), myb(mypar%dim%l)
    real(8) :: myc,myhet(mypar%dim%het)
    integer :: i,ii,j
    integer :: Ix(mypar%dim%het),myIx(mypar%dim%het) !number of observations for each x category
    real(8) :: mypklx(mypar%dim%k,mypar%dim%l,mypar%dim%het),pklx(mypar%dim%k,mypar%dim%l,mypar%dim%het) !p_{kl}(x) in variance decomposition
    real(8) :: pklx_un(mypar%dim%k,mypar%dim%l,mypar%dim%het)
    real(8), dimension(mypar%dim%het) :: myew,myww,ew,ww,varw
    real(8) :: wwtot,ewtot,varwtot
    real(8) :: linstat(10)

    mypklx=0.d0
    myIx=0
    myew=0.d0
    myww=0.d0

    do i=1,dim_I
       if(associated(work(i)%first)) then !if parallel, work(i)%first is only allocated for cpu that did this history.
          call pklcount(work(i)%first)
       end if
    end do

    call mpi_allreduce(mypklx,pklx,mypar%dim%k*mypar%dim%l*mypar%dim%het,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(myIx,Ix,mypar%dim%het,mpi_integer,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(myew,ew,mypar%dim%het,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
    call mpi_allreduce(myww,ww,mypar%dim%het,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)

    ewtot=sum(ew)
    wwtot=sum(ww)
    ew=ew/dble(Ix)
    ewtot=ewtot/sum(dble(Ix))
    varw=ww/dble(Ix)-ew*ew
    varwtot=wwtot/sum(dble(Ix))-ewtot*ewtot

    pklx_un=pklx/sum(pklx)

    call linproj_common(mypar%muw,pklx_un,mya,myb,myc,myhet) !calculates a and b for given model parameters
  contains
    recursive subroutine pklcount(mysp)
      type(spell), pointer :: mysp
      integer :: i,k 
      do i=1,size(mysp%wage)
      	 if(mysp%fid>0) then
            myIx(hetmap(mysp,i))=myIx(hetmap(mysp,i))+1
            myew(hetmap(mysp,i))=myew(hetmap(mysp,i))+mysp%wage(i)
            myww(hetmap(mysp,i))=myww(hetmap(mysp,i))+mysp%wage(i)*mysp%wage(i)
         end if
         if(mysp%fid>0) mypklx(work(mysp%wid)%k,firm(mysp%fid)%l,hetmap(mysp,i))=&
              &mypklx(work(mysp%wid)%k,firm(mysp%fid)%l,hetmap(mysp,i))+1.0d0
      end do
      if(associated(mysp%next)) call pklcount(mysp%next)
    end subroutine pklcount
  end subroutine pklxcalc

  subroutine linproj_common(mypar,pklx_un,outa,outb,outc,outhet)
    !expecting to match dimensions included in mypar. Notice ordering is pklx_un(k,l,x).
    !pklx_un is the joint (k,l,x) distribution conditional on employment.
    implicit none
    real(8), intent(in), dimension(:,:,:) :: mypar !take a 3 dimensional input to be linearly projected on
    real(8), dimension(:,:,:), intent(in) :: pklx_un !joint distribution over 3 dimensions of mypar
    real(8), intent(out) :: outa(size(mypar,dim=1)),outb(size(mypar,dim=2)), outc, outhet(size(mypar,dim=3))
    integer(1), allocatable :: xmat(:,:)
    real(8), allocatable :: w(:,:),y(:,:),res(:,:),xxinv(:,:),pk(:),pl(:),px(:)
    integer :: k,l,i,x
    integer :: ck,cl,cx !dimensions

    ck=size(mypar,dim=1)
    cl=size(mypar,dim=2)
    cx=size(mypar,dim=3)

    allocate(xmat(ck*cl*cx,ck+cl+cx-2),w(ck*cl*cx,ck*cl*cx),y(ck*cl*cx,1),&
         &res(ck+cl+cx-2,1),xxinv(ck+cl+cx-2,ck+cl+cx-2),pk(ck),pl(cl),px(cx))
    
    i=0
    w=0.d0
    y=0.d0
    xmat=0
    pk=0.d0
    pl=0.d0
    px=0.d0

    do x=1,cx
       do l=1,cl
          do k=1,ck
             pk(k)=pk(k)+pklx_un(k,l,x) !k marginal
             pl(l)=pl(l)+pklx_un(k,l,x) !l marginal
             px(x)=px(x)+pklx_un(k,l,x) !x marginal
             i=i+1
             w(i,i)=pklx_un(k,l,x)
             y(i,1)=mypar(k,l,x) !y is stacked k-l-x (k innermost, then l, then x)
             call set_dummies(i,k,l,x) !build the dummy matrix
          end do
       end do
    end do

    xxinv=matmul(matmul(transpose(xmat),w),xmat)
    xxinv=matinv_lapack(xxinv)
    res=matmul(matmul(matmul(xxinv,transpose(xmat)),w),y)

    outc=res(1,1)
    outa(1)=0.d0
    outb(1)=0.d0
    outhet(1)=0.d0
    outa(2:ck)=res(2:ck,1)
    outb(2:cl)=res(ck+1:ck+cl-1,1)
    outhet(2:cx)=res(ck+cl:ck+cl+cx-2,1)

    outc=outc+avg(outa,pk)+avg(outb,pl)+avg(outhet,px) !is this necessary?
    outa=outa-avg(outa,pk)
    outb=outb-avg(outb,pl)
    outhet=outhet-avg(outhet,px)
  contains
    subroutine set_dummies(r,myk,myl,myx)
      integer, intent(in) :: r !act on row r in xmat
      integer, intent(in) :: myk,myl,myx
      xmat(r,1)=1 !constant in regression.
      if(myk>1) xmat(r,myk)=1 !skip if myk=1 (reference category)
      if(myl>1) xmat(r,ck+myl-1)=1 !skip if myl=1 (reference category)
      if(myx>1) xmat(r,ck+cl+myx-2)=1 !skip if myx=1 (reference category).
    end subroutine set_dummies

    function avg(x,w) result(res)
      real(8), dimension(:), intent(in) :: x(:),w(:)
      real(8) :: res
      integer :: i
      if(size(x) .ne. size(w)) stop 'avg: size mismatch.'
      res=0.d0
      do i=1,size(x)
         res=res+x(i)*w(i)
      end do
    end function avg
  end subroutine linproj_common

  subroutine fig1
    type(tcdf) :: byt(5)
    integer, parameter :: evalpoints=100
    integer :: k,l,x,t,cc
    real(8) :: vec(8),ev(2),eval,asup(2),bsup(2)
    do t=1,5
       byt(t)%est=>est(t)
       byt(t)%kdim=est(t)%par%dim%k
       byt(t)%ldim=est(t)%par%dim%l
       allocate(byt(t)%krank(0:est(t)%par%dim%k),byt(t)%pkcdf(0:est(t)%par%dim%k),&
            &byt(t)%lrank(0:est(t)%par%dim%l),byt(t)%plcdf(0:est(t)%par%dim%l))
       byt(t)%krank=0.d0
       byt(t)%lrank=0.d0
       byt(t)%pkcdf=0.d0
       byt(t)%plcdf=0.d0
       do k=1,est(t)%par%dim%k
          byt(t)%krank(k)=byt(t)%est%ak(k)
          byt(t)%pkcdf(k)=byt(t)%pkcdf(k-1)+est(t)%pek(k)
       end do
       do l=1,est(t)%par%dim%l
          byt(t)%lrank(l)=byt(t)%est%bl(l)
          byt(t)%plcdf(l)=byt(t)%plcdf(l-1)+est(t)%pel(l)
       end do
    end do

    !define evaluation support: union
    asup(1)=byt(1)%est%ak(1)
    asup(2)=byt(1)%est%ak(byt(1)%kdim)
    bsup(1)=byt(1)%est%bl(1)
    bsup(2)=byt(1)%est%bl(byt(1)%ldim)
    do t=2,5
       asup(1)=min(asup(1),byt(t)%est%ak(1))
       asup(2)=max(asup(2),byt(t)%est%ak(byt(t)%kdim))
       bsup(1)=min(bsup(1),byt(t)%est%bl(1))
       bsup(2)=max(bsup(2),byt(t)%est%bl(byt(t)%ldim))
    end do

    open(12,file=figfold//"fig1c.dat",status="replace")
    write(12,'(a4,15a9)') "k","ak1","ak2","ak3","ak4","ak5","pk1","pk2","pk3","pk4","pk5","t1","t2","t3","t4","t5"
    do k=1,24
       write(12,'(I4,10f9.5,5I9)') k,(byt(t)%est%ak(k),t=1,5),(byt(t)%pkcdf(k),t=1,5),(t,t=1,5)
    end do
    close(12)

    open(12,file=figfold//"fig1d1.dat",status="replace")
    write(12,'(a4,6a9)') "l","bl1","bl2","pl1","pl2","t1","t2"
    do l=1,14
       write(12,'(I4,4f9.5,2I9)') l,(byt(t)%est%bl(l),t=1,2),(byt(t)%plcdf(l),t=1,2),(t,t=1,2)
    end do
    close(12)

    open(12,file=figfold//"fig1d2.dat",status="replace")
    write(12,'(a4,9a9)') "l","bl3","bl4","bl5","pl3","pl4","pl5","t3","t4","t5"
    do l=1,22
       write(12,'(I4,6f9.5,3I9)') l,(byt(t)%est%bl(l),t=3,5),(byt(t)%plcdf(l),t=3,5),(t,t=3,5)
    end do
    close(12)

    open(12,file=figfold//"fig1e.dat",status="replace")
    write(12,'(3a9)') "rank","a","b"
    do l=1,50
       eval=dble(l-1)/dble(49)
       ev=0.d0
       do t=1,5
          ev(1)=ev(1)+linint(byt(t)%pkcdf(1:byt(t)%kdim),byt(t)%est%ak,eval)
          ev(2)=ev(2)+linint(byt(t)%plcdf(1:byt(t)%ldim),byt(t)%est%bl,eval)
       end do
       ev=ev/dble(5)
       write(12,'(3f9.5)') eval,ev
    end do
    close(12)

    open(12,file=figfold//"fig1f.dat",status="replace")
    write(12,'(10a9)') "pa","pb","muw1","muw2","muw3","muw4","muw5","muw6","muw7","muw8"
    do k=1,evalpoints
       do l=1,evalpoints
          ev=(/ asup(1)+ dble(k-1)*(asup(2)-asup(1))/dble(evalpoints-1), bsup(1)&
               & + dble(l-1)*(bsup(2)-bsup(1))/dble(evalpoints-1) /)
          write(12,'(10f9.5)') ev(1),ev(2),(muwsmooth(x,ev,.25d0,10),x=1,8)
       end do
    end do
    close(12)    
  end subroutine fig1

  subroutine fig2
    type(tcdf) :: byt(5)
    integer :: k,l,x,t,z
    real(8) :: vec(14),ev
    do t=1,5
       byt(t)%kdim=est(t)%par%dim%k
       byt(t)%ldim=est(t)%par%dim%l
       allocate(byt(t)%zcdf(0:byt(t)%kdim,6),byt(t)%krank(0:byt(t)%kdim),byt(t)%expcdf(0:byt(t)%kdim,4),&
            &byt(t)%tencdf(0:byt(t)%kdim,2,2))
       byt(t)%krank=0.d0
       byt(t)%zcdf=0.d0
       byt(t)%expcdf=0.d0
       byt(t)%tencdf=0.d0
       do k=1,byt(t)%kdim
          do l=0,byt(t)%ldim
             do x=1,8
                byt(t)%krank(k)=byt(t)%krank(k)+est(t)%pp(k,l,x)
             end do
             do x=1,4
                byt(t)%expcdf(k,x)=byt(t)%expcdf(k,x)+est(t)%pp(k,l,x)+est(t)%pp(k,l,x+4)
             end do
             do z=1,2
                do x=1,4
                   if(l.eq.0) then
                      byt(t)%tencdf(k,z,1)=byt(t)%tencdf(k,z,1)+est(t)%pp(k,l,x+(z-1)*4)
                   else
                      byt(t)%tencdf(k,z,2)=byt(t)%tencdf(k,z,2)+est(t)%pp(k,l,x+(z-1)*4)
                   end if
                end do
             end do
          end do
          do z=1,6
             byt(t)%zcdf(k,z)=byt(t)%zcdf(k-1,z)+est(t)%par%piw(z,k)*byt(t)%krank(k)
          end do
          byt(t)%krank(k)=byt(t)%krank(k)+byt(t)%krank(k-1)
          byt(t)%expcdf(k,:)=byt(t)%expcdf(k,:)+byt(t)%expcdf(k-1,:)
          byt(t)%tencdf(k,:,:)=byt(t)%tencdf(k,:,:)+byt(t)%tencdf(k-1,:,:)
       end do
       byt(t)%krank=byt(t)%krank/byt(t)%krank(byt(t)%kdim) !shouldn't be necessary.
       do z=1,6
          byt(t)%pz(z)=byt(t)%zcdf(byt(t)%kdim,z)
          byt(t)%zcdf(:,z)=byt(t)%zcdf(:,z)/byt(t)%pz(z)
       end do
       do x=1,4
          byt(t)%pexp(x)=byt(t)%expcdf(byt(t)%kdim,x)
          byt(t)%expcdf(:,x)=byt(t)%expcdf(:,x)/byt(t)%pexp(x)
       end do
       do z=1,2
          byt(t)%pten(z,:)=byt(t)%tencdf(byt(t)%kdim,z,:)
          byt(t)%tencdf(:,z,1)=byt(t)%tencdf(:,z,1)/byt(t)%pten(z,1)
          byt(t)%tencdf(:,z,2)=byt(t)%tencdf(:,z,2)/byt(t)%pten(z,2)
       end do
    end do

    open(12,file=figfold//"fig2.dat",status="replace")
    write(12,'(15a9)') "krank","z1","z2","z3","z4","z5","z6","exp1","exp2","exp3","exp4","eten1","eten2","uten1","uten2"
    do l=1,51
       ev=dble(l-1)/50
       vec=0.d0
       do t=1,5
          do z=1,6
             vec(z)=vec(z)+byt(t)%zeval(z,ev) !by z
          end do
          do z=1,4
             vec(z+6)=vec(z+6)+byt(t)%expeval(z,ev) !by experience
          end do
          do z=1,2
             vec(z+10)=vec(z+10)+byt(t)%eteneval(z,ev) !by employed tenure
          end do
          do z=1,2
             vec(z+12)=vec(z+12)+byt(t)%uteneval(z,ev) !by employed tenure
          end do
       end do
       vec=vec/dble(5)
       write(12,'(15f9.5)') ev,vec
    end do
    close(12)
  end subroutine fig2

  subroutine fig3
    type(txcdf) :: tx(5,8)
    integer :: l,k,x,q,t
    integer, parameter :: evdim=51
    real(8) :: ev,vec(6)

    do x=1,8
       open(12,file=figfold//"fig3_"//inttochar(1,x)//".dat",status="replace")
       write(12,'(7a9)') "lrank","q1","q2","q3","q4","q5","q6"
       do l=1,evdim
          ev=dble(l-1)/dble(evdim-1)
          write(12,'(7f9.5)') ev,(quantsmooth(x,(/ ev, dble(q)/dble(6) /),.1d0,20),q=1,5),dble(1)
       end do
       close(12)
    end do
  end subroutine fig3

  subroutine readest(ch,myest)
    integer, intent(in) :: ch
    type(fullest), intent(inout) :: myest
    type(pardim) :: mydim
    integer :: ii,jj,tt,kk
    real(8) :: x
    read(ch,*) mydim !read data dimensions
    allocate(myest%p(mydim%k,0:mydim%l,mydim%sig),myest%ak(mydim%k),myest%bl(mydim%l),&
         &myest%varwx(mydim%het),myest%het(mydim%het),&
         &myest%wfe(mydim%k,6),myest%ffe(mydim%l,6))
    myest%par=new_modelpar_new(mydim)
    !read muw
    do tt=1,myest%par%dim%het
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%muw(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read sigma
    do tt=1,myest%par%dim%sig
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%sigmaw(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read gam
    do tt=1,myest%par%dim%mob
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%gam(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read lam
    do tt=1,myest%par%dim%mob
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%lam(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read psi
    do tt=1,myest%par%dim%mob
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%psi(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read delta
    do tt=1,myest%par%dim%mob
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%del(ii,jj,tt), jj=1,myest%par%dim%l)
       end do
    end do

    !read piw
    do tt=1,myest%par%dim%z
       read(ch,'(*(f15.8))') (myest%par%piw(tt,ii),ii=1,myest%par%dim%k)
    end do

    !read pif
    do tt=1,myest%par%dim%s
       read(ch,'(*(f15.8))') (myest%par%pif(tt,jj),jj=1,myest%par%dim%l)
    end do

    !read m (initial distribution)
    do tt=1,myest%par%dim%ini
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%m(ii,jj,tt), jj=0,myest%par%dim%l)
       end do
    end do

    !read q
    read(ch,'(*(f15.8))') (myest%par%q(jj),jj=1,myest%par%dim%l) !notice: only writing 1:L. Q includes 0 index that is by default set to 1.

    !read p
    read(ch,'(*(f15.8))') (myest%par%pk(ii),ii=1,myest%par%dim%k)

    !read pz (marginal z distribution)
    read(ch,'(*(f15.8))') (myest%par%pz(jj),jj=1,myest%par%dim%z)

    !read ps (marginal s distribution)
    read(ch,'(*(f15.8))') (myest%par%ps(jj),jj=1,myest%par%dim%s)

    !read rho
    read(ch,'(f15.8)') myest%par%rho

    !read MM
    do tt=1,myest%par%dim%mob
       do kk=1,myest%par%dim%k
          do ii=0,myest%par%dim%l
             read(ch,'(*(f15.8))') (myest%par%MM(kk,ii,jj,tt), jj=0,myest%par%dim%l)
          end do
       end do
    end do

    !read Mbar
    do tt=1,myest%par%dim%mob
       do kk=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%par%Mbar(kk,jj,tt), jj=0,myest%par%dim%l)
       end do
    end do

    !read p(k,l,x)
    do tt=1,myest%par%dim%sig
       do ii=1,myest%par%dim%k
          read(ch,'(*(f15.8))') (myest%p(ii,jj,tt), jj=0,myest%par%dim%l)
       end do
    end do

    !read varwx
    read(ch,'(*(f15.8))') (myest%varwx(tt), tt=1,myest%par%dim%het)
    
    !read a
    read(ch,'(*(f15.8))') (myest%ak(ii), ii=1,myest%par%dim%k)

    !read b
    read(ch,'(*(f15.8))') (myest%bl(jj), jj=1,myest%par%dim%l)

    !check end-of-file
    read(ch,*,iostat=ii) x
    if(ii .ge. 0) stop 'Fatal: read did not reach end of file as expected'
  end subroutine readest
end module postest
