module datbysim
  !contains simulation specific parameters as well as simulated data structure
  use mpimod
  use estim
  use sort3 !contains hpsort
  use marsaglia !contains kiss
  use tools
  implicit none !all objects in module inherit implicit none.
  integer, parameter :: mydim_I  =   4000000 !# of workers.
  integer, parameter :: mydim_J  =    400000 !# of firms.
  integer, parameter :: sim_T    =     1560 !# of periods i.e. 30 years
  integer, parameter :: yr       =       52 !number of periods in a year.
  integer, parameter :: TT=1040, npvdim=1000000
  real(8), parameter :: beta=(.95d0)**(dble(1)/dble(52)) !discount factor
  real(8), allocatable :: cdf(:,:,:,:)
  type simstate
     integer :: k,l
     integer :: exp !experience
     integer :: ten !tenure
     integer :: t !time
     integer :: empc !number of periods in employment.
     real(8) :: npv
  end type simstate

  type(simstate) :: now  
  type(modelpar) :: par !globally set collection of model parameters. Must be initialized.
contains
  subroutine simNPV(npvout) !par is set outside of call
    implicit none
    real(8), intent(out) :: npvout(par%dim%k,0:par%dim%l,par%dim%het)
    integer :: k,l,ll,x,t,s,myseed,yy,i
    type(simstate) :: now
    real(8) :: xx
    real(8),allocatable :: NPV(:,:,:)
    if(allocated(cdf)) deallocate(cdf)
    allocate(cdf(par%dim%k,0:par%dim%l,par%dim%het,-1:par%dim%l),NPV(par%dim%k,0:par%dim%l,par%dim%het))

    do k=1,par%dim%K
       do l=0,par%dim%L
          do x=1,par%dim%het
             cdf(k,l,x,-1)=0.d0
             do ll=0,par%dim%l
                cdf(k,l,x,ll)=cdf(k,l,x,ll-1)+par%MM(k,l,ll,x)
             end do
             cdf(k,l,x,:)=cdf(k,l,x,:)/cdf(k,l,x,par%dim%l) !just to make sure that rounding errors are corrected.
          end do
       end do
    end do

    yy=-1
    NPV=0.d0
    do x=1,par%dim%het
       do k=1,par%dim%k
          do l=0,par%dim%l
             yy=yy+1
             if(mod(yy,numprocs) .ne. myid) cycle !Divide up (k,l,x) tasks
             call setseed(kisseed(2345678+2*yy,87656789+200*yy,5434567+100*yy,43212345+500*yy))
             !warm up kiss
             do i=1,10000+yy
                xx=kissuni()
             end do
             !go
             do s=1,npvdim
                now%k=k
                now%l=l
                call inidraw(now,x)
                do while(now%t<=TT)
                   now=Next(now)
                end do
                NPV(k,l,x)=NPV(k,l,x)+now%npv
             end do
             NPV(k,l,x)=NPV(k,l,x)/dble(npvdim) !average of npvdim simulations.
             write(6,'(i1,$)') x
          end do
       end do
    end do
    call mpi_barrier(mpi_comm_world,ierr)
    call mpi_allreduce(npv,npvout,par%dim%k*(1+par%dim%l)*par%dim%het,mpi_double_precision,mpi_sum,mpi_comm_world,ierr)
  contains
    function Next(my) result(res)
      type(simstate), intent(in) :: my
      type(simstate) :: res
      integer :: dur,cap,t
      cap=dur_until_change(my)
      dur=int(log(1.d0-kissuni())/log(par%mbar(my%k,my%l,StateToX(my))))+1.d0
      if(dur>cap) then
         dur=cap
         res%l=my%l
         res%ten=my%ten+dur
      else !Now, draw destination state
         res%l=findndx(cdf(my%k,my%l,StateTox(my),:),kissuni()) - 1
         res%ten=0
      end if
      if(my%l==0) then
         res%empc=my%empc
      else
         res%empc=my%empc+dur
      end if
      res%k=my%k
      res%exp=my%exp+dur
      res%npv=my%npv
      do t=my%t,my%t+dur-1
         res%npv=res%npv+(beta**(t-1))*WageDraw(my)
      end do
      res%t=t
    end function Next

    subroutine inidraw(my,myx)
      implicit none
      type(simstate), intent(inout) :: my
      integer, intent(in) :: myx
      real(8) :: u,uu
      u=kissuni() !used for the experience draw
      uu=kissuni() !used for the tenure draw
      my%npv=0.d0
      my%t=1
      my%empc=0
      if(myx<1 .or. myx>8) stop 'inidraw got bad input.'
      select case(myx)
      case(1)
         my%exp=int(u*260)+1 !uniformly set initial experience within the x category
         if(my%l==0) then
            my%ten=int(uu*26)+1
         else
            my%ten=int(uu*100)+1
         end if
      case(2)
         my%exp=int(u*260)+261
         if(my%l==0) then
            my%ten=int(uu*26)+1
         else
            my%ten=int(uu*100)+1
         end if
      case(3)
         my%exp=int(u*260)+521
         if(my%l==0) then
            my%ten=int(uu*26)+1
         else
            my%ten=int(uu*100)+1
         end if
      case(4)
         my%exp=782 !Any experience level above 782 is treated the same.
         if(my%l==0) then
            my%ten=int(uu*26)+1
         else
            my%ten=int(uu*100)+1
         end if
      case(5)
         my%exp=int(u*260)+1
         my%ten=101
      case(6)
         my%exp=int(u*260)+261
         my%ten=101
      case(7)
         my%exp=int(u*260)+521
         my%ten=101
      case(8)
         my%exp=782
         my%ten=101
      end select
    end subroutine inidraw

    function WageDraw(my) result(res)
      implicit none
      type(simstate), intent(in) :: my
      real(8) :: res
      res=0.d0
      if(my%l==0) return !if non-employed log wage is zero.
      res=par%muw(my%k,my%l,StateToX(my))/dble(52) !just do the average wage draw.
    end function WageDraw

    function StateToX(my) result(res)
      implicit none
      type(simstate), intent(in) :: my
      integer :: res
      select case(my%exp)
      case(: 260)
         res=1
      case(261:520)
         res=2
      case(521:780)
         res=3
      case(781:)
         res=4
      end select
      if(my%l==0) then
         if(my%ten>26) res=res+4
      else
         if(my%ten>100) res=res+4
      end if
    end function StateToX

    function dur_until_change(my) result(res) !How long until worker's state changes or end of time.
      type(simstate), intent(in) :: my
      integer :: res
      select case(StateToX(my)) !can take values 1 through 8
      case(1)
         if(my%l==0) then
            res=min(261-my%exp,27-my%ten)
         else
            res=min(261-my%exp,101-my%ten)
         end if
      case(2)
         if(my%l==0) then
            res=min(521-my%exp,27-my%ten)
         else
            res=min(521-my%exp,101-my%ten)
         end if
      case(3)
         if(my%l==0) then
            res=min(781-my%exp,27-my%ten)
         else
            res=min(781-my%exp,101-my%ten)
         end if
      case(4)
         if(my%l==0) then
            res=27-my%ten
         else
            res=101-my%ten
         end if
      case(5)
         res=261-my%exp
      case(6)
         res=521-my%exp
      case(7)
         res=781-my%exp
      case(8)
         res=TT+1
      end select
      res=min(res,TT-my%t+1)
    end function dur_until_change
  end subroutine simNPV

  subroutine cleansim
    integer :: i
    call mpi_barrier(mpi_comm_world,ierr)
    do i=1,dim_I !clean up the simulated spells
       if(associated(work(i)%first)) then
          call remove_spell(work(i)%first)
          work(i)%first=>null()
          work(i)%last=>null()
       end if
    end do

    do i=1,dim_J
       if(associated(firm(i)%first)) then
          firm(i)%first=>null()
          firm(i)%last=>null()
       end if
    end do
    call modelpar_deallocate(par)
  end subroutine cleansim

  subroutine simdat_ten_exp(seed,ser,sim_ck,sim_cl)
    !this simulation expects par to contain 8 categories as well as an intial distribution over states.
    !It takes dimensions from par. Simulation does not simulate permanent worker and firm heterogeneity.
    !Initializes by young and untenured. 
    use mpimod
    use tools
    implicit none
    integer, intent(in) :: seed, sim_ck, sim_cl
    logical, intent(in) :: ser !if true run simulation serially.
    type simstate
       integer :: ftype
       integer :: ten
    end type simstate
    integer(4) :: i,j !counters
    integer(4) :: kk,ll,xx,lll,yy
    integer(4) :: widcdf(0:sim_ck,1),fidcdf(0:sim_cl)
    real(8) :: inicdf(sim_ck,0:sim_cl+1) !Contains k type conditional steady state match cdf. sscdf(.,0) is by def zero.
    real(8) :: destcdf(sim_ck,0:sim_cl,0:1,0:3,0:sim_cl+1) !(k,l) conditional destination cdf.
    real(8) :: haz(sim_ck,0:sim_cl,0:1,0:3) !(k,l,ten,exp) conditional hazard rate
    real(8) :: inim(sim_ck,0:sim_cl) !m(l |k,x)

    !Panel in estim module is populated
    dim_I=mydim_I
    dim_J=mydim_J
    true_ck=sim_ck
    true_cl=sim_cl
    if(allocated(work)) deallocate(work,firm)
    allocate(work(dim_I),firm(0:dim_J))

    !Initialize simulation related objects
    do i=1,80
       yrtoweek(i)=(i-1)*yr+1
    end do
    !set RNG
    call setseed(kisseed(2345678+2*seed,87656789+200*seed,5434567+100*seed,43212345+500*seed))
    !First, take care of preliminaries. Determine distributions and firm types.

    !rescale m(k,l|x) to get m(l|k,x)
    do kk=1,par%dim%k
       inim(kk,:)=par%m(kk,:,1)/par%pk(kk) !here, the 1 index selects young and short tenure.
    end do
    do kk=1,par%dim%k
       inim(kk,:)=inim(kk,:)/sum(inim(kk,:))
    end do
    do kk=1,par%dim%k
       inicdf(kk,0)=0.d0
       lll=0
       do ll=0,par%dim%l
          lll=lll+1
          ! par%m is now m(k,l|x) , we want m(l|k,x)
          inicdf(kk,lll)=inicdf(kk,lll-1)+inim(kk,ll) !simulate from young-short tenure onwards
       end do
       inicdf(kk,:)=inicdf(kk,:)/inicdf(kk,par%dim%l+1)

       do yy=0,3 !experience level
          do xx=0,1 !xx=0 is the low tenure state, and xx=1 is the high tenure state.
             do ll=0,par%dim%l
                lll=lll+1
                haz(kk,ll,xx,yy)=1.d0-par%Mbar(kk,ll,yy+1+xx*4) !expecting 8 categories
                destcdf(kk,ll,xx,yy,0)=0.0d0
                do j=0,par%dim%l
                   destcdf(kk,ll,xx,yy,j+1)=destcdf(kk,ll,xx,yy,j)+par%MM(kk,ll,j,yy+1+xx*4) !mym(kk,ll,j)
                end do
                if(abs(sum(par%MM(kk,ll,:,yy+1+xx*4))-destcdf(kk,ll,xx,yy,par%dim%l+1))>1.0e-15) then
                   !allow for some numerical precision differences
                   stop 'destcdf calculation failed'
                end if
                destcdf(kk,ll,xx,yy,:)=destcdf(kk,ll,xx,yy,:)/destcdf(kk,ll,xx,yy,par%dim%l+1)
             end do
          end do
       end do
    end do
    do kk=1,par%dim%k
       if(inicdf(kk,par%dim%l+1)<1.d0) then
          write(6,'(I5,f20.10)') 1,inicdf(kk,par%dim%l+1)
          stop
       end if
    end do

    fidcdf(0)=0
    forall(ll=1:sim_cl-1) fidcdf(ll)=int(sum(par%q(1:ll))*dble(dim_J))
    fidcdf(sim_cl)=dim_J

    firm(0)%l=0
    ll=1
    do j=1,dim_J
       if(j .gt. fidcdf(ll)) ll=ll+1
       if(ll>sim_cl) then
          write(6,'(4i10)') j,dim_j,ll,fidcdf(ll-1)
          stop 'type index exceeds number of types.'
       end if
       firm(j)%l=ll
    end do

    !Now, for the panel simulation and assignment of types to workers.

    kk=1
    do i=1,dim_I
       if(i .gt. sum(par%pk(1:kk))*dim_I+1) kk=kk+1
       work(i)%z=1
       work(i)%k=kk !worker i's type
       if(kk>par%dim%k) print *,"Warning: kk greater than dim%k",kk,i
       if(ser) then !simulate the worker's spell history
          call spellsim(inidraw(work(i)%k),1,.true.)
       else
          if(mod(i-1,numprocs)==myid) call spellsim(inidraw(work(i)%k),1,.true.)
       end if
    end do
  contains
    recursive subroutine spellsim(now,start,create) !use worker index i from parent.
      implicit none
      type(simstate), intent(in) :: now
      integer, intent(in) :: start
      logical, intent(in) :: create
      integer :: mycount,myten,myexp !myexp runs from 0 to 3
      integer :: dur,durnc,startyr,endyr,duryr
      integer :: dur_at_haz_change,dur_at_left_cens
      if(create) then
         call newspell !creates a spell and makes this point to it.
         this%wid=i
         this%fid=fiddraw(now%ftype) !random draw of firm type from ss, then assign firm id from group of type ll firms.
         this%ten=0
         if(firm(this%fid)%l .ne. now%ftype) then
            write(6,'(3I10)') this%fid,firm(this%fid)%l,now%ftype
            stop 'type assignment failed'
         end if
      end if

      !map into tenure category
      myten=1
      if(now%ftype>0) then
         if(now%ten<101) myten=0
      else
         if(now%ten<27) myten=0
      end if

      !map into experience category {0,1,2,3}
      myexp=0
      if(start+now%ten>260) myexp=myexp+1 !5-10 year cat
      if(start+now%ten>520) myexp=myexp+1 !10=15 year cat
      if(start+now%ten>780) myexp=myexp+1 !>15 year cat

      !find shortest duration cutoff until haz changes
      dur_at_left_cens=sim_t-start
      select case(myexp)
      case(0)
         dur_at_haz_change=min(dur_at_left_cens,260-start)
      case(1)
         dur_at_haz_change=min(dur_at_left_cens,520-start)
      case(2)
         dur_at_haz_change=min(dur_at_left_cens,780-start)
      case(3)
         dur_at_haz_change=dur_at_left_cens
      end select
      if(myten==0) then !haz can change due to tenure changing category
         if(now%ftype>0) then
            dur_at_haz_change=min(dur_at_haz_change,100)
         else
            dur_at_haz_change=min(dur_at_haz_change,26)
         end if
      end if

      durnc=now%ten+int(log(1.0d0-kissuni())/log(1.0d0-haz(work(i)%k,now%ftype,myten,myexp))) !draw exponential duration and add to existing duration

      if(durnc>dur_at_haz_change .and. dur_at_haz_change<dur_at_left_cens) then
         call spellsim(simstate(now%ftype,dur_at_haz_change+1),start,.false.) !continue existing spell, but allow haz to be updated
      else !either spell ends at left censoring or by itself without hazard change
         dur=min(durnc,1+dur_at_left_cens) !duration of spell
         this%end=start+dur-1
         this%strt=start

         startyr=int(real(this%strt)/real(yr))+1
         if(this%strt<yrtoweek(startyr)) startyr=startyr-1
         this%yr=startyr

         endyr=int(real(this%end)/real(yr))+1
         if(this%end<yrtoweek(endyr)) endyr=endyr-1

         duryr=endyr-startyr+1
         if(duryr<=0) then
            print *,startyr,endyr,this%end
            stop 'this will end badly.'
         end if
         allocate(this%wage(duryr),this%tenure(duryr))
         this%wage=0.d0
         if(now%ftype>0) call wagedraw(duryr,start,work(i)%k,firm(this%fid)%l,this%wage,this%tenure) !wage draw where first 2 years are low tenure
         call add_worker(work(i),this)
         call add_firm(firm(this%fid),this)
         if(this%end<sim_T) then
            ll=findndx(destcdf(work(i)%k,firm(this%fid)%l,myten,myexp,0:par%dim%l+1),kissuni(),0)-1 !determine type of destination firm.
            call spellsim(simstate(ll,1),this%end+1,.true.) !do a new spell duration and wages.
         end if
      end if
    end subroutine spellsim

    function inidraw(myk) result(res) !return a random firm type realization from myk's steady state match distribution.
      integer, intent(in) :: myk
      type(simstate) :: res
      integer :: ini,mykk
      ini=findndx(inicdf(myk,0:par%dim%l+1),kissuni(),1)-1 !1st element in inicdf(k,:) is g_k1, 2nd element is g_k1+g_k2, etc...
      res%ftype=ini
      res%ten=1
    end function inidraw

    function fiddraw(myl)
      integer, intent(in) :: myl
      integer :: fiddraw
      if(myl==0) then
         fiddraw=0
      else
         fiddraw=kissuni()*(fidcdf(myl)-fidcdf(myl-1))+fidcdf(myl-1)+1
         if(fiddraw==0) stop "drew fid=0 for employer."
      end if
    end function fiddraw

    subroutine wagedraw(mydur,mystart,myk,myl,reswage,resten) !normally distributed random draws with variance par%sigmaw(myk,myl)*2 and mean par%muw(myk,myl)
      integer, intent(in) :: mydur,mystart,myl,myk
      real(4), dimension(mydur), intent(out) :: reswage
      integer(1), dimension(mydur), intent(out) :: resten
      integer :: myi,myten,myexp,mysigten
      real(8) :: prev_ar, new_ar, epsilon
      do myi=1,mydur
         if(myi==1) then
            myten=0
            mysigten=1
         elseif(myi==2) then
            myten=0
            mysigten=1
         else
            myten=1
            mysigten=2
         end if
         myexp=0
         if(mystart+52*myi .gt. 260) myexp=myexp+1
         if(mystart+52*myi .gt. 520) myexp=myexp+1
         if(mystart+52*myi .gt. 780) myexp=myexp+1
         resten(myi)=myten
         if(myi==1) then
            prev_ar = kissgaus()*par%sigmaw(myk,myl,myexp+1+mysigten*4)
            reswage(myi) = prev_ar+par%muw(myk,myl,myexp+1+myten*4)
         else
            epsilon = kissgaus()*par%sigmaw(myk,myl,myexp+1+mysigten*4)
            new_ar = par%rho*prev_ar+epsilon
            reswage(myi) = new_ar+par%muw(myk,myl,myexp+1+myten*4)
            prev_ar = new_ar
         end if
      end do
    end subroutine wagedraw
  end subroutine simdat_ten_exp

  recursive subroutine SortCF(ch,klvar) !Counterfactual parameter settings for sorting counterfactuals
    integer, intent(in) :: ch !which sorting channel (1-4)
    integer, intent(in) :: klvar !which kind of variation to eliminate (1 - no k, 2 - no l, 3 - no k and no l)
    integer :: xi,li,ki

    select case(ch) !which channel
    case(0) !benchmark
       select case(klvar)
       case(1)
       case(2)
       case(3)
       end select
    case(1) !gamma
       select case(klvar)
       case(1)
          do xi=1,par%dim%het
             do li=1,par%l
                par%gam(:,li,xi)=dot_product(par%gam(:,li,xi),par%pk)
             end do
          end do
       case(2) !Removing l variation in gamma also removes k variation
          par%gam=dble(1)/dble(par%l)
       case(3)
          par%gam=dble(1)/dble(par%l)
       end select
       !rescale so that gam sum to one across ell
       do xi=1,par%dim%het
          forall(ki=1:par%k) par%gam(ki,:,xi)=par%gam(ki,:,xi)/sum(par%gam(ki,:,xi))
       end do
    case(2) !delta
       select case(klvar)
       case(1)
          do xi=1,par%dim%het
             do li=1,par%l
                par%del(:,li,xi)=dot_product(par%del(:,li,xi),par%pk)
             end do
          end do
       case(2)
          do xi=1,par%dim%het
             do ki=1,par%k
                par%del(ki,:,xi)=dot_product(par%del(ki,:,xi),par%q(1:par%l))
             end do
          end do
       case(3)
          call SortCF(2,2) !no l variation
          call SortCF(2,1) !no k variation
       end select
    case(3) !lambda
       select case(klvar)
       case(1)
          do xi=1,par%dim%het
             do li=1,par%l
                par%lam(:,li,xi)=dot_product(par%lam(:,li,xi),par%pk)
             end do
          end do
       case(2)
          do xi=1,par%dim%het
             do ki=1,par%k
                par%lam(ki,:,xi)=dot_product(par%lam(ki,:,xi),par%q(1:par%l))
             end do
          end do
       case(3)
          call SortCF(3,1)
          call SortCF(3,2)
       end select
    case(4) !psi
       select case(klvar)
       case(1)
          do xi=1,par%dim%het
             do li=1,par%l
                par%psi(:,li,xi)=dot_product(par%psi(:,li,xi),par%pk)
             end do
          end do
       case(2)
          do xi=1,par%dim%het
             do ki=1,par%k
                par%psi(ki,:,xi)=dot_product(par%psi(ki,:,xi),par%q(1:par%l))
             end do
          end do
       case(3)
          call SortCF(4,1)
          call SortCF(4,2)
       end select
    case(5) !gam-del
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(2,1)
       case(2)
          call SortCF(1,2)
          call SortCF(2,2)
       case(3)
          call SortCF(1,1)
          call SortCF(2,1)
          call SortCF(1,2)
          call SortCF(2,2)
       end select
    case(6) !gam-lam
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(3,1)
       case(2)
          call SortCF(1,2)
          call SortCF(3,2)
       case(3)
          call SortCF(1,1)
          call SortCF(3,1)
          call SortCF(1,2)
          call SortCF(3,2)
       end select
    case(7) !gam-psi
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(4,1)
       case(2)
          call SortCF(1,2)
          call SortCF(4,2)
       case(3)
          call SortCF(1,1)
          call SortCF(4,1)
          call SortCF(1,2)
          call SortCF(4,2)
       end select
    case(8) !del-lam
       select case(klvar)
       case(1)
          call SortCF(2,1)
          call SortCF(3,1)
       case(2)
          call SortCF(2,2)
          call SortCF(3,2)
       case(3)
          call SortCF(2,1)
          call SortCF(3,1)
          call SortCF(2,2)
          call SortCF(3,2)
       end select
    case(9) !del-psi
       select case(klvar)
       case(1)
          call SortCF(2,1)
          call SortCF(4,1)
       case(2)
          call SortCF(2,2)
          call SortCF(4,2)
       case(3)
          call SortCF(2,1)
          call SortCF(4,1)
          call SortCF(2,2)
          call SortCF(4,2)
       end select
    case(10) !psi-lam
       select case(klvar)
       case(1)
          call SortCF(3,1)
          call SortCF(4,1)
       case(2)
          call SortCF(3,2)
          call SortCF(4,2)
       case(3)
          call SortCF(3,1)
          call SortCF(4,1)
          call SortCF(3,2)
          call SortCF(4,2)
       end select
    case(11) !gam-del-lam
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(2,1)
          call SortCF(3,1)
       case(2)
          call SortCF(1,2)
          call SortCF(2,2)
          call SortCF(3,2)
       case(3)
          call SortCF(1,1)
          call SortCF(2,1)
          call SortCF(3,1)
          call SortCF(1,2)
          call SortCF(2,2)
          call SortCF(3,2)
       end select
    case(12) !gam-del-psi
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(2,1)
          call SortCF(4,1)
       case(2)
          call SortCF(1,2)
          call SortCF(2,2)
          call SortCF(4,2)
       case(3)
          call SortCF(1,1)
          call SortCF(2,1)
          call SortCF(4,1)
          call SortCF(1,2)
          call SortCF(2,2)
          call SortCF(4,2)
       end select
    case(13) !gam-lam-psi
       select case(klvar)
       case(1)
          call SortCF(1,1)
          call SortCF(3,1)
          call SortCF(4,1)
       case(2)
          call SortCF(1,2)
          call SortCF(3,2)
          call SortCF(4,2)
       case(3)
          call SortCF(1,1)
          call SortCF(3,1)
          call SortCF(4,1)
          call SortCF(1,2)
          call SortCF(3,2)
          call SortCF(4,2)
       end select
    case(14) !lam-del-psi
       select case(klvar)
       case(1)
          call SortCF(2,1)
          call SortCF(3,1)
          call SortCF(4,1)
       case(2)
          call SortCF(2,2)
          call SortCF(3,2)
          call SortCF(4,2)
       case(3)
          call SortCF(2,1)
          call SortCF(3,1)
          call SortCF(4,1)
          call SortCF(2,2)
          call SortCF(3,2)
          call SortCF(4,2)
       end select
    end select

    call M_update(par)

  end subroutine SortCF
end module datbysim
